import dayjs from 'dayjs';
import DayjsCustomParseFormat from 'dayjs/plugin/customParseFormat.js';
dayjs.extend(DayjsCustomParseFormat);
import ipaddrjs from 'ipaddr.js';
const { fromByteArray: buf2ipaddr } = ipaddrjs;
import 'xshell/prototype.browser.js';
import { blue, cyan, green, grey, magenta } from 'xshell/chalk.browser.js';
import { concat, assert, Lock, genid } from 'xshell/utils.browser.js';
import { connect_websocket } from 'xshell/net.browser.js';
import { t } from './i18n/index.js';
import { DdbDecimal128Serializor } from './data-types/decimal-128.js';
import { is_decimal_type, is_decimal_null_value } from './shared/utils/decimal-type.js';
import { nulls, DdbChartType, DdbDurationUnit, DdbForm, DdbFunctionType, DdbType } from './shared/constants.js';
export * from './shared/constants.js';
/** 可以表示所有 DolphinDB 数据库中的数据类型  Can represent data types in all DolphinDB databases */
export class DdbObj {
    static dec = new TextDecoder('utf-8');
    static enc = new TextEncoder();
    /** 维护已解析的 symbol base，比如流数据中后续的 symbol 向量可能只发送一个 base.id, base.size == 0, 依赖之前发送的 symbol base ？
        只是暂存，如果一张表有多个 symbol 列，可能这个 symbol base 会被复用，不同的对象之间 symbol base 一般不复用
    */
    static symbol_bases = {};
    /** little endian (client) */
    static le_client = Boolean(new Uint8Array(Uint32Array.of(1).buffer)[0]);
    /** 是否为小端 (little endian) */
    le = DdbObj.le_client;
    /** 数据形式 https://www.dolphindb.cn/cn/help/DataTypesandStructures/DataForms/index.html */
    form;
    /** 数据类型 https://www.dolphindb.cn/cn/help/DataTypesandStructures/DataTypes/index.html */
    type;
    /** 占用 parse 时传入的 buf 的长度 */
    length;
    /** table name / column name */
    name;
    /**
        最低维、第 1 维
        - vector: rows = n, cols = 1
        - pair:   rows = 2, cols = 1
        - matrix: rows = n, cols = m
        - set:    同 vector
        - dict:   包含 keys, values 向量
        - table:  同 matrix
    */
    rows;
    /** 第 2 维 */
    cols;
    /** 实际数据。不同的 DdbForm, DdbType 使用 DdbValue 中不同的类型来表示实际数据
        The actual data. Different DdbForm, DdbType use different types in DdbValue to represent actual data */
    value;
    /** 原始二进制数据，仅在 parse_object 为 false 时通过 parse_message 生成的顶层对象有这个属性 */
    buffer;
    constructor(data) {
        Object.assign(this, data);
    }
    static parse(buf, le) {
        if (!buf.length)
            return new this({
                le,
                form: DdbForm.scalar,
                type: DdbType.void,
                length: 0,
                value: null
            });
        const type = buf[0];
        const form = buf[1];
        if (buf.length <= 2)
            return new this({
                le,
                form,
                type,
                length: 2,
                value: null,
            });
        // set 里面 data 嵌套了一个 vector, 跳过 vector 的 type 和 form
        const i_data = form === DdbForm.set ? 4 : 2;
        const buf_data = buf.subarray(i_data);
        switch (form) {
            case DdbForm.scalar: {
                const [length, value] = this.parse_scalar(buf_data, le, type);
                return new this({
                    le,
                    form,
                    type,
                    length: i_data + length,
                    value,
                });
            }
            case DdbForm.vector:
            case DdbForm.pair:
            case DdbForm.set: {
                let vector = this.parse_vector(buf_data, le, type);
                vector.length += i_data;
                vector.form = form;
                return vector;
            }
            case DdbForm.table: {
                // table([
                //     [1, 2] as a,
                //     [1, 2] as b
                // ])
                // <Buffer 
                // 00 06 form = table
                // 02 00 00 00 02 00 00 00 rows = 2, cols = 2
                // 00 行名称
                // 61 00 62 00 列名称 a, b
                // 04 01 form = vector, type = int
                // 02 00 00 00 01 00 00 00 cols = 2, rows = 1
                // 01 00 00 00 02 00 00 00 
                // 04 01 
                // 02 00 00 00 01 00 00 00 
                // 01 00 00 00 02 00 00 00>
                const dv = new DataView(buf.buffer, buf.byteOffset + i_data);
                const rows = dv.getUint32(0, le);
                const cols = dv.getUint32(4, le);
                const i_name_tail = buf_data.indexOf(0, 8);
                const name = this.dec.decode(buf_data.subarray(8, i_name_tail));
                const i_items_start = i_name_tail + 1;
                const [len_items, colnames] = this.parse_vector_items(buf_data.subarray(i_items_start), le, DdbType.string, cols);
                let value = new Array(cols);
                let i_start = i_items_start + len_items;
                for (let i = 0; i < cols; i++) {
                    const type = buf_data[i_start];
                    let col = this.parse_vector(buf_data.subarray(i_start + 2), le, type);
                    col.length += 2;
                    col.name = colnames[i];
                    value[i] = col;
                    i_start += col.length;
                }
                return new this({
                    le,
                    form,
                    type,
                    length: i_data + i_start,
                    name,
                    rows,
                    cols,
                    value,
                });
            }
            case DdbForm.dict:
            case DdbForm.chart: {
                // <Buffer 19 05 type = any, form = dict
                // 12 01 keys.type = string, keys.form = vector
                // 03 00 00 00 01 00 00 00 keys.cols = 3, keys.rows = 1
                // 63 00 62 00 61 00 
                // 19 01 values.type = any, values.form = vector
                // 03 00 00 00 01 00 00 00 values.cols = 3, values.rows = 1
                // 04 00 03 00 00 00 04 00 02 00 00 00 04 00 01 00 00 00>
                let keys = this.parse_vector(buf_data.subarray(2), le, buf_data[0]);
                keys.length += 2;
                let values = this.parse_vector(buf_data.subarray(keys.length + 2), le, buf_data[keys.length]);
                values.length += 2;
                let dict = new this({
                    le,
                    form: DdbForm.dict,
                    type,
                    length: i_data + keys.length + values.length,
                    rows: keys.rows,
                    cols: 2,
                    value: [
                        keys,
                        values
                    ],
                });
                if (form === DdbForm.dict)
                    return dict;
                else {
                    const { chartType: type, stacking, binStart: bin_start, binEnd: bin_end, binCount: bin_count, title: titles, extras, data, ...others } = dict.to_dict();
                    const [chart, x_axis, y_axis] = titles.value;
                    dict.form = DdbForm.chart;
                    dict.value = {
                        type: type.value,
                        stacking: stacking.value,
                        titles: {
                            chart,
                            x_axis,
                            y_axis,
                        },
                        ...bin_start ? { bin_start, bin_end, } : {},
                        ...bin_count ? { bin_count } : {},
                        ...extras ? (() => {
                            const { multiYAxes: multi_y_axes = false, ...extras_others } = extras.to_dict({ strip: true });
                            return {
                                extras: {
                                    multi_y_axes,
                                    ...extras_others,
                                }
                            };
                        })() : {},
                        data,
                        ...others,
                    };
                    return dict;
                }
            }
            case DdbForm.matrix: {
                // rename!(
                //     1..9$3:3,
                //     [1, 2, 3],
                //     ['c1', 'c2', 'c3']
                // )
                // <Buffer 04 03 type = int, form = matrix
                // 03 has_row_label (& 0x01) = 1, has_col_label (& 0x02) = 1
                // row labels
                // 04 01 type = int, form = vector
                // 03 00 00 00 01 00 00 00 rows = 3, cols = 1
                // 01 00 00 00 02 00 00 00 03 00 00 00 vector values
                // col labels
                // 12 01 type = string, form = vector
                // 03 00 00 00 01 00 00 00 rows = 3, cols = 1
                // 63 31 00 63 32 00 63 33 00 
                // matrix data
                // 04 03 type = matrix.type, form = matrix
                // 03 00 00 00 03 00 00 00 rows = 3, cols = 3
                const dv = new DataView(buf.buffer, buf.byteOffset + i_data);
                const label_flags = buf_data[0];
                const has_row_labels = Boolean(label_flags & 0x01);
                const has_col_labels = Boolean(label_flags & 0x02);
                let row_labels = null;
                let col_labels = null;
                let offset = 1;
                if (has_row_labels) {
                    row_labels = this.parse_vector(buf_data.subarray(offset + 2), le, buf_data[offset]);
                    row_labels.length += 2;
                    offset += row_labels.length;
                }
                if (has_col_labels) {
                    col_labels = this.parse_vector(buf_data.subarray(offset + 2), le, buf_data[offset]);
                    col_labels.length += 2;
                    offset += col_labels.length;
                }
                assert(buf_data[offset] === type, 'matrix.datatype === matrix.type');
                const rows = dv.getUint32(offset + 2, le);
                const cols = dv.getUint32(offset + 6, le);
                const [len_items, data] = this.parse_vector_items(buf_data.subarray(offset + 10), le, type, rows * cols // 假设小于 2**32
                );
                offset += 10 + len_items;
                return new this({
                    le,
                    form,
                    type,
                    length: i_data + offset,
                    rows,
                    cols,
                    value: {
                        rows: row_labels,
                        cols: col_labels,
                        data,
                    },
                });
            }
            default:
                return new this({
                    le,
                    form,
                    type,
                    length: i_data + buf_data.length,
                    value: buf_data
                });
        }
    }
    static parse_scalar(buf, le, type) {
        switch (type) {
            case DdbType.void:
                return [1, null];
            case DdbType.bool: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const value = dv.getInt8(0);
                return [1, value === nulls.int8 ? null : Boolean(value)];
            }
            case DdbType.char: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const value = dv.getInt8(0);
                return [1, value === nulls.int8 ? null : value];
            }
            case DdbType.short: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const value = dv.getInt16(0, le);
                return [2, value === nulls.int16 ? null : value];
            }
            case DdbType.int:
            // datetime
            case DdbType.date:
            case DdbType.month:
            case DdbType.time:
            case DdbType.minute:
            case DdbType.second:
            case DdbType.datetime:
            case DdbType.datehour: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const value = dv.getInt32(0, le);
                return [4, value === nulls.int32 ? null : value];
            }
            case DdbType.float: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const value = dv.getFloat32(0, le);
                return [4, value === nulls.float32 ? null : value];
            }
            case DdbType.double: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const value = dv.getFloat64(0, le);
                return [8, value === nulls.double ? null : value];
            }
            case DdbType.long:
            // timestamp
            case DdbType.timestamp:
            case DdbType.nanotime:
            case DdbType.nanotimestamp: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const value = dv.getBigInt64(0, le);
                return [8, value === nulls.int64 ? null : value];
            }
            case DdbType.string:
            case DdbType.symbol:
            case DdbType.code:
            case DdbType.handle:
            // sqlDS 函数会返回包含 datasource 的 any vector
            case DdbType.datasource:
            case DdbType.functiondef:
            // mysql 插件 connect 方法会返回 resource 类型的对象
            case DdbType.resource: {
                const i_head = type === DdbType.functiondef ? 1 : 0;
                let i_zero = buf.indexOf(0, i_head);
                let i_end; // 整个字符串（包括 0）的末尾，excluding
                if (i_zero === -1)
                    i_end = i_zero = buf.length;
                else
                    i_end = i_zero + 1;
                // 调整了 i_zero 到字符串（不包括 0）的末尾，excluding
                const str = this.dec.decode(buf.subarray(i_head, i_zero));
                return [
                    i_end,
                    type === DdbType.functiondef ?
                        {
                            type: buf[0],
                            name: str
                        }
                        :
                            str
                ];
            }
            case DdbType.uuid:
            case DdbType.ipaddr:
            case DdbType.int128:
                return [16, buf.slice(0, 16)];
            case DdbType.blob: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const length = dv.getUint32(0, le);
                return [4 + length, buf.slice(4, 4 + length)];
            }
            case DdbType.complex:
            case DdbType.point: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                return [16, [dv.getFloat64(0, le), dv.getFloat64(8, le)]];
            }
            case DdbType.duration: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                return [8, { unit: dv.getUint32(4, le), data: dv.getInt32(0, le) }];
            }
            case DdbType.decimal32: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const data = dv.getInt32(4, le);
                return [8, { scale: dv.getInt32(0, le), data: data === nulls.int32 ? null : data }];
            }
            case DdbType.decimal64: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const data = dv.getBigInt64(4, le);
                return [12, { scale: dv.getInt32(0, le), data: data === nulls.int64 ? null : data }];
            }
            case DdbType.decimal128: {
                return DdbDecimal128Serializor.parse_as_scalar(buf, le);
            }
            default:
                throw new Error(String(DdbType[type] || type) + t(' 暂时不支持解析'));
        }
    }
    /** parse: rows, cols, items
        返回的 ddbobj.length 不包括 vector 的 type 和 form
    */
    static parse_vector(buf, le, type) {
        const dv = new DataView(buf.buffer, buf.byteOffset);
        const rows = dv.getUint32(0, le);
        let i_items_start = 8;
        if (type < 64 || type >= 128) {
            const [len_items, value] = this.parse_vector_items(buf.subarray(i_items_start), le, type, rows);
            return new this({
                le,
                form: DdbForm.vector,
                type,
                length: i_items_start + len_items,
                cols: 1,
                rows,
                value,
            });
        }
        // array vector
        // av = array(INT[], 0, 3)
        // append!(av, [1..4])
        // append!(av, [1..70000])
        // av
        // <Buffer 44 01  type = array vector, form = vector
        // 02 00 00 00 74 11 01 00 rows = 2, cols = 70004 (0x011174)
        // block 0
        // 01 00 block.rows = 1
        // 04 block.unit = 4
        // 00 reserved
        // 04 00 00 00 block.lengths = [4]
        // 01 00 00 00 02 00 00 00 03 00 00 00 04 00 00 00 block.data
        // block 1
        // 01 00 block.rows = 1
        // 04 block.unit = 4
        // 00 reserved
        // 70 11 01 00 block.lengths = [70000 (0x00011170)]
        // 01 00 00 00 02 00 00 00 ... 279992 more bytes> block.data
        // decimal array vector
        // a = array(DECIMAL32(2)[], 0, 10)
        // append!(a, [
        //     [1, 2, 3], 
        //     [4, 5], 
        //     [6, 7, 8],
        //     [9, 10]
        // ])
        // print(a)
        // a
        // [[1.00,2.00,3.00],[4.00,5.00],[6.00,7.00,8.00],[9.00,10.00]]
        // <Buffer 65 01 type = 101 = 64 + 37 (decimal32) , form = vector
        // 04 00 00 00 0a 00 00 00 rows = 4, cols = 10
        // scale (只有 decimal32 才有)
        // 02 00 00 00 scale = 2
        // block 0
        // 04 00 block.rows = 4
        // 01 block.unit = 1
        // 00 reserved
        // 03 02 03 02 block.lengths = [3, 2, 3, 2]
        // 64 00 00 00 c8 00 00 00 2c 01 00 00 90 01 00 00 f4 01 00 00 58 02 00 00 bc 02 00 00 ... 12 more bytes> [100, 200, 300, ...]
        // block 1
        // ...
        const type_ = type - 64;
        const cols = dv.getUint32(4, le);
        let blocks = [];
        // decimal 会在所有 blocks 之前多一个 scale
        if (is_decimal_type(type_)) {
            blocks.scale = dv.getInt32(i_items_start, le);
            i_items_start += 4;
        }
        let i_block_start = i_items_start;
        // 解析一个 block
        for (let i_row = 0; i_row < rows;) {
            /** 对应 array vector 中元素个数 */
            const rows = dv.getUint16(i_block_start, le);
            /** 每个 length 占用的字节数 */
            const unit = dv.getUint8(i_block_start + 2);
            /** array vector 每个元素的子元素长度 */
            let lengths;
            const i_lengths_start = i_block_start + 4;
            const i_data_start = i_lengths_start + rows * unit;
            const lengths_buf = buf.slice(i_lengths_start, i_data_start);
            switch (unit) {
                case 1:
                    lengths = lengths_buf;
                    break;
                case 2:
                    lengths = new Uint16Array(lengths_buf.buffer);
                    break;
                case 4:
                    lengths = new Uint32Array(lengths_buf.buffer);
                    break;
                default:
                    throw new Error(t('array vector 存在非法 unit: {{unit}}', { unit }));
            }
            let total_length = 0;
            for (const x of lengths)
                total_length += x;
            let len_items;
            let data;
            switch (type_) {
                case DdbType.decimal32:
                    len_items = total_length * 4;
                    data = new Int32Array(buf.buffer.slice(buf.byteOffset + i_data_start, buf.byteOffset + i_data_start + len_items));
                    break;
                case DdbType.decimal64:
                    len_items = total_length * 8;
                    data = new BigInt64Array(buf.buffer.slice(buf.byteOffset + i_data_start, buf.byteOffset + i_data_start + len_items));
                    break;
                case DdbType.decimal128:
                    [len_items, data] = DdbDecimal128Serializor.parse_as_same_type_vector_values(buf, i_data_start, total_length);
                    break;
                default:
                    [len_items, data] = this.parse_vector_items(buf.subarray(i_data_start), le, type - 64, total_length);
            }
            blocks.push({
                unit,
                rows,
                lengths,
                data: data
            });
            i_block_start = i_data_start + len_items;
            i_row += rows;
        }
        return new this({
            le,
            form: DdbForm.vector,
            type,
            length: i_block_start,
            cols,
            rows,
            value: blocks
        });
    }
    /** 有可能没有字节对齐，不能直接使用原有 message 的 arraybuffer, 统一复制出来，让原有 arraybuffer 被回收掉比较好 */
    static parse_vector_items(buf, le, type, length) {
        switch (type) {
            case DdbType.void:
                return [0, null];
            case DdbType.bool:
            case DdbType.char:
                return [
                    length,
                    new Int8Array(buf.buffer.slice(buf.byteOffset, buf.byteOffset + length))
                ];
            case DdbType.short:
                return [
                    2 * length,
                    new Int16Array(buf.buffer.slice(buf.byteOffset, buf.byteOffset + 2 * length))
                ];
            case DdbType.int:
            // datetime
            case DdbType.date:
            case DdbType.month:
            case DdbType.time:
            case DdbType.minute:
            case DdbType.second:
            case DdbType.datetime:
            case DdbType.datehour:
                return [
                    4 * length,
                    new Int32Array(buf.buffer.slice(buf.byteOffset, buf.byteOffset + 4 * length))
                ];
            case DdbType.float:
                return [
                    4 * length,
                    new Float32Array(buf.buffer.slice(buf.byteOffset, buf.byteOffset + 4 * length))
                ];
            case DdbType.double:
                return [
                    8 * length,
                    new Float64Array(buf.buffer.slice(buf.byteOffset, buf.byteOffset + 8 * length))
                ];
            case DdbType.long:
            // timestamp
            case DdbType.timestamp:
            case DdbType.nanotime:
            case DdbType.nanotimestamp:
                return [
                    8 * length,
                    new BigInt64Array(buf.buffer.slice(buf.byteOffset, buf.byteOffset + 8 * length))
                ];
            case DdbType.string:
            case DdbType.symbol:
            case DdbType.handle:
            case DdbType.datasource:
            case DdbType.code: {
                let value = new Array(length);
                let i_head = 0, i_tail = i_head;
                for (let i = 0; i < length; i++) {
                    i_tail = buf.indexOf(0, i_head);
                    value[i] = this.dec.decode(buf.subarray(i_head, i_tail));
                    i_head = i_tail + 1;
                }
                return [i_head, value];
            }
            case DdbType.symbol_extended: {
                // <Buffer 91 01 type = symbol extended, form = vector
                // 05 00 00 00 01 00 00 00 row = 5, col = 1
                // buf:
                // 00 00 00 00 symbol base id = 0 (uint32)
                // 02 00 00 00 symbol base size = 2
                // 00 61 61 00 以 \0 分割的字符串
                // 01 00 00 00 01 00 00 00 01 00 00 00 01 00 00 00 01 00 00 00>
                const dv = new DataView(buf.buffer, buf.byteOffset);
                const base_id = dv.getUint32(0, le);
                const base_size = dv.getUint32(4, le);
                let base_length = 0;
                let base = this.symbol_bases[base_id];
                // base_size 为 0 时复用之前的 symbol base
                if (base_size) {
                    [base_length, base] = this.parse_vector_items(buf.subarray(8), le, DdbType.string, base_size);
                    this.symbol_bases[base_id] = base;
                }
                const value_start = 8 + base_length;
                const value_end = value_start + length * 4;
                const data = new Uint32Array(buf.buffer.slice(buf.byteOffset + value_start, buf.byteOffset + value_end));
                return [
                    value_end,
                    {
                        base_id,
                        base,
                        data
                    }
                ];
            }
            case DdbType.uuid:
            case DdbType.ipaddr:
            case DdbType.int128:
                return [
                    16 * length,
                    new Uint8Array(buf.buffer.slice(buf.byteOffset, buf.byteOffset + 16 * length))
                ];
            case DdbType.blob: {
                // <Buffer 20 01 type = blob, form = vector
                // 02 00 00 00 01 00 00 00 cols = 2, rows = 1
                // 04 00 00 00 61 62 63 64 
                // 04 00 00 00 61 62 63 64>
                let value = new Array(length);
                const dv = new DataView(buf.buffer, buf.byteOffset);
                let i_head = 0;
                for (let i = 0; i < length; i++) {
                    const sublen = dv.getUint32(i_head, le);
                    const i_blob_head = i_head + 4;
                    const i_blob_tail = i_blob_head + sublen;
                    value[i] = buf.slice(i_blob_head, i_blob_tail);
                    i_head = i_blob_tail;
                }
                return [i_head, value];
            }
            case DdbType.complex:
            case DdbType.point:
                return [
                    16 * length,
                    new Float64Array(buf.buffer.slice(buf.byteOffset, buf.byteOffset + 16 * length))
                ];
            case DdbType.any: {
                // [1, 2, 'a', 'aaa']
                // any[4](<Buffer 
                // 04 00 DdbType.int, DdbForm.scalar
                // 01 00 00 00 
                // 04 00 DdbType.int, DdbForm.scalar
                // 02 00 00 00 
                // 02 00 DdbType.char, DdbForm.scalar
                // 61 
                // 12 00 DdbType.string, DdbForm.scalar
                // 61 61 61 00>)
                let values = new Array(length);
                let i_head = 0;
                for (let i = 0; i < length; i++) {
                    const obj = this.parse(buf.subarray(i_head), le);
                    values[i] = obj;
                    i_head += obj.length;
                }
                return [
                    i_head,
                    values
                ];
            }
            // 25 01 type = decimal32, form = vector
            // 02 00 00 00 01 00 00 00
            // 00 00 00 00 scale = 0
            // 01 00 00 00 data[0] = 1
            // 3a 01 00 00 data[1] = 0x013a = 314
            case DdbType.decimal32: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                return [
                    4 + 4 * length,
                    {
                        scale: dv.getInt32(0, le),
                        data: new Int32Array(buf.buffer.slice(buf.byteOffset + 4, buf.byteOffset + 4 + 4 * length))
                    }
                ];
            }
            case DdbType.decimal64: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                return [
                    4 + 8 * length,
                    {
                        scale: dv.getInt32(0, le),
                        data: new BigInt64Array(buf.buffer.slice(buf.byteOffset + 4, buf.byteOffset + 4 + 8 * length))
                    }
                ];
            }
            case DdbType.decimal128:
                return DdbDecimal128Serializor.parse_as_vector_items(buf, length, le);
            // 以下情况时, DdbType.duration 实际会返回一个 any vector
            // [2y, 1M, 3d, 7H, 11m, 12s, 15ms, 16us, 17ns]
            // <Buffer 19 01 type = any, form = vector
            // 09 00 00 00 01 00 00 00 rows = 9, cols = 1
            // 24 00 type = DdbType.duration, form = scalar
            // 02 00 00 00 09 00 00 00 
            // 24 00 01 00 00 00 08 00 00 00 24 00 03 00 00 00 06 00 00 00 24 00 07 00 00 00 05 00 00 00 ... 50 more bytes>
            // 其余情况 (目前仅 pair) 下, 会返回特殊的 duration 序列化
            // 4 bytes data 4 bytes unit
            // 01 00 00 00 data: 1 
            // 01 00 00 00 unit: 1
            // 02 00 00 00 data: 2
            // 02 00 00 00 unit: 2
            case DdbType.duration: {
                const dv = new DataView(buf.buffer, buf.byteOffset);
                let durations = [];
                for (let i = 0; i < length; i++)
                    durations.push({
                        data: dv.getInt32(0 + 8 * i, le),
                        unit: dv.getInt32(4 + 8 * i, le)
                    });
                return [8 * length, durations];
            }
            default:
                throw new Error(t('vector<{{type}}> 暂时不支持解析', { type: String(DdbType[type] || type) }));
        }
    }
    pack() {
        const { form, type, value } = this;
        let header = new Uint8Array(2);
        header[0] = type;
        header[1] = form;
        const body = (() => {
            switch (form) {
                case DdbForm.scalar:
                    switch (type) {
                        case DdbType.void:
                            return [Uint8Array.of(1)];
                        case DdbType.bool:
                            return [
                                Int8Array.of(value === null ?
                                    nulls.int8
                                    :
                                        Number(value))
                            ];
                        case DdbType.char:
                            return [
                                Int8Array.of((value === null ?
                                    nulls.int8
                                    :
                                        value))
                            ];
                        case DdbType.short:
                            return [
                                Int16Array.of(value === null ?
                                    nulls.int16
                                    :
                                        value)
                            ];
                        case DdbType.int:
                        // datetime
                        case DdbType.date:
                        case DdbType.month:
                        case DdbType.time:
                        case DdbType.minute:
                        case DdbType.second:
                        case DdbType.datetime:
                        case DdbType.datehour:
                            return [
                                Int32Array.of(value === null ?
                                    nulls.int32
                                    :
                                        value)
                            ];
                        case DdbType.float:
                            return [
                                Float32Array.of(value === null ?
                                    nulls.float32
                                    :
                                        value)
                            ];
                        case DdbType.double:
                            return [
                                Float64Array.of(value === null ?
                                    nulls.double
                                    :
                                        value)
                            ];
                        case DdbType.long:
                        // timestamp
                        case DdbType.timestamp:
                        case DdbType.nanotime:
                        case DdbType.nanotimestamp:
                            return [
                                BigInt64Array.of(value === null ?
                                    nulls.int64
                                    :
                                        value)
                            ];
                        case DdbType.string:
                        case DdbType.symbol:
                        case DdbType.code:
                        case DdbType.handle:
                        case DdbType.datasource:
                        case DdbType.resource:
                            return [
                                DdbObj.enc.encode(value),
                                Uint8Array.of(0),
                            ];
                        case DdbType.functiondef:
                            return [
                                Uint8Array.of(value.type),
                                DdbObj.enc.encode(value.name),
                                Uint8Array.of(0)
                            ];
                        case DdbType.uuid:
                        case DdbType.ipaddr:
                        case DdbType.int128:
                            return [this.value];
                        case DdbType.blob:
                            return [Uint32Array.of(this.value.byteLength), this.value];
                        case DdbType.complex:
                        case DdbType.point:
                            return [Float64Array.from(this.value)];
                        case DdbType.duration: {
                            const { unit, data } = this.value;
                            return [Int32Array.of(data, unit)];
                        }
                        case DdbType.decimal32: {
                            const { scale, data } = this.value;
                            return [Int32Array.of(scale, data === null ? nulls.int32 : data)];
                        }
                        case DdbType.decimal64: {
                            const { scale, data } = this.value;
                            return [Int32Array.of(scale), BigInt64Array.of(data === null ? nulls.int64 : data)];
                        }
                        case DdbType.decimal128:
                            return DdbDecimal128Serializor.pack(this.value);
                        default:
                            throw new Error(String(DdbType[type] || type) + t(' 暂时不支持序列化'));
                    }
                case DdbForm.vector:
                case DdbForm.pair:
                    // pack array vector
                    if (form === DdbForm.vector && 64 <= type && type < 128)
                        return [
                            Uint32Array.of(this.rows, this.cols),
                            ...(is_decimal_type(type - 64)) ?
                                [Int32Array.of(this.value.scale)]
                                :
                                    [],
                            ...this.value.map(block => ([
                                Uint16Array.of(block.rows),
                                Uint8Array.of(block.unit, 0),
                                block.lengths,
                                block.data
                            ])).flat()
                        ];
                    return [
                        Uint32Array.of(this.rows, 1),
                        ...DdbObj.pack_vector_body(value, type, this.rows)
                    ];
                case DdbForm.set:
                    return [new DdbObj({ ...this, form: DdbForm.vector }).pack()];
                case DdbForm.table:
                    return [
                        Uint32Array.of(this.rows, this.cols),
                        DdbObj.enc.encode(this.name || ''),
                        // column names
                        Uint8Array.of(0),
                        ...DdbObj.pack_vector_body(this.value.map((col, i) => col.name || `col${i}`), DdbType.string, this.cols),
                        // column vectors
                        ...this.value.map(col => col.pack())
                    ];
                case DdbForm.dict:
                    return [value[0].pack(), value[1].pack()];
                case DdbForm.chart: {
                    const { type, stacking, bin_start, bin_end, bin_count, titles: { chart, x_axis, y_axis }, extras, data } = this.value;
                    const { value: [keys, values] } = new DdbDict({
                        chartType: new DdbInt(type),
                        stacking,
                        ...bin_start ? { binStart: bin_start, binEnd: bin_end } : {},
                        ...bin_count ? { binCount: bin_count } : {},
                        title: new DdbVectorString([chart, x_axis, y_axis]),
                        ...extras ? (() => {
                            const { multi_y_axes, ...extras_other } = extras;
                            return { extras: new DdbDict({ multiYAxes: multi_y_axes, ...extras_other }) };
                        })() : {},
                        data,
                    });
                    return [keys.pack(), values.pack()];
                }
                case DdbForm.matrix: {
                    const { rows, cols, data } = value;
                    return [
                        Uint8Array.of((rows ? 0x01 : 0x00) | (cols ? 0x02 : 0x00)),
                        ...rows ? [rows.pack()] : [],
                        ...cols ? [cols.pack()] : [],
                        Uint8Array.of(this.type, this.form),
                        Uint32Array.of(this.rows, this.cols),
                        ...DdbObj.pack_vector_body(data, this.type, this.rows * this.cols)
                    ];
                }
                default:
                    throw new Error(t('vector {{type}} 暂不支持序列化', { type: DdbType[type] || type }));
            }
        })();
        if (!body)
            return new Uint8Array(0);
        return concat([header, ...body]);
    }
    static pack_vector_body(value, type, length) {
        switch (type) {
            case DdbType.void:
                return [];
            case DdbType.bool:
                return [value];
            case DdbType.char:
                return [value];
            case DdbType.short:
                return [value];
            case DdbType.int:
            // datetime
            case DdbType.date:
            case DdbType.month:
            case DdbType.time:
            case DdbType.minute:
            case DdbType.second:
            case DdbType.datetime:
            case DdbType.datehour:
                return [value];
            case DdbType.float:
                return [value];
            case DdbType.double:
                return [value];
            case DdbType.long:
            // timestamp
            case DdbType.timestamp:
            case DdbType.nanotime:
            case DdbType.nanotimestamp:
                return [value];
            case DdbType.string:
            case DdbType.symbol:
            case DdbType.handle:
            case DdbType.datasource:
            case DdbType.code: {
                let bufs = new Array(length * 2);
                for (let i = 0; i < length; i++) {
                    bufs[2 * i] = this.enc.encode(value[i]);
                    bufs[2 * i + 1] = Uint8Array.of(0);
                }
                return bufs;
            }
            case DdbType.uuid:
            case DdbType.ipaddr:
            case DdbType.int128:
                return [value];
            case DdbType.blob: {
                let bufs = new Array(length * 2);
                for (let i = 0; i < length; i++) {
                    const blob_value = value[i];
                    bufs[2 * i] = Uint32Array.of(blob_value.length);
                    bufs[2 * i + 1] = blob_value;
                }
                return bufs;
            }
            case DdbType.complex:
            case DdbType.point:
                return [value];
            case DdbType.any: {
                // [1, 2, 'a', 'aaa']
                // any[4](<Buffer 
                // 04 00 DdbType.int, DdbForm.scalar
                // 01 00 00 00 
                // 04 00 DdbType.int, DdbForm.scalar
                // 02 00 00 00 
                // 02 00 DdbType.char, DdbForm.scalar
                // 61 
                // 12 00 DdbType.string, DdbForm.scalar
                // 61 61 61 00>)
                let bufs = new Array(length);
                for (let i = 0; i < length; i++)
                    bufs[i] = value[i].pack();
                return bufs;
            }
            case DdbType.symbol_extended: {
                const { base_id, base, data } = value;
                return [
                    Uint32Array.of(base_id || 0, base.length),
                    ...this.pack_vector_body(base, DdbType.string, base.length),
                    data
                ];
            }
            case DdbType.decimal32:
            case DdbType.decimal64:
            case DdbType.decimal128: {
                const { scale, data } = value;
                return [Int32Array.of(scale), data];
            }
            default:
                throw new Error(t('vector {{type}} 暂不支持序列化', { type: String(DdbType[type] || type) }));
        }
    }
    toString(options = { nullstr: true, quote: true }) {
        const type = this.inspect_type();
        const data = (() => {
            switch (this.form) {
                case DdbForm.scalar:
                    return format(this.type, this.value, this.le, options);
                case DdbForm.vector:
                case DdbForm.pair:
                case DdbForm.set: {
                    const form = this.form;
                    function format_array(items, ellipsis) {
                        const str_items = items.join(', ') + (ellipsis ? ', ...' : '');
                        return form === DdbForm.pair ?
                            str_items
                            :
                                str_items.bracket('square');
                    }
                    if (64 <= this.type && this.type < 128) { // array vector
                        // 因为 array vector 目前只支持：Logical, Integral（不包括 INT128, COMPRESS 类型）, Floating, Temporal
                        // 都对应 TypedArray 中的一格，所以直接根据 index 去取即可
                        // av = array(INT[], 0, 5)
                        // append!(av, [1..1])
                        // append!(av, [1..70000])
                        // append!(av, [1..1])
                        // append!(av, [1..500])
                        // ...
                        // av
                        const type_ = this.type - 64;
                        const limit = 10;
                        let array_items = new Array(Math.min(limit, this.rows));
                        let i_items = 0;
                        for (const { lengths, data } of this.value) {
                            let acc_len = 0;
                            for (const length of lengths) {
                                let items = new Array(Math.min(limit, length));
                                for (let i = 0; i < items.length; i++)
                                    switch (type_) {
                                        case DdbType.decimal32:
                                        case DdbType.decimal64:
                                        case DdbType.decimal128:
                                            const x = data[acc_len + i];
                                            if (is_decimal_null_value(type_, x)) {
                                                items[i] = '';
                                                break;
                                            }
                                            const { scale } = this.value;
                                            const s = String(x < 0 ? -x : x).padStart(scale, '0');
                                            const str = (x < 0 ? '-' : '') + (scale ? `${s.slice(0, -scale) || '0'}.${s.slice(-scale)}` : s);
                                            items[i] = options.colors ? green(str) : str;
                                            break;
                                        case DdbType.complex:
                                        case DdbType.point: {
                                            const index = acc_len + i;
                                            items[i] = format(type_, data.subarray(2 * index, 2 * (index + 1)), this.le, options);
                                            break;
                                        }
                                        case DdbType.uuid:
                                        case DdbType.int128:
                                        case DdbType.ipaddr: {
                                            const index = acc_len + i;
                                            items[i] = format(type_, data.subarray(16 * index, 16 * (index + 1)), this.le, options);
                                            break;
                                        }
                                        default:
                                            items[i] = format(type_, data[acc_len + i], this.le, options);
                                            break;
                                    }
                                array_items[i_items++] = format_array(items, length > limit);
                                acc_len += length;
                            }
                            if (i_items >= limit)
                                break;
                        }
                        return format_array(array_items, this.rows > limit);
                    }
                    switch (this.type) {
                        case DdbType.symbol_extended: {
                            const limit = 50;
                            const { base, data } = this.value;
                            let items = new Array(Math.min(limit, data.length));
                            for (let i = 0; i < items.length; i++)
                                items[i] = base[data[i]].quote('single');
                            return format_array(items, data.length > limit);
                        }
                        case DdbType.void:
                            return format(this.type, this.value, this.le, options);
                        case DdbType.uuid:
                        case DdbType.int128:
                        case DdbType.ipaddr: {
                            const limit = 10;
                            const value = this.value;
                            const len_data = value.length / 16;
                            let items = new Array(Math.min(limit, len_data));
                            for (let i = 0; i < items.length; i++)
                                items[i] = format(this.type, value.subarray(16 * i, 16 * (i + 1)), this.le, options);
                            return format_array(items, len_data > limit);
                        }
                        case DdbType.complex:
                        case DdbType.point: {
                            const limit = 20;
                            const value = this.value;
                            const len_data = value.length / 2;
                            let items = new Array(Math.min(limit, len_data));
                            for (let i = 0; i < items.length; i++)
                                items[i] = format(this.type, value.subarray(2 * i, 2 * (i + 1)), this.le, options);
                            return format_array(items, len_data > limit);
                        }
                        case DdbType.decimal32:
                        case DdbType.decimal64:
                        case DdbType.decimal128: {
                            const limit = 50;
                            const { data } = this.value;
                            let items = new Array(Math.min(limit, data.length));
                            for (let i = 0; i < items.length; i++)
                                items[i] = formati(this, i, options);
                            return format_array(items, data.length > limit);
                        }
                        default: {
                            const limit = 50;
                            let items = new Array(Math.min(limit, this.value.length));
                            for (let i = 0; i < items.length; i++)
                                items[i] = format(this.type, this.value[i], this.le, options);
                            return format_array(items, this.value.length > limit);
                        }
                    }
                }
                case DdbForm.chart:
                    return JSON.stringify(this.value, (key, value) => key === 'data' ? value.toString(options) : value, 4);
            }
            return String(this.value);
        })();
        return `${options?.colors ? blue(type) : type}(${this.name ? `'${this.name}', ` : ''}${data})`;
    }
    inspect_type() {
        const tname = DdbType[this.type];
        switch (this.form) {
            case DdbForm.scalar:
                if (this.type === DdbType.functiondef)
                    return `functiondef<${DdbFunctionType[this.value.type]}>`;
                return tname;
            case DdbForm.vector:
                if (64 <= this.type && this.type < 128)
                    return `${DdbType[this.type - 64]}[][${this.rows}]`;
                return `${tname}[${this.rows}]`;
            case DdbForm.pair:
                return `pair<${tname}>`;
            case DdbForm.set:
                return `set<${tname}>[${this.rows}]`;
            case DdbForm.table:
                return `table[${this.rows}r][${this.cols}c]`;
            case DdbForm.dict:
                return `dict<${DdbType[this.value[0].type]}, ${DdbType[this.value[1].type]}>[${this.rows}]`;
            case DdbForm.chart:
                return `chart<${DdbChartType[this.value.type]}>`;
            case DdbForm.matrix:
                return `matrix<${tname}>[${this.rows}r][${this.cols}c]`;
            default:
                return `${DdbForm[this.form]} ${tname}`;
        }
    }
    /** 自动转换 js string, boolean 为 DdbObj */
    static to_ddbobj(value) {
        if (value && value instanceof DdbObj)
            return value;
        const type = typeof value;
        switch (type) {
            case 'string':
                return new DdbString(value);
            case 'boolean':
                return new DdbBool(value);
            default:
                throw new Error(t('不能自动转换 {{type}} 至 DdbObj', { type }));
        }
    }
    /** 转换 js 数组为 DdbObj[] */
    static to_ddbobjs(values) {
        return values.map(value => this.to_ddbobj(value));
    }
    to_cols() {
        assert(this.form === DdbForm.table, t('form 必须是 DdbForm.table, 否则不能 to_cols'));
        return this.value.map(col => ({
            title: col.name,
            dataIndex: col.name,
            render: value => format(col.type, value, col.le, { nullstr: false, quote: false })
        }));
    }
    to_rows() {
        assert(this.form === DdbForm.table, t('form 必须是 DdbForm.table, 否则不能 to_rows'));
        let rows = new Array(this.rows);
        for (let i = 0; i < this.rows; i++) {
            let row = {};
            for (let j = 0; j < this.cols; j++) {
                const { type, name, value: values } = this.value[j]; // column
                switch (type) {
                    case DdbType.bool: {
                        const value = values[i];
                        row[name] = value === nulls.int8 ?
                            null
                            :
                                Boolean(value);
                        break;
                    }
                    case DdbType.ipaddr:
                        row[name] = values.subarray(16 * i, 16 * (i + 1));
                        break;
                    case DdbType.symbol_extended: {
                        const { base, data } = values;
                        row[name] = base[data[i]];
                        break;
                    }
                    default:
                        row[name] = values[i];
                }
            }
            rows[i] = row;
        }
        return rows;
    }
    to_dict({ strip, deep, } = {}) {
        assert(this.form === DdbForm.dict, t('this.form 必须是 DdbForm.dict, 否则不能调用 to_dict 转换为 js object'));
        const [{ value: keys, type: key_type }, { value: values, type: value_type }] = this.value;
        assert(key_type === DdbType.string && value_type === DdbType.any, t('当前只支持自动转换 dict<string, any> 为 js object'));
        assert(!(deep && !strip), t('deep = true 时必须设置 strip = true'));
        let obj = {};
        for (let i = 0; i < this.rows; i++) {
            let value = values[i];
            if (deep && value.form === DdbForm.dict)
                obj[keys[i]] = value.to_dict({ strip, deep });
            else
                obj[keys[i]] = strip ? value.value : value;
        }
        return obj;
    }
}
/** 整数一定用这个 number formatter, InspectOptions.decimals 不传也用这个  Integer must use this number formatter, InspectOptions.decimals also use this if not passed */
let default_formatter = Intl.NumberFormat('en-US', { maximumFractionDigits: 20 });
let _decimals = 20;
/** 缓存，为了优化性能，通常 options.decimals 都是不变的  Cache, in order to optimize performance, usually options.decimals are unchanged */
let _formatter = Intl.NumberFormat('en-US', { maximumFractionDigits: 20 });
/** 根据 DdbType 格式化单个元素 (value) 为字符串  Formats a single element (value) as a string according to DdbType, null returns a 'null' string */
export function format(type, value, le, options = {}) {
    const { decimals, nullstr = false, colors = false, quote = false } = options;
    const formatter = (() => {
        if (decimals === undefined || decimals === null)
            return default_formatter;
        if (decimals !== _decimals) {
            _decimals = decimals;
            _formatter = Intl.NumberFormat('en-US', { maximumFractionDigits: decimals, minimumFractionDigits: decimals });
        }
        return _formatter;
    })();
    function get_nullstr() {
        return nullstr ?
            colors ? grey('null') : 'null'
            :
                '';
    }
    switch (type) {
        case DdbType.bool:
            if (value === null || value === nulls.int8)
                return get_nullstr();
            else {
                const str = String(Boolean(value));
                return colors ? blue(str) : str;
            }
        case DdbType.char:
            if (value === null || value === nulls.int8)
                return get_nullstr();
            else {
                let str = 
                // ascii printable
                // http://facweb.cs.depaul.edu/sjost/it212/documents/ascii-pr.htm
                (32 <= value && value <= 126) ?
                    String.fromCharCode(value)
                    :
                        '\\' + String(value);
                if (quote)
                    str = str.quote();
                return colors ? green(str) : str;
            }
        case DdbType.short:
            if (value === null || value === nulls.int16)
                return get_nullstr();
            else {
                const str = default_formatter.format(value);
                return colors ? green(str) : str;
            }
        case DdbType.int:
            if (value === null || value === nulls.int32)
                return get_nullstr();
            else {
                const str = default_formatter.format(value);
                return colors ? green(str) : str;
            }
        case DdbType.long:
            if (value === null || value === nulls.int64)
                return get_nullstr();
            else {
                const str = default_formatter.format(value);
                return colors ? green(str) : str;
            }
        case DdbType.date:
            if (value === null || value === nulls.int32)
                return get_nullstr();
            else {
                const str = date2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.month:
            if (value === null || value === nulls.int32)
                return get_nullstr();
            else {
                const str = month2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.time:
            if (value === null || value === nulls.int32)
                return get_nullstr();
            else {
                const str = time2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.minute:
            if (value === null || value === nulls.int32)
                return get_nullstr();
            else {
                const str = minute2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.second:
            if (value === null || value === nulls.int32)
                return get_nullstr();
            else {
                const str = second2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.datetime:
            if (value === null || value === nulls.int32)
                return get_nullstr();
            else {
                const str = datetime2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.timestamp:
            if (value === null || value === nulls.int64)
                return get_nullstr();
            else {
                const str = timestamp2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.nanotime:
            if (value === null || value === nulls.int64)
                return get_nullstr();
            else {
                const str = nanotime2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.nanotimestamp:
            if (value === null || value === nulls.int64)
                return get_nullstr();
            else {
                const str = nanotimestamp2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.float:
            if (value === null || value === nulls.float32)
                return get_nullstr();
            else {
                const str = formatter.format(value);
                return colors ? green(str) : str;
            }
        case DdbType.double:
            if (value === null || value === nulls.double)
                return get_nullstr();
            else {
                const str = formatter.format(value);
                return colors ? green(str) : str;
            }
        case DdbType.symbol:
        case DdbType.string: {
            let str = value;
            if (quote)
                str = str.quote('single');
            return colors ? cyan(str) : str;
        }
        case DdbType.uuid: {
            const str = uuid2str(value, le);
            return colors ? cyan(str) : str;
        }
        case DdbType.functiondef: {
            const str = value.name.quote('single');
            return colors ? cyan(str) : str;
        }
        case DdbType.handle:
        case DdbType.code:
        case DdbType.datasource:
        case DdbType.resource: {
            const str = value.quote('single');
            return colors ? cyan(str) : str;
        }
        case DdbType.datehour:
            if (value === null || value === nulls.int32)
                return get_nullstr();
            else {
                const str = datehour2str(value);
                return colors ? magenta(str) : str;
            }
        case DdbType.ipaddr: {
            const str = ipaddr2str(value, le);
            return colors ? cyan(str) : str;
        }
        case DdbType.int128: {
            const str = int1282str(value, le);
            return colors ? cyan(str) : str;
        }
        case DdbType.blob: {
            const str = value.length > 100 ?
                DdbObj.dec.decode(value.subarray(0, 98)) + '…'
                :
                    DdbObj.dec.decode(value);
            return colors ? cyan(str) : str;
        }
        case DdbType.point: {
            const [x, y] = value;
            return `(${format(DdbType.double, x, le, options)}, ${format(DdbType.double, y, le, options)})`;
        }
        case DdbType.complex: {
            const [x, y] = value;
            return `${format(DdbType.double, x, le, options)}+${format(DdbType.double, y, le, options)}i`;
        }
        case DdbType.duration: {
            const { data, unit } = value;
            const str = `${data}${DdbDurationUnit[unit]}`;
            return colors ? magenta(str) : str;
        }
        case DdbType.decimal32:
        case DdbType.decimal64:
        case DdbType.decimal128: {
            const { scale, data } = value;
            if (data === null ||
                is_decimal_null_value(type, data))
                return get_nullstr();
            const s = String(data < 0 ? -data : data).padStart(scale, '0');
            const str = (data < 0 ? '-' : '') + (scale ? `${s.slice(0, -scale) || '0'}.${s.slice(-scale)}` : s);
            return colors ? green(str) : str;
        }
        default:
            return value === null ? get_nullstr() : String(value);
    }
}
/** 格式化向量、集合中的第 index 项为字符串  formatted vector, the index-th item in the collection is a string */
export function formati(obj, index, options = {}) {
    assert(index < obj.rows, 'index < obj.rows');
    if (64 <= obj.type && obj.type < 128) { // array vector
        // 因为 array vector 目前只支持：Logical, Integral（不包括 INT128, COMPRESS 类型）, Floating, Temporal
        // 都对应 TypedArray 中的一格，所以 lengths.length 等于 block 中的 row 的个数
        // av = array(INT[], 0, 5)
        // append!(av, [1..1])
        // append!(av, [1..70000])
        // append!(av, [1..1])
        // append!(av, [1..500])
        // ...
        // av
        const type_ = obj.type - 64;
        let offset = 0;
        for (const { lengths, data, rows } of obj.value) {
            let acc_len = 0;
            if (offset + rows <= index) {
                offset += rows;
                continue; // 跳过这个 block
            }
            for (const length of lengths) {
                if (offset < index) {
                    offset++;
                    acc_len += length;
                    continue;
                }
                const limit = 10;
                let items = new Array(Math.min(limit, length));
                for (let i = 0; i < items.length; i++)
                    switch (type_) {
                        case DdbType.decimal32:
                        case DdbType.decimal64:
                        case DdbType.decimal128:
                            const x = data[acc_len + i];
                            if (is_decimal_null_value(type_, x)) {
                                items[i] = '';
                                break;
                            }
                            const { scale } = obj.value;
                            const s = String(x < 0 ? -x : x).padStart(scale, '0');
                            const str = (x < 0 ? '-' : '') + (scale ? `${s.slice(0, -scale) || '0'}.${s.slice(-scale)}` : s);
                            items[i] = options.colors ? green(str) : str;
                            break;
                        case DdbType.complex:
                        case DdbType.point: {
                            const index = acc_len + i;
                            items[i] = format(type_, data.subarray(2 * index, 2 * (index + 1)), obj.le, options);
                            break;
                        }
                        case DdbType.uuid:
                        case DdbType.int128:
                        case DdbType.ipaddr: {
                            const index = acc_len + i;
                            items[i] = format(type_, data.subarray(16 * index, 16 * (index + 1)), obj.le, options);
                            break;
                        }
                        default:
                            items[i] = format(type_, data[acc_len + i], obj.le, options);
                            break;
                    }
                return (items.join(', ') + (length > limit ? ', ...' : '')).bracket('square');
            }
        }
    }
    switch (obj.type) {
        case DdbType.symbol_extended: {
            const { base, data } = obj.value;
            return format(DdbType.string, base[data[index]], obj.le, options);
        }
        case DdbType.void:
            return format(obj.type, obj.value, obj.le, options);
        case DdbType.uuid:
        case DdbType.int128:
        case DdbType.ipaddr:
            return format(obj.type, obj.value.subarray(16 * index, 16 * (index + 1)), obj.le, options);
        case DdbType.blob: {
            const value = obj.value[index];
            return value.length > 100 ?
                DdbObj.dec.decode(value.subarray(0, 98)) + '…'
                :
                    DdbObj.dec.decode(value);
        }
        case DdbType.complex:
        case DdbType.point:
            return format(obj.type, obj.value.subarray(2 * index, 2 * (index + 1)), obj.le, options);
        case DdbType.decimal32:
        case DdbType.decimal64:
        case DdbType.decimal128: {
            const { scale, data } = obj.value;
            const x = data[index];
            if (is_decimal_null_value(obj.type, x))
                return '';
            const s = String(x < 0 ? -x : x).padStart(scale, '0');
            const str = (x < 0 ? '-' : '') + (scale ? `${s.slice(0, -scale) || '0'}.${s.slice(-scale)}` : s);
            return options.colors ? green(str) : str;
        }
        default:
            return format(obj.type, obj.value[index], obj.le, options);
    }
}
export class DdbVoid extends DdbObj {
    constructor() {
        super({
            form: DdbForm.scalar,
            type: DdbType.void,
            value: null,
        });
    }
}
export class DdbBool extends DdbObj {
    constructor(value) {
        super({
            form: DdbForm.scalar,
            type: DdbType.bool,
            value,
        });
    }
}
export class DdbChar extends DdbObj {
    constructor(value) {
        super({
            form: DdbForm.scalar,
            type: DdbType.char,
            value: typeof value === 'string' ?
                value.charCodeAt(0)
                :
                    value
        });
    }
}
export class DdbInt extends DdbObj {
    constructor(value) {
        super({
            form: DdbForm.scalar,
            type: DdbType.int,
            value,
        });
    }
}
export class DdbString extends DdbObj {
    constructor(value) {
        super({
            form: DdbForm.scalar,
            type: DdbType.string,
            value
        });
    }
}
export class DdbLong extends DdbObj {
    constructor(value) {
        super({
            form: DdbForm.scalar,
            type: DdbType.long,
            value
        });
    }
}
export class DdbDouble extends DdbObj {
    constructor(value) {
        super({
            form: DdbForm.scalar,
            type: DdbType.double,
            value
        });
    }
}
export class DdbDateTime extends DdbObj {
    constructor(value) {
        super({
            form: DdbForm.scalar,
            type: DdbType.datetime,
            value
        });
    }
}
export class DdbTimeStamp extends DdbObj {
    constructor(value) {
        super({
            form: DdbForm.scalar,
            type: DdbType.timestamp,
            value
        });
    }
}
export class DdbNanoTimeStamp extends DdbObj {
    constructor(value) {
        super({
            form: DdbForm.scalar,
            type: DdbType.nanotimestamp,
            value
        });
    }
}
export class DdbBlob extends DdbObj {
    constructor(value) {
        assert(value, t('new DdbBlob 不能传空的 value'));
        super({
            form: DdbForm.scalar,
            type: DdbType.blob,
            value: value instanceof Uint8Array ? value : new Uint8Array(value)
        });
    }
}
export class DdbPair extends DdbObj {
    constructor(l, r = null) {
        super({
            form: DdbForm.pair,
            type: DdbType.int,
            rows: 2,
            cols: 1,
            value: Int32Array.of(l === null ? nulls.int32 : l, r === null ? nulls.int32 : r)
        });
    }
}
export class DdbFunction extends DdbObj {
    constructor(name, type) {
        super({
            form: DdbForm.scalar,
            type: DdbType.functiondef,
            value: { type, name }
        });
    }
}
export class DdbVectorInt extends DdbObj {
    constructor(ints, name) {
        super({
            form: DdbForm.vector,
            type: DdbType.int,
            rows: ints.length,
            cols: 1,
            value: ints instanceof Int32Array ?
                ints
                :
                    Int32Array.from(ints, v => v === null ?
                        nulls.int32
                        :
                            v),
            name,
        });
    }
}
export class DdbVectorLong extends DdbObj {
    constructor(longs, name) {
        super({
            form: DdbForm.vector,
            type: DdbType.long,
            rows: longs.length,
            cols: 1,
            value: longs instanceof BigInt64Array ?
                longs
                :
                    BigInt64Array.from(longs, v => v === null ?
                        nulls.int64
                        :
                            BigInt(v)),
            name,
        });
    }
}
export class DdbVectorDouble extends DdbObj {
    constructor(doubles, name) {
        super({
            form: DdbForm.vector,
            type: DdbType.double,
            rows: doubles.length,
            cols: 1,
            value: doubles instanceof Float64Array ?
                doubles
                :
                    Float64Array.from(doubles, v => v === null ?
                        nulls.double
                        :
                            v),
            name,
        });
    }
}
export class DdbVectorString extends DdbObj {
    constructor(strings, name) {
        super({
            form: DdbForm.vector,
            type: DdbType.string,
            rows: strings.length,
            cols: 1,
            value: strings,
            name,
        });
    }
}
export class DdbVectorAny extends DdbObj {
    constructor(objs, name) {
        super({
            form: DdbForm.vector,
            type: DdbType.any,
            rows: objs.length,
            cols: 1,
            value: DdbObj.to_ddbobjs(objs),
            name,
        });
    }
}
export class DdbVectorSymbol extends DdbObj {
    constructor(strings, name) {
        let map = new Map([
            ['', 0]
        ]);
        let data = new Uint32Array(strings.length);
        for (let i = 0; i < strings.length; i++) {
            const x = strings[i];
            let index = map.get(x);
            if (index === undefined) {
                index = map.size;
                map.set(x, index);
            }
            data[i] = index;
        }
        super({
            form: DdbForm.vector,
            type: DdbType.symbol_extended,
            rows: strings.length,
            cols: 1,
            value: {
                base: [...map.keys()],
                base_id: null,
                data,
            },
            name
        });
    }
}
export class DdbSetInt extends DdbObj {
    constructor(ints) {
        super({
            form: DdbForm.set,
            type: DdbType.int,
            rows: ints instanceof Set ?
                ints.size
                :
                    ints.length,
            cols: 1,
            value: ints instanceof Int32Array ?
                ints
                :
                    Int32Array.from(ints, v => v === null ?
                        nulls.int32
                        :
                            v),
        });
    }
}
export class DdbSetDouble extends DdbObj {
    constructor(doubles) {
        super({
            form: DdbForm.set,
            type: DdbType.double,
            rows: doubles instanceof Set ?
                doubles.size
                :
                    doubles.length,
            cols: 1,
            value: doubles instanceof Int32Array ?
                doubles
                :
                    Int32Array.from(doubles, v => v === null ?
                        nulls.int32
                        :
                            v),
        });
    }
}
export class DdbSetString extends DdbObj {
    constructor(strings) {
        if (strings instanceof Set)
            strings = [...strings];
        super({
            form: DdbForm.set,
            type: DdbType.string,
            rows: strings.length,
            cols: 1,
            value: strings
        });
    }
}
/** 构造 DdbDict 对象，支持两种用法:  Constructs a DdbDict object, which supports two usages:
     - 传入类型是 DdbVectorObj 的 keys, values 两个参数直接组成 dict<keys.type, values.type> 的 DdbDict
         The incoming type is the keys of DdbObj<DdbVectorValue>, and the two parameters of values directly form the DdbDict of dict<keys.type, values.type>
     - 传入 js object (类型是 Record<string, boolean | string | DdbObj>), 自动转换为 dict<string, any> 的 DdbDict
         Pass in js object (type is Record<string, boolean | string | DdbObj>), automatically converted to DdbDict of dict<string, any> */
export class DdbDict extends DdbObj {
    constructor(arg0, arg1) {
        if (arg1)
            super({
                form: DdbForm.dict,
                type: arg1.type,
                rows: arg0.rows,
                cols: 2,
                value: [arg0, arg1]
            });
        else {
            const keys = Object.keys(arg0);
            super({
                form: DdbForm.dict,
                type: DdbType.any,
                rows: keys.length,
                cols: 2,
                value: [
                    new DdbVectorString(keys),
                    new DdbVectorAny(Object.values(arg0))
                ]
            });
        }
    }
}
export class DdbTable extends DdbObj {
    constructor(columns, name = '') {
        super({
            form: DdbForm.table,
            type: DdbType.void,
            rows: columns[0].rows,
            cols: columns.length,
            name,
            value: columns,
        });
    }
}
export function date2ms(date) {
    // 将 server 的本地时间 (以 ms 为单位，1970.01.01 00:00:00 作为零点) 当成是客户端这里的本地的时间，然后根据本地的时区信息转换为 utc 时间
    // 本地的时区与实际的时间值相关，timezone offset 可能会受到夏令时 (DST) 的影响
    // 得到的 utc 毫秒数交给 js date 或者 dayjs 去格式化
    if (date === null || date === nulls.int32)
        return null;
    const ms = 1000 * 3600 * 24 * date;
    return 1000 * 60 * new Date(ms).getTimezoneOffset() + ms;
}
export function date2str(date, format = 'YYYY.MM.DD') {
    return (date === null || date === nulls.int32) ?
        'null'
        :
            dayjs(date2ms(date)).format(format);
}
export function month2ms(month) {
    return (month === null || month === nulls.int32) ?
        null
        :
            dayjs(month2str(month), 'YYYY.MM[M]').valueOf();
}
export function month2str(month) {
    if (month === null || month === nulls.int32)
        return 'null';
    if (month < 0)
        return String(month);
    const _month = month % 12;
    const year = Math.trunc(month / 12);
    return `${String(year).padStart(4, '0')}.${String(_month + 1).padStart(2, '0')}M`;
}
export function time2ms(time) {
    return (time === null || time === nulls.int32) ?
        null
        :
            1000 * 60 * new Date(time).getTimezoneOffset() + time;
}
export function time2str(time, format = 'HH:mm:ss.SSS') {
    return (time === null || time === nulls.int32) ?
        'null'
        :
            dayjs(time2ms(time)).format(format);
}
export function minute2ms(minute) {
    if (minute === null || minute === nulls.int32)
        return null;
    const ms = 60 * 1000 * minute;
    return 1000 * 60 * new Date(ms).getTimezoneOffset() + ms;
}
export function minute2str(minute, format = 'HH:mm[m]') {
    return (minute === null || minute === nulls.int32) ?
        'null'
        :
            dayjs(minute2ms(minute)).format(format);
}
export function second2ms(second) {
    if (second === null || second === nulls.int32)
        return null;
    const ms = 1000 * second;
    return 1000 * 60 * new Date(ms).getTimezoneOffset() + ms;
}
export function second2str(second, format = 'HH:mm:ss') {
    return (second === null || second === nulls.int32) ?
        'null'
        :
            dayjs(second2ms(second)).format(format);
}
export function datetime2ms(datetime) {
    if (datetime === null || datetime === nulls.int32)
        return null;
    const ms = 1000 * datetime;
    return 1000 * 60 * new Date(ms).getTimezoneOffset() + ms;
}
export function datetime2str(datetime, format = 'YYYY.MM.DD HH:mm:ss') {
    return (datetime === null || datetime === nulls.int32) ?
        'null'
        :
            dayjs(datetime2ms(datetime)).format(format);
}
export function timestamp2ms(timestamp) {
    if (timestamp === null || timestamp === nulls.int64)
        return null;
    const ms = Number(timestamp);
    return 1000 * 60 * new Date(ms).getTimezoneOffset() + ms;
}
/** format timestamp (bigint) to string
    - timestamp: bigint value
    - format?:
        格式串，默认是 `YYYY.MM.DD HH:mm:ss.SSS`  format string, default to `YYYY.MM.DD HH:mm:ss.SSS`
        https://day.js.org/docs/en/parse/string-format#list-of-all-available-parsing-tokens
*/
export function timestamp2str(timestamp, format = 'YYYY.MM.DD HH:mm:ss.SSS') {
    return (timestamp === null || timestamp === nulls.int64) ?
        'null'
        :
            dayjs(timestamp2ms(timestamp)).format(format);
}
export function datehour2ms(datehour) {
    if (datehour === null || datehour === nulls.int32)
        return null;
    const ms = 1000 * 3600 * datehour;
    return 1000 * 60 * new Date(ms).getTimezoneOffset() + ms;
}
export function datehour2str(datehour, format = 'YYYY.MM.DDTHH') {
    if (datehour === null || datehour === nulls.int32)
        return 'null';
    const ms = 1000 * 3600 * datehour;
    return dayjs(1000 * 60 * new Date(ms).getTimezoneOffset() + ms).format(format);
}
/** parse timestamp string to bigint value
    - str: timestamp string, 如果为空字符串或 'null' 会返回对应的空值 (nulls.int64)
        timestamp string, If it is an empty string or 'null', it will return the corresponding empty value (nulls.int64)
    - format?:
        对应传入字符串的格式串，默认是 `YYYY.MM.DD HH:mm:ss.SSS`
        The format string corresponding to the incoming string, the default is `YYYY.MM.DD HH:mm:ss.SSS`
        https://day.js.org/docs/en/parse/string-format#list-of-all-available-parsing-tokens
*/
export function str2timestamp(str, format = 'YYYY.MM.DD HH:mm:ss.SSS') {
    if (!str || str === 'null')
        return nulls.int64;
    assert(str.length === format.length, t('timestamp 字符串长度必须等于格式串长度'));
    const ms = dayjs(str, format).valueOf();
    return BigInt(-(1000 * 60 * new Date(ms).getTimezoneOffset()) +
        ms);
}
export function nanotime2ns(nanotime) {
    return nanotimestamp2ns(nanotime);
}
export function nanotime2str(nanotime, format = 'HH:mm:ss.SSSSSSSSS') {
    if (nanotime === null || nanotime === nulls.int64)
        return 'null';
    if (nanotime < 0n)
        return String(nanotime);
    const i_second_start = format.indexOf('ss');
    assert(i_second_start !== -1, t('格式串必须包含秒的格式 (ss)'));
    const i_second_end = i_second_start + 2;
    const i_nanosecond_start = format.indexOf('SSSSSSSSS', i_second_end);
    assert(i_nanosecond_start !== -1, t('格式串必须包含纳秒的格式 (SSSSSSSSS)'));
    const ms = Number(nanotime) / 1000000;
    return (dayjs(1000 * 60 * new Date(ms).getTimezoneOffset() + ms).format(format.slice(0, i_second_end)) +
        format.slice(i_second_end, i_nanosecond_start) +
        String(nanotime % 1000000000n).padStart(9, '0'));
}
export function nanotimestamp2ns(nanotimestamp) {
    if (nanotimestamp === null || nanotimestamp === nulls.int64)
        return null;
    const ms = Number(nanotimestamp / 1000000n);
    return BigInt(1000 * 60 * new Date(ms).getTimezoneOffset()) * 1000000n + nanotimestamp;
}
/** format nanotimestamp value (bigint) to string
    - nanotimestamp: bigint value
    - format?:
        格式串，默认是 `YYYY.MM.DD HH:mm:ss.SSSSSSSSS`  format string, default is `YYYY.MM.DD HH:mm:ss.SSSSSSSSS`
        秒的格式为 ss (必须包含); 纳秒的格式为 SSSSSSSSS (必须包含)  Seconds are in the format ss (must be included); nanoseconds are in the format SSSSSSSSS (must be included)
        https://day.js.org/docs/en/parse/string-format#list-of-all-available-parsing-tokens
*/
export function nanotimestamp2str(nanotimestamp, format = 'YYYY.MM.DD HH:mm:ss.SSSSSSSSS') {
    // tests:
    // nanotimestamp2str(0n)
    // nanotimestamp2str(-1n)
    // nanotimestamp2str(-9_9999_9999n)
    // nanotimestamp2str(-10_0000_0000n)
    // nanotimestamp2str(-10_0000_0001n)
    if (nanotimestamp === null || nanotimestamp === nulls.int64)
        return 'null';
    const i_second_start = format.indexOf('ss');
    assert(i_second_start !== -1, t('格式串必须包含秒的格式 (ss)'));
    const i_second_end = i_second_start + 2;
    const i_nanosecond_start = format.indexOf('SSSSSSSSS', i_second_end);
    assert(i_nanosecond_start !== -1, t('格式串必须包含纳秒的格式 (SSSSSSSSS)'));
    const remainder = nanotimestamp % 1000000000n;
    const borrow = remainder < 0n;
    const ms = Number(nanotimestamp / 1000000n);
    return (dayjs(1000 * 60 * new Date(ms).getTimezoneOffset() +
        // 去掉 9 位的纳秒部分，转化为毫秒
        Number((nanotimestamp - remainder + (borrow ? -1000000000n : 0n)) / 1000000n)).format(format.slice(0, i_second_end)) +
        format.slice(i_second_end, i_nanosecond_start) +
        String(borrow ?
            (remainder + 1000000000n) % 1000000000n
            :
                remainder).padStart(9, '0'));
}
/** parse nano timestamp string to bigint value
    - str: nano timestamp string, 如果为空字符串或 'null' 会返回对应的空值 (nulls.int64)
        nano timestamp string, If it is an empty string or 'null', it will return the corresponding empty value (nulls.int64)
    - format?:
        对应传入字符串的格式串，默认是 `YYYY.MM.DD HH:mm:ss.SSSSSSSSS`
        秒的格式为 ss (必须包含); 纳秒的格式为 SSSSSSSSS (必须包含)
        https://day.js.org/docs/en/parse/string-format#list-of-all-available-parsing-tokens
        The format string corresponding to the incoming string, the default is `YYYY.MM.DD HH:mm:ss.SSSSSSSSS`
        Seconds are in the format ss (must be included); nanoseconds are in the format SSSSSSSSS (must be included) */
export function str2nanotimestamp(str, format = 'YYYY.MM.DD HH:mm:ss.SSSSSSSSS') {
    if (!str || str === 'null')
        return nulls.int64;
    assert(str.length === format.length, t('nanotimestamp 字符串长度必须等于格式串长度'));
    const i_second_start = format.indexOf('ss');
    assert(i_second_start !== -1, t('格式串必须包含秒的格式 (ss)'));
    const i_second_end = i_second_start + 2;
    const i_nanosecond_start = format.indexOf('SSSSSSSSS', i_second_end);
    assert(i_nanosecond_start !== -1, t('格式串必须包含纳秒的格式 (SSSSSSSSS)'));
    const ms = dayjs(str.slice(0, i_second_end), format.slice(0, i_second_end)).valueOf();
    return (BigInt(-(1000 * 60 * new Date(ms).getTimezoneOffset()) +
        ms) * 1000000n
        +
            BigInt(str.slice(i_nanosecond_start, i_nanosecond_start + 9)));
}
export function ipaddr2str(buffer, le = true, ipv6) {
    let buf = buffer;
    if (le)
        buf = buffer.slice().reverse();
    const i_non_zero = buf.findIndex(x => x);
    if (ipv6 ||
        i_non_zero !== -1 && i_non_zero < 12) // ipv6
        return buf2ipaddr([...buf]).toString();
    else // ipv4
        return buf.subarray(12).join('.');
}
export function uuid2str(buffer, le = true) {
    const str = int1282str(buffer, le);
    return `${str.slice(0, 8)}-${str.slice(8, 12)}-${str.slice(12, 16)}-${str.slice(16, 20)}-${str.slice(20)}`;
}
export function int1282str(buffer, le = true) {
    let buf = buffer;
    if (le)
        buf = buffer.slice().reverse();
    return [...buf].map(x => x.toString(16)
        .padStart(2, '0')).join('');
}
export const winsize = 10_0000;
export class DdbConnectionError extends Error {
    name = 'DdbConnectionError';
    cause;
    ddb;
    constructor(ddb, error) {
        super(error?.message || `${ddb.url} ${t('连接出错了，可能由于网络原因连接已被关闭，或服务器断开连接')}`, { cause: error });
        if (error)
            this.cause = error;
        this.ddb = ddb;
    }
}
export class DdbDatabaseError extends Error {
    name = 'DdbDatabaseError';
    ddb;
    type;
    options;
    constructor(message, ddb, type, options) {
        super(message);
        this.ddb = ddb;
        this.type = type;
        this.options = options;
    }
}
export class DDB {
    /** 当前的 session id (http 或 tcp) */
    sid = '0';
    /** utf-8 text decoder */
    dec = new TextDecoder('utf-8');
    enc = new TextEncoder();
    /** DolphinDB WebSocket URL
        e.g. `ws://127.0.0.1:8848/`, `wss://dolphindb.com`
    */
    url;
    /** 为所有 websocket 操作加锁，包括设置 this.on_message, this.on_error, websocket.send */
    lwebsocket = new Lock();
    /** little endian (server) */
    le = true;
    /** little endian (client) */
    static le_client = Boolean(new Uint8Array(Uint32Array.of(1).buffer)[0]);
    /** 是否在建立连接后自动登录，默认 true  Whether to automatically log in after the connection is established, the default is true */
    autologin = true;
    /** DolphinDB 登录用户名  DolphinDB username */
    username = 'admin';
    /** DolphinDB 登录密码  DolphinDB password */
    password = '123456';
    /** python session flag (2048) */
    python = false;
    /** 是否为流数据连接，非流数据这个字段恒为 null  Whether it is a streaming data connection, this field is always null for non-streaming data */
    streaming = null;
    /** 是否打印每个 rpc 的信息用于调试 */
    verbose = false;
    // --- 内部选项, 状态
    print_message_buffer = false;
    print_object_buffer = false;
    print_message = true;
    parse_object = true;
    pnode_run_defined = false;
    /** 在 websocket 收到的第一个 error 时，
        在 connect_websocket 的 on_error 回调中构造 DdbConnectionError 并保存到 DDB 对象上，
        这个 error 的错误信息最准确
        this.errored 为 true 时建议 throw this.error || new DdbConnectionError(this) */
    error;
    /** DdbMessage listeners */
    listeners = [];
    /** 首次 connect 连接的 promise, 后面的 connect 调用都返回这个 */
    pconnect;
    /** 首次定义 pnode_run 的 promise，保证并发调用 rpc 时只定义一次 pnode_run */
    ppnode_run;
    get connected() {
        return !this.error && this.lwebsocket.resource?.readyState === WebSocket.OPEN;
    }
    get errored() {
        return Boolean(this.error || (this.lwebsocket.resource && this.lwebsocket.resource.readyState !== WebSocket.OPEN));
    }
    /**
        使用 WebSocket URL 初始化连接到 DolphinDB 的实例（不建立实际的网络连接）
        Initialize an instance of DolphinDB Client using the WebSocket URL (without establishing an actual network connection)
        - url?: DolphinDB WebSocket URL，如：`ws://127.0.0.1:8848`, 默认为当前页面 URL
            DolphinDB WebSocket URL. e.g.：`ws://127.0.0.1:8848`, Defaults to the current page URL
        - options?: DdbOptions
            - autologin?: 是否在建立连接后自动登录，默认 `true`  Whether to log in automatically after establishing a connection, default `true`
            - username?: DolphinDB 登录用户名，默认 `'admin'`  DolphinDB username, default `'admin'`
            - password?: DolphinDB 登录密码，默认 `'123456'`  DolphinDB password, default `'123456'`
            - python?: 设置 python session flag，默认 `false`  set python session flag, default `false`
            - streaming?: 设置该选项后，该 WebSocket 连接只用于流数据  When this option is set, the WebSocket connection is only used for streaming data
            - verbose?: 是否打印每个 rpc 的信息用于调试
        
        @example
        let ddb = new DDB('ws://127.0.0.1:8848')
        
        // 使用 HTTPS 加密  Encrypt with HTTPS
        let ddbsecure = new DDB('wss://dolphindb.com', {
            autologin: true,
            username: 'admin',
            password: '123456',
            python: false
        }) */
    constructor(url, options = {}) {
        this.url = url;
        if (options.verbose !== undefined)
            this.verbose = options.verbose;
        if (options.autologin !== undefined)
            this.autologin = options.autologin;
        if (options.username !== undefined)
            this.username = options.username;
        if (options.password !== undefined)
            this.password = options.password;
        if (options.python !== undefined)
            this.python = options.python;
        if (options.streaming !== undefined)
            this.streaming = options.streaming;
    }
    on_message(buffer, websocket) {
        throw new Error(t('这是在调用 this.rpc 之前默认的 on_message, 不应该被调用到，除非建立连接后 server 先推送了 message'));
    }
    on_error() {
        // 这里的实现一定会被 connect, rpc 中的实现覆盖
    }
    /** 调用后会确保和数据库的连接是正常的 (this.connected === true && this.errored === false)，否则抛出错误
        这个方法是幂等的，首次调用建立实际的 WebSocket 连接到 URL 对应的 DolphinDB，然后执行自动登录，
        如果是流数据连接，还会调用 publishTable 订阅流表
        后续调用检查上面的条件
        连接断开后禁止再次调用 connect 重连原有 ddb 对象，应该通过 new DDB() 的方式新建连接对象，原因是：
        1. on_error 回调和某个 websocket 绑定了，不方便解绑后重新绑定
        2. session 是有状态的，重连也无法恢复之前的状态
        3. 断线后所有之前的 ddb.call, ddb.eval 都应该抛出连接错误
        
        After calling, it will ensure that the connection with the database is normal (this.connected === true && this.errored === false), otherwise an error will be thrown
        This method is idempotent, the first call establishes an actual WebSocket connection to the DolphinDB corresponding to the URL, and subsequent calls check the above conditions
        After the connection is disconnected, it is forbidden to call connect again to reconnect the original ddb object. You should use new DDB() to create a new connection object because:
        1. The on_error callback is bound to a certain websocket, and it is inconvenient to unbind and rebind
        2. The session is stateful, and the previous state cannot be restored even after reconnection
        3. After disconnection, all previous ddb.call, ddb.eval should throw a connection error */
    async connect() {
        if (this.errored)
            throw this.error || new DdbConnectionError(this);
        if (this.connected)
            return;
        return this.pconnect ??= new Promise(async (resolve, reject) => {
            this.on_error = () => {
                reject(this.error /* 一定有，不需要再 || new DdbConnectionError(this)  */);
            };
            try {
                // 连接建立之前应该不会有别的调用占用 this.lwebsocket
                this.lwebsocket.resource = await connect_websocket(this.url, {
                    protocols: (() => {
                        if (this.streaming)
                            return 'streaming';
                        if (this.python)
                            return 'python';
                    })(),
                    on_message: (buffer, websocket) => {
                        this.on_message(buffer, websocket);
                    },
                    on_error: error => {
                        this.error ??= new DdbConnectionError(this, error);
                        this.on_error();
                    }
                });
            }
            catch (error) {
                this.error ??= new DdbConnectionError(this, error);
                reject(this.error);
                return;
            }
            try {
                assert(this.connected);
                await this.rpc('connect', { skip_connection_check: true });
                if (this.streaming)
                    await this.subscribe();
                resolve();
            }
            catch (error) {
                this.error ??= error;
                reject(error);
            }
        });
    }
    get_rpc_options({ urgent = false, secondary = false, async: _async = false, pickle = false, clear = false, api = false, compress = false, cancellable = true, priority = urgent ? 8 : 4, parallelism = 8, root_id = '', limit, } = {}) {
        let flag = 0;
        if (urgent)
            flag += 1;
        if (secondary)
            flag += 2;
        if (_async)
            flag += 4;
        if (pickle)
            flag += 8;
        if (clear)
            flag += 16;
        if (api)
            flag += 32;
        if (compress)
            flag += 64;
        // python session
        if (this.python)
            flag += 2048;
        const options = [
            flag,
            cancellable ? 1 : 0,
            priority,
            parallelism,
            ...limit ? [
                root_id,
                limit,
            ] : [],
        ];
        return options.join('_');
    }
    disconnect() {
        const { resource } = this.lwebsocket;
        if (resource) {
            const { readyState } = resource;
            if (readyState !== WebSocket.CLOSED && readyState !== WebSocket.CLOSING)
                // 这里不获取 lock，直接关闭连接
                resource.close(1000);
        }
    }
    /** (内部使用的方法) rpc through websocket (function/script/variable command)
        未连接到 DDB 时调用会自动连接，连接断开时调用会抛出 DdbConnectionError
        When the DDB is not connected, the call will be automatically connected. When the connection is disconnected, the call will throw the DdbConnectionError
        - type: API 类型: 'script' | 'function' | 'variable' | 'connect'
        - options:
            - urgent?: 决定 `行为标识` 那一行字符串的取值（只适用于 script 和 function）
            - vars?: type === 'variable' 时必传，variable 指令中待上传的变量名
            - listener?: 处理本次 rpc 期间的消息 (DdbMessage)
            - parse_object?: 在本次 rpc 期间设置 parse_object, 结束后恢复原有
                为 false 时返回的 DdbObj 仅含有 buffer 和 le，不做解析，以便后续转发、序列化
            - skip_connection_check?: 在首次 await ddb.connect() 建立连接时不能再次调用 await this.connect() 确保连接状态，会导致循环依赖，
                将这个 flag 设为 true 跳过连接状态检查 */
    async rpc(type, options) {
        // 保留调用栈信息
        let error = new DdbDatabaseError('', this, type, options);
        if (!options.skip_connection_check)
            await this.connect();
        const { script, func, args: _args = [], vars = [], urgent, listener, } = options;
        if (func === 'pnode_run')
            await (this.ppnode_run ??= this.eval(this.python ?
                '\n' +
                    'def pnode_run (nodes, func_name, args, add_node_alias):\n' +
                    '    nargs = size(args)\n' +
                    '    func = funcByName(func_name)\n' +
                    '    \n' +
                    '    if not nargs:\n' +
                    '        return pnodeRun(func, nodes, add_node_alias)\n' +
                    '    \n' +
                    '    args_partial = [ ]\n' +
                    '    args_partial.append(func)\n' +
                    '    for a in args:\n' +
                    '        args_partial.append(a)\n' +
                    '    \n' +
                    '    return pnodeRun(\n' +
                    '        unifiedCall(partial, args_partial),\n' +
                    '        nodes,\n' +
                    '        add_node_alias\n' +
                    '    )\n'
                :
                    // 这个开头的空行很重要，应该可以绕过 webLoginRequired = true 时禁止执行代码
                    // 搜一下 APISocketConsole::execute
                    // https://dolphindb1.atlassian.net/browse/D20-4991
                    '\n' +
                        'def pnode_run (nodes, func_name, args, add_node_alias = true) {\n' +
                        '    nargs = size(args)\n' +
                        '    func = funcByName(func_name)\n' +
                        '    \n' +
                        '    if (!nargs)\n' +
                        '        return pnodeRun(func, nodes, add_node_alias)\n' +
                        '    \n' +
                        '    args_partial = array(any, 1 + nargs, 1 + nargs)\n' +
                        '    args_partial[0] = func\n' +
                        '    args_partial[1:] = args\n' +
                        '    return pnodeRun(\n' +
                        '        unifiedCall(partial, args_partial),\n' +
                        '        nodes,\n' +
                        '        add_node_alias\n' +
                        '    )\n' +
                        '}\n', { urgent: true }));
        // this 上的当前配置需要在 message 到达后使用，先保存起来
        const _listeners = [...this.listeners].reverse();
        // rpc 请求期间需要独占 websocket，所以设计了一个锁，申请之后才能使用
        // ddb 世界观：需要等待上一个 rpc 结果从 server 返回之后才能发起下一个调用  
        // 违反世界观可能造成:  
        // 1. 并发多个请求只返回第一个结果（阻塞，需后续请求疏通）
        // 2. windows 下 ddb server 返回多个相同的结果
        // 既然上一个请求没有出现 websocket error，且函数开头已经调用了 await this.connect() 检查过，
        // 这里也乐观的认为 this.connected && !this.errored 为 true
        return this.lwebsocket.request(async (websocket) => {
            // 独占资源后先检查状态
            if (this.errored || !this.connected)
                throw this.error || new DdbConnectionError(this);
            const args = DdbObj.to_ddbobjs(_args);
            const id = genid() % 1000;
            const rpc_id = ` (id = ${id})`;
            const command = this.enc.encode((() => {
                switch (type) {
                    case 'function':
                        if (this.verbose)
                            console.log(func + args.map(arg => arg.toString()).join(', ').bracket() + rpc_id);
                        return 'function\n' +
                            `${func}\n` +
                            `${args.length}\n` +
                            `${Number(DDB.le_client)}\n`;
                    case 'script':
                        if (this.verbose)
                            console.log(script + rpc_id);
                        return 'script\n' +
                            script;
                    case 'variable':
                        if (this.verbose)
                            for (let i = 0; i < vars.length; i++)
                                console.log(`${vars[i]} = ${args[i].toString()}`);
                        return 'variable\n' +
                            `${vars.join(',')}\n` +
                            `${vars.length}\n` +
                            `${Number(DDB.le_client)}\n`;
                    case 'connect':
                        if (this.verbose)
                            console.log('connect()' +
                                (this.autologin ?
                                    '\n' +
                                        `login(${this.username.quote()}, ${this.password.quote()})`
                                    :
                                        '') +
                                rpc_id);
                        return 'connect\n' +
                            // 详见 InterProcessIO.cpp#APISocketConsole::parseScript 中的
                            // Util::startWith "connect"
                            (this.autologin ?
                                'login\n' +
                                    this.username + '\n' +
                                    this.password /* encrypted (可选参数) + '\n' + 'false' */
                                :
                                    '');
                }
            })());
            // 使用资源发送请求并等待请求完成
            const result = await new Promise((resolve, reject) => {
                this.on_error = () => {
                    // 这里一定有了 this.error, 不需要再 || new DdbConnectionError(this)
                    reject(this.error);
                };
                this.on_message = buffer => {
                    try {
                        const buf = new Uint8Array(buffer);
                        if (this.print_message_buffer)
                            console.log(buf);
                        const message = this.parse_message(buf, error);
                        listener?.(message, this);
                        for (const listener of _listeners)
                            listener(message, this);
                        const { type, data } = message;
                        switch (type) {
                            case 'print':
                                if (this.print_message)
                                    console.log(data);
                                break;
                            case 'object':
                                resolve(data);
                                break;
                            case 'error':
                                reject(data);
                                break;
                        }
                    }
                    catch (error) {
                        // 这里的错误并非 websocket 错误，而是 rpc 错误
                        reject(error);
                    }
                };
                websocket.send(concat([
                    this.enc.encode(`API2 ${this.sid} ${command.length} / ${this.get_rpc_options({ urgent })}\n`),
                    command,
                    ...args.map(arg => arg.pack())
                ]));
            });
            if (this.verbose)
                console.log(result.toString() + rpc_id);
            return result;
        });
    }
    /** eval script through websocket (script command)
        - script?: 执行的脚本  Script to execute
        - options?: 执行选项  execution options
            - urgent?: 紧急 flag，确保提交的脚本使用 urgent worker 处理，防止被其它作业阻塞
                Urgent flag to ensure that submitted scripts are processed by urgent workers to prevent being blocked by other jobs
            - listener?: 处理本次 rpc 期间的消息 (DdbMessage)  Process messages during this rpc (DdbMessage)
            - parse_object?: 在该次 rpc 期间设置 parse_object, 结束后恢复原有，为 false 时返回的 DdbObj 仅含有 buffer 和 le，
                不做解析，以便后续转发、序列化
                Set parse_object during this rpc, and restore the original after the end.
                When it is false, the returned DdbObj only contains buffer and le without parsing,
                so as to facilitate subsequent forwarding and serialization
    */
    async eval(script, { urgent, listener, parse_object, } = {}) {
        return this.rpc('script', { script, urgent, listener, parse_object });
    }
    /** call function through websocket (function command)
        - func: 函数名  function name
        - args?: `[ ]` 调用参数 (传入的原生 string 和 boolean 会被自动转换为 DdbObj<string> 和 DdbObj<boolean>)
            Call parameters (the incoming native string and boolean will be automatically converted to DdbObj<string> and DdbObj<boolean>)
        - options?: 调用选项  call options
            - urgent?: 紧急 flag。使用 urgent worker 执行，防止被其它作业阻塞
                Emergency flag. Use urgent worker execution to prevent being blocked by other jobs
            - node?: 设置结点 alias 时发送到集群中对应的结点执行 (使用 DolphinDB 中的 rpc 方法)
                When the node alias is set, it is sent to the corresponding node in the cluster for execution (using the rpc method in DolphinDB)
            - nodes?: 设置多个结点 alias 时发送到集群中对应的多个结点执行 (使用 DolphinDB 中的 pnodeRun 方法)
                When setting multiple node aliases, send them to the corresponding multiple nodes in the cluster for execution (using the pnodeRun method in DolphinDB)
            - func_type?: 设置 node 参数时必传，需指定函数类型，其它情况下不传
                It must be passed when setting the node parameter, the function type needs to be specified, and it is not passed in other cases
            - add_node_alias?: 设置 nodes 参数时选传，其它情况不传
                Select to pass when setting the nodes parameter, otherwise not pass
            - listener?: 处理本次 rpc 期间的消息 (DdbMessage)
                Process messages during this rpc (DdbMessage)
            - parse_object?: 在该次 rpc 期间设置 parse_object, 结束后恢复原有，为 false 时返回的 DdbObj 仅含有 buffer 和 le，
                不做解析，以便后续转发、序列化
                Set parse_object during this rpc, and restore the original after the end.
                When it is false, the returned DdbObj only contains buffer and le without parsing,
                so as to facilitate subsequent forwarding and serialization
            - skip_connection_check?: (内部使用) 在首次 await ddb.connect() 建立连接时不能再次调用 await this.connect() 确保连接状态，会导致循环依赖，
                将这个 flag 设为 true 跳过连接状态检查
                (internal use) When await ddb.connect() establishes a connection for the first time, you cannot call await this.connect() again to ensure the connection status, which will lead to circular dependencies.
                 Set this flag to true to skip connection status checks */
    async call(func, args = [], { urgent, node, nodes, func_type, add_node_alias, listener, parse_object, skip_connection_check } = {}) {
        if (node) {
            assert(func_type in DdbFunctionType, t('指定 node 时必须设置 func_type'));
            args = [
                node,
                new DdbFunction(func, func_type),
                ...args
            ];
            func = 'rpc';
        }
        if (nodes) {
            args = [
                new DdbVectorString(nodes),
                func,
                new DdbVectorAny(args),
                ...(() => {
                    if (typeof add_node_alias !== 'undefined')
                        return [add_node_alias];
                    if (this.python)
                        return [true];
                    return [];
                })()
            ];
            func = 'pnode_run';
        }
        return this.rpc('function', {
            func,
            args,
            urgent,
            listener,
            parse_object,
            skip_connection_check
        });
    }
    /** upload variable through websocket (variable command) */
    async upload(
    /** 上传的变量名  Uploaded variables' name */
    vars, 
    /** 上传的变量值  Uploaded variables' value */
    args, { listener, parse_object, } = {}) {
        assert(args.length && args.length === vars.length, t('variable 指令参数不能为空且参数名不能为空，且数量应该匹配'));
        return this.rpc('variable', { vars, args, listener, parse_object });
    }
    /** 取消当前 session id 对应的所有 console jobs  Cancel all console jobs corresponding to the current session id */
    async cancel() {
        let ddb = new DDB(this.url, this);
        try {
            // 因为是新建的连接，而且执行完脚本之后马上就关闭了，所以不用考虑变量泄漏的问题
            await ddb.eval(`jobs = exec rootJobId from getConsoleJobs() where sessionId = ${this.sid}\n` +
                (this.python ? 'if size(jobs):\n' : 'if (size(jobs))\n') +
                '    cancelConsoleJob(jobs)\n', { urgent: true });
        }
        finally {
            ddb.disconnect();
        }
    }
    /** 解析服务端响应报文，返回去掉 header 的 data buf */
    parse_message(buf, error) {
        // MSG\n
        // <message>\0
        // 'M'.codePointAt(0).to_hex_str()
        if (buf[0] === 0x4d && buf[1] === 0x53 && buf[2] === 0x47 && buf[3] === 0x0a) {
            assert(buf.at(-1) === 0, t('print 消息的 buffer 应该以 \\0 结束'));
            if (buf.indexOf(0) !== buf.length - 1)
                console.warn(t('print 消息的 buffer 中间不应该有 \\0'));
            return {
                type: 'print',
                data: this.dec.decode(buf.subarray(4, -1))
            };
        }
        // '1166953221 1 1\n'
        // 'OK\n'
        // '\x04\x00\x02\x00\x00\x00'
        /** index of line feed 0 */
        const ilf0 = buf.indexOf(0x0a); // '\n'
        const parts = this.dec.decode(buf.subarray(0, ilf0)).split(' ');
        /** session id */
        const sid = parts[0];
        if (sid !== this.sid) {
            console.log(t('session id 从 {{sid}} 变为 {{sid_}}', { sid: this.sid, sid_: sid }));
            this.sid = sid;
        }
        /** 返回对象的数量 */
        const nobj = Number(parts[1]);
        /** 大小端: 协议中大端为 0, 小端为 1 */
        this.le = Number(parts[2]) !== 0;
        const ils1 = ilf0 + 1;
        const ilf1 = buf.indexOf(0x0a, ils1); // '\n'
        /** 'OK' 表示成功，其它文本表示失败 */
        const message = this.dec.decode(buf.subarray(ils1, ilf1));
        if (message !== 'OK') {
            error.message = message;
            return { type: 'error', data: error };
        }
        const bufobj = buf.subarray(ilf1 + 1);
        if (this.print_object_buffer)
            console.log(bufobj);
        return {
            type: 'object',
            data: error.options.parse_object ?? this.parse_object ?
                DdbObj.parse(bufobj, this.le)
                :
                    new DdbObj({
                        form: DdbForm.scalar,
                        type: DdbType.void,
                        length: 0,
                        le: this.le,
                        buffer: bufobj,
                    })
        };
    }
    /** 内部的流订阅方法  Internal stream subscription method */
    async subscribe() {
        console.log(t('订阅流表成功') + ', colnames:', 
        // string[3](['time', 'stock', 'price'])
        this.streaming.colnames = (await this.call('publishTable', [
            'localhost',
            new DdbInt(0),
            this.streaming.table,
            (this.streaming.action ||= `api_js_${new Date().getTime()}`),
            // new DdbInt(-1),  // offset
            // filter
            // allow exists
        ], { skip_connection_check: true })).value);
        this.streaming.window = {
            offset: 0,
            rows: 0,
            segments: [],
        };
        // 先准备好收到 websocket message 的 callback
        this.on_message = buffer => {
            // 复制一份，避免 handler 中异步直接使用传入的 streaming 这个对象时属性发生变化
            let streaming = { ...this.streaming, error: null };
            try {
                const dv = new DataView(buffer);
                const buf = new Uint8Array(buffer);
                streaming.time = dv.getBigInt64(1, this.le);
                streaming.id = dv.getBigInt64(9, this.le);
                const i_topic_end = buf.indexOf(0, 17);
                streaming.topic = this.dec.decode(buf.subarray(17, i_topic_end));
                // 是 column 片段组成的 any vector
                const data = streaming.data = DdbObj.parse(buf.subarray(i_topic_end + 1), this.le);
                let { window: win } = streaming;
                win.rows += streaming.rows = data.value[0].rows;
                win.segments.push(data);
                if (win.rows >= winsize * 2 && win.segments.length >= 2) {
                    let winsize_ = 0;
                    let i = win.segments.length - 1;
                    // 往前移动至首个累计 winsize_ 超过 winsize 的位置
                    for (; winsize_ < winsize; i--)
                        winsize_ += win.segments[i].value[0].rows;
                    win.segments = win.segments.slice(i);
                    win.offset += win.rows - winsize_;
                    win.rows = winsize_;
                }
            }
            catch (error) {
                // 将 error 交给 handler 处理
                streaming.error = error;
            }
            this.streaming.handler(streaming);
        };
        return this.streaming;
    }
}
//# sourceMappingURL=browser.js.map