/// <reference types="node" resolution-mode="require"/>
/// <reference types="node" resolution-mode="require"/>
/// <reference types="node" resolution-mode="require"/>
import { promises as fsp, default as fs } from 'fs';
type FileHandle = fsp.FileHandle & {
    fp: string;
};
export type Encoding = 'utf-8' | 'gb18030' | 'shift-jis' | 'utf-16le';
export declare const encodings: readonly ["utf-8", "gb18030", "shift-jis", "utf-16le"];
/** fp 所指向的 文件/ 文件夹 是否存在
    Does the file/folder pointed to by fp exist?  */
export declare function fexists(fp: string, { print }?: {
    print?: boolean;
}): boolean;
/** 打开文件，返回 FileHandle
    open file, return FileHandle
    Some characters (`< > : " / \ | ? *`) are reserved under Windows as documented
    by [Naming Files, Paths, and Namespaces](https://docs.microsoft.com/en-us/windows/desktop/FileIO/naming-a-file). Under NTFS, if the filename contains
    a colon, Node.js will open a file system stream, as described by [this MSDN page](https://docs.microsoft.com/en-us/windows/desktop/FileIO/using-streams).
    
    - flags: `'r'`
    - options?:
        - mode?: `'0o666'` Sets the file mode (permission and sticky bits) if the file is created. */
export declare function fopen(fp: string, flags: string | number, { mode, print }?: {
    mode?: fs.Mode;
    print?: boolean;
}): Promise<fsp.FileHandle & {
    fp: string;
    flags: string | number;
    mode: fs.Mode;
}>;
export declare function fread(fp: string): Promise<string>;
export declare function fread(fp: string, { dir, encoding, print }?: {
    dir?: string;
    encoding: 'binary';
    print?: boolean;
}): Promise<Buffer>;
export declare function fread(fp: string, { dir, encoding, print }?: {
    dir?: string;
    encoding?: Encoding;
    print?: boolean;
}): Promise<string>;
export declare function fread_lines(fp: string, options?: {
    dir?: string;
    encoding?: Encoding;
    print?: boolean;
}): Promise<string[]>;
export declare function fread_json<T = any>(fp: string, options?: {
    dir?: string;
    encoding?: Encoding;
    print?: boolean;
}): Promise<T>;
/** 写入 data 到 fp 路径所指的文件
    - fp: 目标文件完整路径
    - data: 支持下面几种类型
        - string: 写入文本
        - Uint8Array, Buffer: 写入二进制 buffer
        - any: 通过 JSON.stringify 转为文本后写入文件 */
export declare function fwrite(fp: string | FileHandle, data: string | Uint8Array | any, { dir, print, mkdir, }?: {
    dir?: string;
    print?: boolean;
    mkdir?: boolean;
}): Promise<void>;
export declare function fappend(fp: string, data: string | Uint8Array, { dir, print }?: {
    dir?: string;
    print?: boolean;
}): Promise<void>;
export type FStats = fs.BigIntStats & {
    fp: string;
};
export interface FListOptions {
    filter?: RegExp | ((fp: string) => any);
    deep?: boolean;
    absolute?: boolean;
    print?: boolean;
    stats?: boolean;
}
/**
    - fpd: 文件夹完整路径 absolute path of directory
    - options?:
        - deep?: `false` 递归遍历 recursively
        - absolute?: `false` 返回、打印完整路径而不是相对路径 Return, print full path instead of relative path
        - print?: `true`
        - filter?: `true`  RegExp | (fp: string) => any 注意当 deep = true 时被 filter 过滤掉的目录及目录中的文件不会包含在结果中
                    Note that when deep = true, directories and files in directories that are filtered out by the filter will not be included in the results
        - stats?: `false` 启用后返回 FStats 列表, 不能和 deep 一起使用
                   Returns the FStats list when enabled, cannot be used with deep */
export declare function flist(fpd: string): Promise<string[]>;
export declare function flist(fpd: string, options?: FListOptions & {
    stats: true;
}): Promise<FStats[]>;
export declare function flist(fpd: string, options?: FListOptions): Promise<string[]>;
export declare function fstat(fp: string): Promise<FStats>;
/** 删除文件或文件夹  delete files or folders
    - fp: 文件或文件夹的完整路径  The full path to the file or folder
    - options?:
        - print?: `true`
    
    返回是否实际进行了删除操作
    Returns whether the delete operation actually took place */
export declare function fdelete(fp: string, { print }?: {
    print?: boolean;
}): Promise<boolean>;
/** 复制文件或文件夹  copy file or direcotry
    - fp_src: 源 文件/文件夹 完整路径  src file/directory absolute path
    - fp_dst: 目标 文件/文件夹 完整路径  dst file/directory absolute path
    - options?:
        - print?: `true`
        - overwrite?: `true`
    
    @example
    fcopy('d:/temp/camera/', 'd:/camera/') */
export declare function fcopy(fp_src: string, fp_dst: string, { print, overwrite, }?: {
    print?: boolean;
    overwrite?: boolean;
}): Promise<void>;
/** 移动文件或文件夹  move file or direcotry
    - src: 源 文件/文件夹 完整路径  src file/directory absolute path
    - dst: 目标 文件/文件夹 完整路径  dst file/directory absolute path
    @example
    fmove('d:/temp/camera/', 'd:/camera/') */
export declare function fmove(src: string, dst: string, { overwrite, print }?: {
    overwrite?: boolean;
    print?: boolean;
}): Promise<void>;
/** 重命名文件  rename file
    - fp:  当前文件名／路径  current filename/path
    - fp_: 新的文件名／路径  new filename/path
    - options?:
        - fpd?: fp 和 fp_ 在同一文件夹内  fp and fp_ is in same directory
        - print?: `true`
        - overwrite?: `true`  默认覆盖（不检查效率更高）  better performance without check */
export declare function frename(fp: string, fp_: string, { fpd, print, overwrite }?: {
    fpd?: string;
    print?: boolean;
    overwrite?: boolean;
}): Promise<void>;
/**
    递归创建文件夹，确保 fpd 指向的文件夹存在  Create folders recursively, make sure the folder pointed to by fpd exists
    返回首个创建的文件夹或 undefined  Returns the first created folder or undefined
    
    - fpd: 文件夹完整路径  Folder full path
    - options?:
        - print?: `true`
        - mode?: `'0o777'` */
export declare function fmkdir(fpd: string, { print, mode, }?: {
    print?: boolean;
    /** `0o777` A file mode. If a string is passed, it is parsed as an octal integer. */
    mode?: string | number;
}): Promise<string>;
/** 创建软链接  Create soft links
    - fp_real: 现在真实文件/文件夹的路径  current real file/directory path
    - fp_link: 目标链接文件/文件夹的路径  target file/directory path */
export declare function flink(fp_real: string, fp_link: string, { junction, print }?: {
    junction?: boolean;
    print?: boolean;
}): Promise<void>;
/** 打开一个文件并搜索替换某个 pattern  open a file and replace certain pattern */
export declare function freplace(fp: string, pattern: string | RegExp, replacement: string): Promise<void>;
export {};
