import { promises as fsp, default as fs, } from 'fs';
import { isUint8Array } from 'util/types';
import path from 'upath';
import fse from 'fs-extra';
import { t } from './i18n/instance.js';
import { to_json } from './prototype.js';
import { assert } from './utils.js';
export const encodings = ['utf-8', 'gb18030', 'shift-jis', 'utf-16le'];
/** fp 所指向的 文件/ 文件夹 是否存在
    Does the file/folder pointed to by fp exist?  */
export function fexists(fp, { print = true } = {}) {
    const exists = fs.existsSync(fp);
    if (print)
        console.log(exists ? t('已存在') : t('不存在'), fp);
    return exists;
}
/** 打开文件，返回 FileHandle
    open file, return FileHandle
    Some characters (`< > : " / \ | ? *`) are reserved under Windows as documented
    by [Naming Files, Paths, and Namespaces](https://docs.microsoft.com/en-us/windows/desktop/FileIO/naming-a-file). Under NTFS, if the filename contains
    a colon, Node.js will open a file system stream, as described by [this MSDN page](https://docs.microsoft.com/en-us/windows/desktop/FileIO/using-streams).
    
    - flags: `'r'`
    - options?:
        - mode?: `'0o666'` Sets the file mode (permission and sticky bits) if the file is created. */
export async function fopen(fp, flags, { mode, print } = {}) {
    if (print)
        console.log(t('打开文件'), fp);
    return Object.assign(await fsp.open(fp, flags, mode), { fp, flags, mode });
}
export async function fread(fp, { dir, encoding = 'utf-8', print = true } = {}) {
    if (dir) {
        assert(dir.endsWith('/'), t('dir 必须以 / 结尾'));
        fp = dir + fp;
    }
    assert(!fp.endsWith('/'), t('fp 必须是文件，不能以 / 结尾'));
    assert(path.isAbsolute(fp), `${t('fp 必须是绝对路径:')} ${fp}`);
    assert(encoding !== 'auto');
    if (print)
        console.log(t('读取'), fp);
    switch (encoding) {
        case 'utf-8':
            return fsp.readFile(fp, { encoding: 'utf-8' });
        case 'binary':
            return fsp.readFile(fp);
        default:
            return new TextDecoder(encoding)
                .decode(await fsp.readFile(fp));
    }
}
export async function fread_lines(fp, options = {}) {
    return (await fread(fp, options))
        .split_lines();
}
export async function fread_json(fp, options = {}) {
    return JSON.parse(await fread(fp, options));
}
/** 写入 data 到 fp 路径所指的文件
    - fp: 目标文件完整路径
    - data: 支持下面几种类型
        - string: 写入文本
        - Uint8Array, Buffer: 写入二进制 buffer
        - any: 通过 JSON.stringify 转为文本后写入文件 */
export async function fwrite(fp, data, { dir, print = true, mkdir = false, } = {}) {
    const is_handle = typeof fp === 'object' && fp && 'fd' in fp;
    if (is_handle) {
        if (print)
            console.log(t('写入'), fp.fp);
    }
    else {
        if (dir) {
            assert(dir.endsWith('/'), t('dir 必须以 / 结尾'));
            fp = dir + fp;
        }
        assert(path.isAbsolute(fp), `${t('fp 必须是绝对路径，当前为:')} ${fp}`);
        if (print)
            console.log(t('写入'), fp);
    }
    if (!isUint8Array(data) && typeof data !== 'string')
        data = to_json(data);
    try {
        await fsp.writeFile(fp, data);
    }
    catch (error) {
        if (!mkdir || error.code !== 'ENOENT' || is_handle)
            throw error;
        await fmkdir(fp.fdir);
        await fsp.writeFile(fp, data);
    }
}
export async function fappend(fp, data, { dir, print = true } = {}) {
    if (dir) {
        assert(dir.endsWith('/'), t('dir 必须以 / 结尾'));
        fp = dir + fp;
    }
    assert(path.isAbsolute(fp), `${t('fp 必须是绝对路径，当前为:')} ${fp}`);
    if (print)
        console.log(t('追加'), fp);
    assert(isUint8Array(data) || typeof data === 'string');
    await fsp.appendFile(fp, data);
}
export async function flist(fpd, options = {}) {
    const { filter, deep = false, absolute = false, print = true, stats = false } = options;
    if (!path.isAbsolute(fpd))
        throw new Error(t('参数 fpd: ') + fpd + t(' 必须是绝对路径'));
    if (!fpd.endsWith('/'))
        throw new Error(t('参数 fpd: ') + fpd + t(' 必须以 / 结尾'));
    // readdir withFileTypes 参数在底层有什么区别，速度上有什么差异
    // 都调用了 uv_fs_scandir, 且调用参数相同，仅仅是 Node.js 侧的回调不同 AfterScanDir / AfterScanDirWithTypes
    // 回调中通过 uv_fs_scandir_next 获取到每个条目的信息，而 uv_fs_scandir_next 中都会读取 type
    // 速度上：都在 0.2 ms 左右就可以完成
    const files = await fsp.readdir(fpd, { withFileTypes: true, encoding: 'utf-8' });
    const filter_regexp = filter instanceof RegExp;
    const filter_fn = Boolean(filter && !filter_regexp);
    let fps = [];
    for (const file of files) {
        const fp = (absolute ? fpd : '') +
            file.name +
            (file.isDirectory() ? '/' : '');
        if (filter_regexp && !filter.test(fp))
            continue;
        if (filter_fn && !filter(fp))
            continue;
        if (print)
            console.log(fp);
        fps.push(fp);
    }
    if (deep)
        return (await Promise.all(fps.map(async (fp) => fp.endsWith('/') ?
            [
                fp,
                ...(await flist(absolute ? fp : fpd + fp, options)).map(fp_ => absolute ? fp_ : fp + fp_)
            ]
            :
                fp))).flat();
    else if (stats)
        return Promise.all(fps.map(async (fp) => fstat(absolute ? fp : fpd + fp)));
    else
        return fps;
}
export async function fstat(fp) {
    if (!path.isAbsolute(fp))
        throw new Error('fp: ' + fp + t(' 必须是绝对路径'));
    let stat = await fsp.stat(fp, { bigint: true });
    stat.fp = fp;
    return stat;
}
/** 删除文件或文件夹  delete files or folders
    - fp: 文件或文件夹的完整路径  The full path to the file or folder
    - options?:
        - print?: `true`
    
    返回是否实际进行了删除操作
    Returns whether the delete operation actually took place */
export async function fdelete(fp, { print = true } = {}) {
    assert(fp.length >= 6, `fp: ${fp} ${t('不能太短，防止误删文件')}`);
    assert(path.isAbsolute(fp), t('fp 必须是绝对路径'));
    try {
        await fsp.rm(fp, { recursive: true });
        if (print)
            if (fp.endsWith('/'))
                console.log((t('删除了文件夹: ') + fp).red);
            else
                console.log((t('删除了文件: ') + fp).red);
        return true;
    }
    catch (error) {
        if (error.code === 'ENOENT') {
            if (print)
                if (fp.endsWith('/'))
                    console.log(t('文件夹已不存在: ') + fp);
                else
                    console.log(t('文件已不存在: ') + fp);
            return false;
        }
        throw error;
    }
}
/** 复制文件或文件夹  copy file or direcotry
    - fp_src: 源 文件/文件夹 完整路径  src file/directory absolute path
    - fp_dst: 目标 文件/文件夹 完整路径  dst file/directory absolute path
    - options?:
        - print?: `true`
        - overwrite?: `true`
    
    @example
    fcopy('d:/temp/camera/', 'd:/camera/') */
export async function fcopy(fp_src, fp_dst, { print = true, overwrite = true, } = {}) {
    assert(fp_src.endsWith('/') === fp_dst.endsWith('/'), t('fp_src 和 fp_dst 必须同为文件路径或文件夹路径'));
    assert(path.isAbsolute(fp_src) && path.isAbsolute(fp_dst), t('fp_src 和 fp_dst 必须为完整路径'));
    if (print)
        console.log(t('复制'), fp_src, '→', fp_dst);
    await fse.copy(fp_src, fp_dst, { overwrite, errorOnExist: true });
}
/** 移动文件或文件夹  move file or direcotry
    - src: 源 文件/文件夹 完整路径  src file/directory absolute path
    - dst: 目标 文件/文件夹 完整路径  dst file/directory absolute path
    @example
    fmove('d:/temp/camera/', 'd:/camera/') */
export async function fmove(src, dst, { overwrite = false, print = true } = {}) {
    if (src.endsWith('/') !== dst.endsWith('/'))
        throw new Error(t('src 和 dst 必须同为文件路径或文件夹路径'));
    if (!path.isAbsolute(src) || !path.isAbsolute(dst))
        throw new Error(t('src 和 dst 必须为完整路径'));
    if (print)
        console.log(t('移动'), src, '→', dst);
    await fse.move(src, dst, { overwrite });
}
/** 重命名文件  rename file
    - fp:  当前文件名／路径  current filename/path
    - fp_: 新的文件名／路径  new filename/path
    - options?:
        - fpd?: fp 和 fp_ 在同一文件夹内  fp and fp_ is in same directory
        - print?: `true`
        - overwrite?: `true`  默认覆盖（不检查效率更高）  better performance without check */
export async function frename(fp, fp_, { fpd, print = true, overwrite = true } = {}) {
    if (fpd) {
        fp = path.join(fpd, fp);
        fp_ = path.join(fpd, fp_);
    }
    else if (!path.isAbsolute(fp) || !path.isAbsolute(fp_))
        throw new Error(t('fp 和 fp_ 必须是绝对路径'));
    if (print)
        console.log(t('重命名'), fp, '→', fp_);
    if (!overwrite && fexists(fp_))
        throw new Error(t('文件已存在：') + fp_);
    await fsp.rename(fp, fp_);
}
/**
    递归创建文件夹，确保 fpd 指向的文件夹存在  Create folders recursively, make sure the folder pointed to by fpd exists
    返回首个创建的文件夹或 undefined  Returns the first created folder or undefined
    
    - fpd: 文件夹完整路径  Folder full path
    - options?:
        - print?: `true`
        - mode?: `'0o777'` */
export async function fmkdir(fpd, { print = true, mode, } = {}) {
    assert(path.isAbsolute(fpd), t('fpd 必须是绝对路径: ') + fpd);
    assert(fpd.endsWith('/'), t('fpd 必须以 / 结尾: ') + fpd);
    // CallingfsPromises.mkdir() when path is a directory that exists results in a rejection only when recursive is false.
    const fpd_ = (await fsp.mkdir(fpd, { recursive: true, mode }))?.replaceAll('\\', '/');
    if (fpd_) {
        if (print)
            console.log(t('已创建文件夹'), fpd);
    }
    else if (print)
        console.log(t('已存在文件夹'), fpd);
    return fpd_;
}
/** 创建软链接  Create soft links
    - fp_real: 现在真实文件/文件夹的路径  current real file/directory path
    - fp_link: 目标链接文件/文件夹的路径  target file/directory path */
export async function flink(fp_real, fp_link, { junction = false, print = true } = {}) {
    assert(path.isAbsolute(fp_real) && path.isAbsolute(fp_link), t('fp 必须是绝对路径'));
    const is_fpd_real = fp_real.endsWith('/');
    const is_fpd_link = fp_link.endsWith('/');
    assert(is_fpd_real === is_fpd_link, t('fp_real 和 fp_link 必须同为文件路径或文件夹路径'));
    if (fexists(fp_link))
        throw new Error(t('存在同名') + (is_fpd_link ? t('文件夹') : t('文件')) + ': ' + fp_link + t('，无法创建链接'));
    if (print)
        console.log(t('已将源文件 ') + fp_real + t(' 链接到 ') + fp_link);
    if (junction)
        fsp.symlink(fp_real, fp_link, 'junction');
    else
        fsp.symlink(fp_real, fp_link, is_fpd_real ? 'dir' : 'file');
}
/** 打开一个文件并搜索替换某个 pattern  open a file and replace certain pattern */
export async function freplace(fp, pattern, replacement) {
    await fwrite(fp, (await fread(fp))
        .replaceAll(pattern, replacement));
}
//# sourceMappingURL=file.js.map