import './prototype.browser.js';
import { Lock } from './utils.browser.js';
export interface BasicAuth {
    type: 'basic';
    username: string;
    password: string;
}
export interface BearerAuth {
    type: 'bearer';
    token: string;
}
export interface RequestOptions {
    method?: 'GET' | 'POST' | 'PUT' | 'HEAD' | 'DELETE' | 'PATCH';
    queries?: Record<string, any>;
    headers?: Record<string, string> | Headers;
    body?: string | Record<string, any> | ArrayBufferView | ArrayBuffer | Blob | URLSearchParams | FormData;
    type?: 'application/json' | 'application/x-www-form-urlencoded' | 'multipart/form-data';
    encoding?: 'binary';
    retries?: true | number;
    timeout?: number;
    auth?: BasicAuth | BearerAuth;
    cookies?: Record<string, string>;
    cors?: boolean;
}
export interface RequestRawOptions extends RequestOptions {
    raw: true;
}
export interface RequestError extends Error {
    url: URL;
    options: RequestInit;
    response?: {
        /** 状态码 */
        status: number;
        url: string;
        headers: Headers;
        ok: boolean;
        type: ResponseType;
        text: string;
        redirected: boolean;
    };
}
/**
    - url: 可以只有 pathname 部分
    - options?:
        - method?: `有 body 时为 POST, 否则为 GET` 'GET' | 'POST' | ···
        - queries?: 添加到 url 上的参数，是 Record<string, any>，true/false 会被转换为 0/1
        - headers?: http 请求头 (Record<string, string> 或者 Headers 类型)，其中 key 必须是小写的
        - body?: http 请求体，可以是 string, Record<string, any> (会自动 JSON.stringify), ArrayBuffer(View), Blob,
            URLSearchParams (type 为 x-www-form-urlencoded), FormData (type 为 form-data)
        - type?: `'application/json'` 有 body 时设置 http 请求头中的 content-type 头
        - encoding?: 设置为 'binary' 时返回 ArrayBuffer 的 body
        - retries?: `false` 可以传入 true (默认 2 次) 或 重试次数
        - timeout?: `5 * 1000`
        - auth?: BasicAuth | BearerAuth
        - raw?: `false` 传入后返回整个 response  */
export declare function request(url: string | URL): Promise<string>;
export declare function request(url: string | URL, options: RequestRawOptions): Promise<Response>;
export declare function request(url: string | URL, options: RequestOptions & {
    encoding: 'binary';
}): Promise<ArrayBuffer>;
export declare function request(url: string | URL, options: RequestOptions): Promise<string>;
/** 发起 http 请求并将响应体作为 json 解析 */
export declare function request_json<T = any>(url: string, options?: RequestOptions): Promise<T>;
export declare class WebSocketConnectionError extends Error {
    name: string;
    websocket: WebSocket;
    event: CloseEvent | /* error 事件时的 event，没有 event.error */ Event;
    type: 'close' | 'error';
    code?: number;
    reason?: string;
    constructor(websocket: WebSocket, event: CloseEvent | Event, message?: string);
}
/** 连接 websocket url, 设置各种事件监听器。在 open 事件后 resolve, 返回 websocket
    遇到 error 时会创建 WebSocketConnectionError:
        - reject 掉返回的 promise (若此时未 settle)
        - 作为参数调用 on_error (已 settle 且有 on_error 回调)
    可以用 WebSocket.bufferedAmount 来显示大消息的发送进度
    https://developer.mozilla.org/en-US/docs/Web/API/WebSocket/bufferedAmount
    - url
    - options:
        - protocols?
        - on_message: 根据 websocket frame 的 opcode 不同 (text frame 或 binary frame)，event 中的 data 对应为 ArrayBuffer 或者 string
          https://datatracker.ietf.org/doc/html/rfc6455#section-5.2
        - on_error?: 在 websocket 出错和非正常关闭 (close, error 事件) 时都调用，可以根据 error.type 来区分，error 的类型是 WebSocketConnectionError，
            type 为 'close' 时有 code 和 reason 属性
        - on_close?: 和 websocket 的 'close' 事件不相同，只在正常关闭 (close code 为 1000) 时才调用，否则都会调用 on_error
            https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent#Status_codes  */
export declare function connect_websocket(url: string | URL, { protocols, on_message, on_error, on_close, }: {
    protocols?: string | string[];
    on_message(data: ArrayBuffer | string, websocket: WebSocket): any;
    on_error?(error: WebSocketConnectionError, websocket: WebSocket): any;
    on_close?(event: CloseEvent, websocket: WebSocket): any;
}): Promise<WebSocket>;
/** 接收到消息后的处理函数
    返回值可以是:
    - 数组: 会自动被封装为 { id: 相同, data: 返回值, done: true } 这样的消息并调用 websocket.send 将其发送
    - void: 什么都不做
    - 以上的 promise */
export type MessageHandler = (message: Message, websocket?: WebSocket) => void | any[] | Promise<void | any[]>;
/** 二进制消息格式
    - json.length (小端序): 4 字节
    - json 数据
    - binary 数据 */
export interface Message<TData extends any[] = any[]> {
    /** rpc id: 在 rpc 系统中认为是唯一的。用来在单个 websocket 连接上复用多个 rpc 请求。多个相同 id 的 message 组成一个请求流 */
    id?: number;
    /** 只在 rpc 发起时指定被调用的 function name，发起时 rpc 时必传 */
    func?: string;
    /** 通过这个 flag 主动表明这是发往对方的最后一个 message, 对方可以销毁 handler 了
        并非强制，可以不说明，由双方的函数自己约定
    */
    done?: boolean;
    /** 通知对方这里产生的错误，本质上类似 data 也是一种数据，并不代表 rpc 的结束，后续可能继续有 rpc message 交换 */
    error?: Error;
    /** data 是一个数组, 作为:
        - rpc 发起方调用 func 的参数，或者请求流 message 携带的数据
        - 结果或者响应流的数据，传给请求发起方
        
        里面是可序列化为 json 的 js 变量，或者是 Uint8Array
        Uint8Array 的参数处理后被替换为 Uint8Array.byteLength, 并将下标记录在 bins 中
        
        注意: 数组中如果有 undefined 值，传输到对面会变成 null 值 */
    data?: TData;
    /** bins: data 中哪些下标对应的原始值是 Uint8Array 类型的，如: [0, 3] */
    bins?: number[];
}
/** 通过创建 remote 对象对 websocket rpc 进行抽象
    调用方使用 remote.call 进行调用
    被调方在创建 remote 对象时传入 funcs 注册处理函数，并使用 remote.handle 方法处理 websocket message
    未连接时自动连接，断开后自动重连 (保证执行 send 时已连接，否则报错)
    从设计上与 rpc 状态与底层连接的状态无关 (通过传入 url 创建的)，底层连接断开后，只要检测到断线，会尝试重连，如果重连成功，则不影响调用的状态
    @example
    // Zero 继承自 Remote 并通过 call 实现了一些方法
    let zero = new Zero({ local: true })
    
    // 一元 rpc
    await zero.repl_ts('1234')
    
    // 订阅流
    const id = genid()
    
    zero.handlers.set(id, ({ data: [chunk] }: Message<[Uint8Array]>) => {
        term.write(chunk)
    })
    
    zero.send({ id, func: 'subscribe_stdio' }) */
export declare class Remote {
    /** 是否为建立 rpc 连接的发起方 */
    initiator: boolean;
    /** websocket url */
    url?: string;
    /** 主动发起 websocket 连接的客户端 (构造函数传 url) 有这个属性 */
    lwebsocket?: Lock<WebSocket>;
    /** 通过 rpc message.func 被调用的 rpc 函数 */
    funcs: Record<string, MessageHandler>;
    /** map<id, message handler>: 通过 rpc message.id 找到对应的 handler, unary rpc 接收方不需要设置 handlers, 发送方需要 */
    handlers: Map<number, MessageHandler>;
    print: boolean;
    pconnect: Promise<any>;
    /** 使用 Uint8Array 作为参数更灵活 https://stackoverflow.com/a/74505197/7609214 */
    static parse<TData extends any[] = any[]>(buffer: Uint8Array): Message<TData>;
    static pack({ id, func, data, done, error }: Message): Uint8Array;
    /** 作为 rpc 发起方，可以通过传入 url 或者 websocket 定义远程 Remote
        作为 rpc 接收方，可以不传 url 和 websocket 定义本地 Remote */
    constructor({ url, funcs, websocket, on_error }?: {
        url?: string;
        funcs?: Remote['funcs'];
        websocket?: WebSocket;
        on_error?(error: WebSocketConnectionError, websocket: WebSocket): void;
    });
    on_error?(error: WebSocketConnectionError, websocket: WebSocket): void;
    /** 幂等，保证 websocket 已连接，否则抛出异常，不需要手动调用，在其它方法中已默认自动重连
        作为接收方需要传入使用的 websocket 连接，确保这个这个连接的状态 */
    connect(websocket?: WebSocket): Promise<void>;
    disconnect(): void;
    /** 接收 websocket 连接的本地 remote 必传 websocket 参数；发起端选传，如果传了必须等于 this.websocket_lock.resource
        发送或连接出错时自动清理 message.id 对应的 handler */
    send(message: Message, websocket?: WebSocket): Promise<void>;
    /** 处理接收到的 websocket message 并解析, 根据 id dispatch 到对应的 handler 进行处理
        如果 message.done == true 则清理 handler
        如果 handler 返回了值，则包装为 message 发送
        使用 Uint8Array 作为参数更灵活 https://stackoverflow.com/a/74505197/7609214  */
    handle(data: Uint8Array, websocket: WebSocket): Promise<void>;
    /** 调用 remote 中的 func, 只适用于最简单的一元 rpc (请求, 响应) */
    call<TReturn extends any[] = any[]>(func: string, args?: any[]): Promise<TReturn>;
}
