import { t } from './i18n/instance.js';
import './prototype.browser.js'; // to_time_str()
import { assert, concat, genid, delay, Lock } from './utils.browser.js';
// ------------------------------------ fetch, request
const drop_request_headers = new Set([
    // : 开头的 key
    // sec-*
    'accept-charset',
    'connection',
    'content-length',
    'keep-alive',
    'trailer',
    'transfer-encoding',
    'upgrade',
]);
async function fetch_retry(url, options, timeout, retries = 0, count = 0) {
    try {
        // 旧的浏览器可能没有这个函数
        if (AbortSignal.timeout)
            options.signal = AbortSignal.timeout(timeout);
        return await fetch(url, options);
    }
    catch (error) {
        if (count >= retries ||
            error.name !== 'TimeoutError' && !['ECONNRESET', 'ETIMEDOUT', 'ESOCKETTIMEDOUT'].includes(error.cause?.code))
            throw error;
        else {
            const duration = 2 ** count;
            console.log(`${t('等待 {{duration}} 秒后重试 fetch ({{_count}}) …', { duration, _count: count })}  ${url.toString()}`);
            await delay(1000 * duration);
            return fetch_retry(url, options, timeout, retries, count + 1);
        }
    }
}
export async function request(url, { method, queries, headers: _headers, body, type = 'application/json', encoding, retries, timeout = 5 * 1000, auth, raw = false, cors, } = {}) {
    url = new URL(url, location.href);
    if (queries)
        for (const key in queries) {
            let value = queries[key];
            if (typeof value === 'boolean')
                value = value ? '1' : '0';
            url.searchParams.append(key, value);
        }
    if (body !== undefined && !method)
        method = 'POST';
    if (retries === true)
        retries = 2;
    // --- headers, http/2 开始都用小写的 headers
    let headers = new Headers();
    if (body !== undefined)
        headers.set('content-type', type);
    if (auth)
        headers.set('authorization', auth.type === 'basic' ? `Basic ${`${auth.username}:${auth.password}`.to_base64()}` : `Bearer ${auth.token}`);
    if (_headers)
        if (_headers instanceof Headers)
            // @ts-ignore: ts 类型不支持，实际上已经有了
            for (const [key, value] of _headers) {
                if (!key.startsWith(':') && !key.startsWith('sec-') && !drop_request_headers.has(key))
                    headers.set(key, value);
            }
        else
            for (const key in _headers)
                if (!key.startsWith(':') && !key.startsWith('sec-') && !drop_request_headers.has(key)) { // 可能在 http/2 的 response 中会有这样开头的保留 headers, 在透传时忽略比较好
                    assert(key === key.toLowerCase(), t('传入 request 的 headers 参数中 key 应该都是小写的，实际为 {{key}}', { key }));
                    headers.set(key, _headers[key]);
                }
    let options = {
        ...method ? { method } : {},
        keepalive: true,
        redirect: 'follow',
        credentials: 'include',
        ...cors ? { mode: 'cors' } : {},
        headers,
        // --- body
        body: (() => {
            if (body === undefined)
                return;
            switch (type) {
                case 'application/json':
                    return typeof body === 'string' ||
                        ArrayBuffer.isView(body) ||
                        body instanceof ArrayBuffer ||
                        body instanceof Blob ?
                        body
                        :
                            JSON.stringify(body);
                case 'application/x-www-form-urlencoded':
                    return body instanceof URLSearchParams ? body : new URLSearchParams(body);
                case 'multipart/form-data':
                    if (body instanceof FormData)
                        return body;
                    else {
                        let form = new FormData();
                        for (const key in body) {
                            let value = body[key];
                            form.set(key, value);
                        }
                        return form;
                    }
            }
        })(),
    };
    let response;
    try {
        response = await fetch_retry(url, options, timeout, retries);
        if (!response.ok)
            throw Object.assign(new Error(t('状态码 {{status}}, 非 2xx', { status: response.status })), { name: 'StatusCodeError' });
    }
    catch (error) {
        error.url = url;
        error.options = options;
        if (response)
            error.response = {
                status: response.status,
                url: response.url,
                headers: response.headers,
                ok: response.ok,
                type: response.type,
                text: await response.text(),
                redirected: response.redirected
            };
        throw error;
    }
    if (raw)
        return response;
    if (!response.body)
        return response.body;
    if (encoding === 'binary')
        return response.arrayBuffer();
    else
        return response.text();
}
/** 发起 http 请求并将响应体作为 json 解析 */
export async function request_json(url, options) {
    const resp = await request(url, options);
    if (!resp)
        return;
    try {
        return JSON.parse(resp);
    }
    catch (error) {
        console.error(resp);
        throw error;
    }
}
let decoder = new TextDecoder();
let encoder = new TextEncoder();
export class WebSocketConnectionError extends Error {
    name = 'WebSocketConnectionError';
    websocket;
    event;
    type;
    code;
    reason;
    constructor(websocket, event, message) {
        super(`${websocket.url} ${t('连接出错了')}. ${message || ''}`);
        this.websocket = websocket;
        this.event = event;
        this.type = event.type;
        if (this.type === 'close') {
            this.code = event.code;
            this.reason = event.reason;
        }
    }
}
/** 连接 websocket url, 设置各种事件监听器。在 open 事件后 resolve, 返回 websocket
    遇到 error 时会创建 WebSocketConnectionError:
        - reject 掉返回的 promise (若此时未 settle)
        - 作为参数调用 on_error (已 settle 且有 on_error 回调)
    可以用 WebSocket.bufferedAmount 来显示大消息的发送进度
    https://developer.mozilla.org/en-US/docs/Web/API/WebSocket/bufferedAmount
    - url
    - options:
        - protocols?
        - on_message: 根据 websocket frame 的 opcode 不同 (text frame 或 binary frame)，event 中的 data 对应为 ArrayBuffer 或者 string
          https://datatracker.ietf.org/doc/html/rfc6455#section-5.2
        - on_error?: 在 websocket 出错和非正常关闭 (close, error 事件) 时都调用，可以根据 error.type 来区分，error 的类型是 WebSocketConnectionError，
            type 为 'close' 时有 code 和 reason 属性
        - on_close?: 和 websocket 的 'close' 事件不相同，只在正常关闭 (close code 为 1000) 时才调用，否则都会调用 on_error
            https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent#Status_codes  */
export async function connect_websocket(url, { protocols, on_message, on_error, on_close, }) {
    let websocket = new WebSocket(url, protocols);
    // https://stackoverflow.com/questions/11821096/what-is-the-difference-between-an-arraybuffer-and-a-blob/39951543
    websocket.binaryType = 'arraybuffer';
    return new Promise((resolve, reject) => {
        let settled = false;
        websocket.addEventListener('open', event => {
            console.log(websocket.url +
                (protocols ?
                    ' ' + (typeof protocols === 'string' ? protocols : protocols.join(', ').bracket('square')).bracket()
                    :
                        '') +
                t(' 已连接'));
            settled = true;
            resolve(websocket);
        });
        websocket.addEventListener('close', event => {
            if (event.code === 1000) // 正常关闭
                if (on_close)
                    on_close(event, websocket);
                else
                    console.log(`${websocket.url} ${t('已正常关闭')}`);
            else { // 异常关闭，认为发生了错误，进行错误处理
                const error = new WebSocketConnectionError(websocket, event, `${t('连接被关闭')}, code: ${event.code}${event.reason ? `, ${t('原因')}: ${event.reason}` : ''}`);
                if (settled)
                    if (on_error)
                        on_error(error, websocket);
                    else // 既然用户不传 on_error, 就当 unhandled error 抛出来
                        throw error;
                else {
                    settled = true;
                    reject(error);
                }
            }
        });
        websocket.addEventListener('error', event => {
            const error = new WebSocketConnectionError(websocket, event);
            // https://blog.insiderattack.net/promises-next-ticks-and-immediates-nodejs-event-loop-part-3-9226cbe7a6aa
            // close 的错误信息比较多，这里延后触发 error 事件，放到微任务队列之后的 timers 队列中
            setTimeout(() => {
                if (settled)
                    if (on_error)
                        on_error(error, websocket);
                    else // 既然用户不传 on_error, 就当 unhandled error 抛出来
                        throw error;
                else {
                    settled = true;
                    reject(error);
                }
            });
        });
        websocket.addEventListener('message', event => {
            on_message(event.data, websocket);
        });
    });
}
/** 通过创建 remote 对象对 websocket rpc 进行抽象
    调用方使用 remote.call 进行调用
    被调方在创建 remote 对象时传入 funcs 注册处理函数，并使用 remote.handle 方法处理 websocket message
    未连接时自动连接，断开后自动重连 (保证执行 send 时已连接，否则报错)
    从设计上与 rpc 状态与底层连接的状态无关 (通过传入 url 创建的)，底层连接断开后，只要检测到断线，会尝试重连，如果重连成功，则不影响调用的状态
    @example
    // Zero 继承自 Remote 并通过 call 实现了一些方法
    let zero = new Zero({ local: true })
    
    // 一元 rpc
    await zero.repl_ts('1234')
    
    // 订阅流
    const id = genid()
    
    zero.handlers.set(id, ({ data: [chunk] }: Message<[Uint8Array]>) => {
        term.write(chunk)
    })
    
    zero.send({ id, func: 'subscribe_stdio' }) */
export class Remote {
    /** 是否为建立 rpc 连接的发起方 */
    initiator;
    /** websocket url */
    url;
    /** 主动发起 websocket 连接的客户端 (构造函数传 url) 有这个属性 */
    lwebsocket;
    /** 通过 rpc message.func 被调用的 rpc 函数 */
    funcs;
    /** map<id, message handler>: 通过 rpc message.id 找到对应的 handler, unary rpc 接收方不需要设置 handlers, 发送方需要 */
    handlers = new Map();
    print = false;
    pconnect;
    /** 使用 Uint8Array 作为参数更灵活 https://stackoverflow.com/a/74505197/7609214 */
    static parse(buffer) {
        const dv = new DataView(buffer.buffer, buffer.byteOffset, buffer.byteLength);
        const len_json = dv.getUint32(0, true);
        let offset = 4 + len_json;
        let message = JSON.parse(decoder.decode(buffer.subarray(4, offset)));
        if (message.bins) {
            const { bins } = message;
            let { data } = message;
            for (const ibin of bins) {
                const len_buf = data[ibin];
                data[ibin] = buffer.subarray(offset, offset + len_buf);
                offset += len_buf;
            }
        }
        if (message.error)
            message.error = Object.assign(new Error(), message.error);
        return message;
    }
    static pack({ id, func, data = [], done, error }) {
        let data_ = new Array(data.length);
        let bins = [];
        let bufs = [];
        for (let i = 0; i < data.length; i++) {
            const item = data[i];
            if (item instanceof Uint8Array) {
                bins.push(i);
                bufs.push(item);
                data_[i] = item.length;
            }
            else
                data_[i] = item;
        }
        // 有可能 data_json 含有循环引用导致 JSON.stringify 报错
        const data_json = {
            id,
            ...func ? { func } : {},
            ...done ? { done } : {},
            ...error ? { error } : {},
            ...data_.length ? { data: data_ } : {},
            ...bins.length ? { bins } : {},
        };
        const str_json = encoder.encode(JSON.stringify(data_json));
        let dv = new DataView(new ArrayBuffer(4));
        dv.setUint32(0, str_json.length, true);
        return concat([dv, str_json, ...bufs]);
    }
    /** 作为 rpc 发起方，可以通过传入 url 或者 websocket 定义远程 Remote
        作为 rpc 接收方，可以不传 url 和 websocket 定义本地 Remote */
    constructor({ url, funcs = {}, websocket, on_error } = {}) {
        assert(!(url && websocket), t('构建 Remote 时 url 和 websocket 最多只能传一个'));
        this.initiator = Boolean(url || websocket);
        this.url = url;
        this.funcs = funcs;
        if (on_error)
            this.on_error = on_error;
        this.lwebsocket = new Lock(websocket || null);
    }
    on_error(error, websocket) {
        throw error;
    }
    /** 幂等，保证 websocket 已连接，否则抛出异常，不需要手动调用，在其它方法中已默认自动重连
        作为接收方需要传入使用的 websocket 连接，确保这个这个连接的状态 */
    async connect(websocket) {
        if (this.initiator)
            if (this.lwebsocket.resource?.readyState === WebSocket.OPEN)
                return;
            else if (!this.url)
                throw new Error(t('创建 Remote 时传入的 websocket 连接已断开'));
            else
                // 假设有多个请求想要并发连接 websocket, 且此时 websocket 是断开的状态
                // 应该排队依次连接，而不是后续的连接直接使用第一次连接的 promise，后续调用还是应该尝试重连（不止连接一次）
                return this.lwebsocket.request(async (websocket) => {
                    // 保存的 rpc 状态在 this.handlers, 与 websocket 无关，因此即使断开重连也不影响 rpc 的运行，即
                    // 底层连接断开后自动重连对上层应该是无感知的，除非再次连接时失败
                    if (websocket?.readyState === WebSocket.OPEN)
                        return;
                    else if (!this.url)
                        throw new Error(t('创建 Remote 时传入的 websocket 连接已断开'));
                    else // 重连
                        this.lwebsocket.resource = await connect_websocket(this.url, {
                            on_message: (data, websocket) => {
                                this.handle(new Uint8Array(data), websocket);
                            },
                            on_error: this.on_error.bind(this)
                        });
                });
        else if (websocket.readyState === WebSocket.OPEN)
            return;
        else
            throw new Error(t('传入的 websocket 连接已断开'));
    }
    disconnect() {
        this.lwebsocket.resource?.close(1000);
    }
    /** 接收 websocket 连接的本地 remote 必传 websocket 参数；发起端选传，如果传了必须等于 this.websocket_lock.resource
        发送或连接出错时自动清理 message.id 对应的 handler */
    async send(message, websocket) {
        try {
            if (this.initiator) {
                assert(!websocket || websocket === this.lwebsocket.resource);
                await this.connect();
                websocket = this.lwebsocket.resource;
            }
            else
                await this.connect(websocket);
            if (!message.id)
                message.id = genid();
            // 不需要独占 websocket
            websocket.send(Remote.pack(message));
        }
        catch (error) {
            if (message.id)
                this.handlers.delete(message.id);
            throw error;
        }
    }
    /** 处理接收到的 websocket message 并解析, 根据 id dispatch 到对应的 handler 进行处理
        如果 message.done == true 则清理 handler
        如果 handler 返回了值，则包装为 message 发送
        使用 Uint8Array 作为参数更灵活 https://stackoverflow.com/a/74505197/7609214  */
    async handle(data, websocket) {
        const message = Remote.parse(data);
        const { id, func, done } = message;
        if (this.print)
            console.log(message);
        let handler;
        if (func)
            handler = this.funcs[func];
        else {
            handler = this.handlers.get(id);
            if (done)
                this.handlers.delete(id);
        }
        try {
            if (handler) {
                const data = await handler(message, websocket);
                if (data)
                    await this.send({ id, data }, websocket);
            }
            else if (message.error)
                throw message.error;
            else
                throw new Error(`${t('找不到 rpc handler')}: ${func ? `func: ${func.quote()}` : `id: ${id}`}`);
        }
        catch (error) {
            // handler 出错并不意味着 rpc 一定会结束，可能 error 是运行中的正常数据，所以不能清理 handler
            if (websocket.readyState === WebSocket.OPEN &&
                !message.error // 防止无限循环往对方发送 error, 只有在对方无错误时才可以发送
            )
                await this.send({ id, error, /* 不能设置 done 清理对面 handler, 理由同上 */ }, websocket);
            throw error;
        }
    }
    /** 调用 remote 中的 func, 只适用于最简单的一元 rpc (请求, 响应) */
    async call(func, args) {
        return new Promise(async (resolve, reject) => {
            const id = genid();
            this.handlers.set(id, (message) => {
                const { error, data } = message;
                if (error)
                    reject(error);
                else
                    resolve(data);
                this.handlers.delete(id);
            });
            try {
                await this.send({ id, func, data: args }); // 不需要 done: true, 因为对面的 remote.handlers 中不会有这个 id 的 handler
            }
            catch (error) {
                reject(error);
            }
        });
    }
}
//# sourceMappingURL=net.browser.js.map