/// <reference types="node" resolution-mode="require"/>
/// <reference types="node" resolution-mode="require"/>
/// <reference types="node" resolution-mode="require"/>
import { type ChildProcess } from 'child_process';
import './prototype.js';
import { Encoding } from './file.js';
import { inspect } from './utils.js';
export declare const exe_nodejs: string;
interface StartOptions {
    /** `继承当前工作目录 process.cwd()` 子进程的工作目录  `inherit the cwd` cwd of the child process. */
    cwd?: string;
    /** `process.env` 覆盖/添加到 process.env 的环境变量  overwrite/add to process.env */
    env?: Record<string, string>;
    /** `'utf-8'` 子进程输出编码  child output encoding */
    encoding?: Encoding | 'binary';
    /** `true` print 选项，支持设置细项  print option (with details) */
    print?: boolean | {
        stdout: boolean;
        stderr: boolean;
        command: boolean;
        code: boolean;
    };
    /** `'pipe'` 设置为 'ignore' 时忽略 stdio 处理  when 'ignore' then ignore stdio processing */
    stdio?: 'pipe' | 'ignore' | ['pipe' | 'ignore' | 'inherit', 'pipe' | 'ignore' | 'inherit', 'pipe' | 'ignore' | 'inherit'];
    /** `false` 是否断开和 child 的关系 (ignore stdio, unref)  whether to break the connection with child (ignore stdio, unref) */
    detached?: boolean;
    /** 为 true 时会设置 UV_PROCESS_WINDOWS_HIDE
        然后设置启动进程的参数 CREATE_NO_WINDOW, 和 SW_HIDE
        d:/0/libuv/src/win/process.c
        CREATE_NO_WINDOW:
        The process is a console application that is being run without a console window.
        Therefore, the console handle for the application is not set.
        This flag is ignored if the application is not a console application, or
        if it is used with either CREATE_NEW_CONSOLE or DETACHED_PROCESS.
        
        具体有什么用还不清楚
    */
    window?: boolean;
}
/** start process
    - exe: .exe 路径或文件名 (建议使用完整路径，跳过 path 搜索，性能更高)  path or filename (full path is recommanded to skip path searching for better perf)
    - args: `[]` 参数列表  arguments list
    - options
        - cwd?: `fp_root`
        - env?: `process.env` 覆盖／添加到 process.env 的环境变量  overwrite/add to process.env
        - encoding?: `'utf-8'` 子进程输出编码  child output encoding
        - print?: `true` print 选项，支持设置细项  print option (with details)
        - stdio?: `'pipe'` 设置为 'ignore' 时忽略 stdio 处理  when 'ignore' then ignore stdio processing
        - detached?: `false` 是否断开和 child 的关系 (ignore stdio, unref)  whether to break the connection with child (ignore stdio, unref)
*/
export declare function start(exe: string, args?: string[], { cwd, encoding, print, stdio, detached, env, window: _window, }?: StartOptions): Promise<ChildProcess>;
export interface CallOptions extends StartOptions {
    throw_code?: boolean;
    input?: string;
}
export interface CallResult<TOutput extends string | Buffer = string> {
    pid: number;
    stdout: TOutput;
    stderr: TOutput;
    code: number | null;
    signal: NodeJS.Signals | null;
    child: ChildProcess;
    [inspect.custom](): string;
}
/** call process for result
    - exe: .exe 路径或文件名 (建议使用路径，跳过 path 搜索，性能更高)  path or filename (full path is recommanded to skip path searching for better perf)
    - args: `[]` 参数列表  arguments list
    - options?:
        - cwd?: `'d:/'`
        - env?: `process.env` 覆盖／添加到 process.env 的环境变量  overwrite/add to process.env
        - encoding?: `'utf-8'` 子进程输出编码, 设置为 binary 时返回 stdout, stderr 类型为 Buffer; 否则是 string
            child output encoding. When set to binary, return stdout, stderr is of type Buffer; otherwise it is string
        - print?: `true` print 选项，支持设置细项  print option (with details)
        - stdio?: `'pipe'` 设置为 'ignore' 时忽略 stdio 处理  when 'ignore' then ignore stdio processing
        - input?: string, start 子进程之后写入到子进程 stdin 中的内容  After starting the sub-process, write to the content of the sub-process STDIN
        - detached?: `false` 是否断开和 child 的关系 (ignore stdio, unref)  whether to break the connection with child (ignore stdio, unref)
        - throw_code?: `true` code 不为 0 时是否抛出异常  whether to throw Error when code is not 0 */
export declare function call(exe: string, args?: string[]): Promise<CallResult<string>>;
export declare function call(exe: string, args?: string[], options?: CallOptions & {
    encoding: 'binary';
}): Promise<CallResult<Buffer>>;
export declare function call(exe: string, args?: string[], options?: CallOptions): Promise<CallResult<string>>;
/** call node <js> for result
    - js: .js 路径 (相对路径根据 cwd 解析)  path (relative path will resolve based on cwd)
    - args: `[]` 参数列表  arguments list
    - options
        - cwd?: `'d:/'`
        - env?: `process.env` 覆盖／添加到 process.env 的环境变量  overwrite/add to process.env
        - encoding?: `'utf-8'` 子进程输出编码  child process output encoding
        - print?: `true` print 选项，支持设置细项  print option (with details)
        - stdio?: `'pipe'` 设置为 'ignore' 时忽略 stdio 处理  when 'ignore' then ignore stdio processing
        - detached?: `false` 是否断开和 child 的关系 (ignore stdio, unref)  whether to break the connection with child (ignore stdio, unref)
        - throw_code?: `true` code 不为 0 时是否抛出异常  whether to throw Error when code is not 0 */
export declare function call_nodejs(js: string, args?: string[], options?: CallOptions): Promise<CallResult<string>>;
export interface TermOptions {
    /** `d:/t/` */
    cwd?: string;
    /** `true` 打印参数 */
    print?: boolean;
    title?: string;
}
export declare const exe_winterm: string;
/** 新建 terminal 窗口执行进程  New Terminal window execution process
    - exe: exe 文件路径  exe file path
    - args: 调用参数  call parameter
    - options?: WinTermOptions
*/
export declare function term(exe: string, args?: string[], { cwd, print, title, }?: TermOptions): Promise<ChildProcess>;
export interface TermNodeOptions extends TermOptions {
    /** nodejs debugger port */
    inspect?: number;
    /** break at first line */
    break?: boolean;
    /** 使用 ts-node 直接运行  use ts-node to run directly */
    tsnode?: boolean;
    /** `false` --experimental-specifier-resolution=node */
    resolve?: boolean;
    /** `false` --trace-warnings */
    trace_warnings?: boolean;
}
/** 在 term tab 中创建 node.exe 进程  create the node.exe process in term tab */
export declare function term_nodejs(fp_js: string, args?: string[], { title, inspect, tsnode, break: _break, resolve, trace_warnings, ..._options }?: TermNodeOptions): Promise<ChildProcess>;
export {};
