import { spawn } from 'child_process';
import { userInfo } from 'os';
import { t } from './i18n/instance.js';
import './prototype.js';
import { inspect, DecoderStream } from './utils.js';
export const exe_nodejs = process.execPath.to_slash();
/** start process
    - exe: .exe 路径或文件名 (建议使用完整路径，跳过 path 搜索，性能更高)  path or filename (full path is recommanded to skip path searching for better perf)
    - args: `[]` 参数列表  arguments list
    - options
        - cwd?: `fp_root`
        - env?: `process.env` 覆盖／添加到 process.env 的环境变量  overwrite/add to process.env
        - encoding?: `'utf-8'` 子进程输出编码  child output encoding
        - print?: `true` print 选项，支持设置细项  print option (with details)
        - stdio?: `'pipe'` 设置为 'ignore' 时忽略 stdio 处理  when 'ignore' then ignore stdio processing
        - detached?: `false` 是否断开和 child 的关系 (ignore stdio, unref)  whether to break the connection with child (ignore stdio, unref)
*/
export async function start(exe, args = [], { cwd, encoding = 'utf-8', print = true, stdio = 'pipe', detached = false, env, window: _window = true, } = {}) {
    const options = {
        cwd,
        shell: false,
        windowsHide: !_window,
        stdio,
        ...env ? {
            env: { ...process.env, ...env }
        } : {},
    };
    if (typeof print === 'boolean')
        print = {
            stdout: print,
            stderr: print,
            command: print,
            code: print,
        };
    if (typeof stdio === 'string')
        stdio = [stdio, stdio, stdio];
    if (print.command)
        console.log(((short_exe_names[exe] || (exe.includes(' ') ? exe.quote() : exe)) +
            (args.length ?
                ' ' + args.filter(arg => arg !== '--suppressApplicationTitle')
                    .map(arg => arg.includes(' ') ? arg.quote() : arg)
                    .join(' ')
                :
                    '')).blue);
    if (detached) {
        let child = spawn(exe, args, {
            ...options,
            stdio: 'ignore',
            detached: true,
        });
        child.unref();
        return child;
    }
    let child = spawn(exe, args, options);
    // 防止 child spawn 失败时 crash nodejs 进程
    child.on('error', error => {
        console.error(error);
    });
    if (stdio[0] === 'pipe')
        child.stdin.setDefaultEncoding('utf8');
    if (stdio.every(s => s === 'ignore'))
        return child;
    if (encoding !== 'binary') {
        if (stdio[1] === 'pipe') {
            if (encoding === 'utf-8')
                child.stdout.setEncoding('utf-8');
            else
                child.stdout = child.stdout.pipe(new DecoderStream(encoding));
            if (print.stdout)
                child.stdout.pipe(process.stdout, { end: false });
        }
        if (stdio[2] === 'pipe') {
            if (encoding === 'utf-8')
                child.stderr.setEncoding('utf-8');
            else
                child.stderr = child.stderr.pipe(new DecoderStream(encoding));
            if (print.stderr)
                child.stderr.pipe(process.stderr, { end: false });
        }
    }
    return child;
}
export async function call(exe, args = [], options = {}) {
    const { encoding = 'utf-8', throw_code = true, input, } = options;
    let { stdio = 'pipe', print = true } = options;
    if (typeof print === 'boolean')
        print = {
            command: print,
            stdout: print,
            stderr: print,
            code: print,
        };
    if (typeof stdio === 'string')
        stdio = [stdio, stdio, stdio];
    const cmd = (short_exe_names[exe] || (exe.includes(' ') ? exe.quote() : exe)) +
        (args.length ? (' ' + args.map(arg => arg.includes(' ') ? arg.quote() : arg)
            .join(' '))
            :
                '');
    if (print.command)
        console.log(cmd.blue);
    let child = await start(exe, args, {
        ...options,
        print: false,
    });
    if (input)
        child.stdin.write(input);
    // --- collect output
    let stdouts = [];
    let stderrs = [];
    let code, signal;
    await Promise.all([
        new Promise(resolve => {
            child.once('exit', (_code, _signal) => {
                code = _code;
                signal = _signal;
                resolve();
            });
        }),
        (async () => {
            if (stdio[1] === 'pipe')
                for await (const chunk of child.stdout) {
                    if (encoding !== 'binary' && print.stdout)
                        process.stdout.write(chunk);
                    stdouts.push(chunk);
                }
        })(),
        (async () => {
            if (stdio[2] === 'pipe')
                for await (const chunk of child.stderr) {
                    if (encoding !== 'binary' && print.stderr)
                        process.stderr.write(chunk);
                    stderrs.push(chunk);
                }
        })()
    ]);
    const message = `${t('进程')} ${cmd} ` + (code ?
        `(pid: ${child.pid}) ${throw_code ? t('异常结束') : t('结束')}, code: ${code}${signal ? `, signal: ${signal}` : ''}`
        :
            t('正常结束'));
    if (print.code || (throw_code && (code || signal)))
        console.log(message[(throw_code && (code || signal)) ? 'red' : 'blue']);
    const result = {
        pid: child.pid,
        stdout: encoding === 'binary' ?
            Buffer.concat(stdouts)
            :
                stdouts.join(''),
        stderr: encoding === 'binary' ?
            Buffer.concat(stderrs)
            :
                stderrs.join(''),
        code,
        signal,
        child,
        [inspect.custom]() {
            return inspect(this, { omit: ['child'] });
        }
    };
    if (code && throw_code)
        throw Object.assign(new Error(message), { result });
    return result;
}
/** call node <js> for result
    - js: .js 路径 (相对路径根据 cwd 解析)  path (relative path will resolve based on cwd)
    - args: `[]` 参数列表  arguments list
    - options
        - cwd?: `'d:/'`
        - env?: `process.env` 覆盖／添加到 process.env 的环境变量  overwrite/add to process.env
        - encoding?: `'utf-8'` 子进程输出编码  child process output encoding
        - print?: `true` print 选项，支持设置细项  print option (with details)
        - stdio?: `'pipe'` 设置为 'ignore' 时忽略 stdio 处理  when 'ignore' then ignore stdio processing
        - detached?: `false` 是否断开和 child 的关系 (ignore stdio, unref)  whether to break the connection with child (ignore stdio, unref)
        - throw_code?: `true` code 不为 0 时是否抛出异常  whether to throw Error when code is not 0 */
export async function call_nodejs(js, args = [], options) {
    return call(exe_nodejs, [js, ...args], options);
}
export const exe_winterm = `C:/Users/${userInfo().username}/AppData/Local/Microsoft/WindowsApps/wt.exe`;
/** 新建 terminal 窗口执行进程  New Terminal window execution process
    - exe: exe 文件路径  exe file path
    - args: 调用参数  call parameter
    - options?: WinTermOptions
*/
export async function term(exe, args = [], { cwd = 'd:/t/', print = true, title,
// env
 } = {}) {
    return start(exe_winterm, [
        'new-tab',
        '-d', cwd,
        ...title ? ['--suppressApplicationTitle', '--title', title] : [],
        exe,
        ...args.map(arg => arg.replaceAll(';', '\\;')),
    ], {
        print,
        detached: true,
        cwd,
        // env
    });
}
/** 在 term tab 中创建 node.exe 进程  create the node.exe process in term tab */
export async function term_nodejs(fp_js, args = [], { title, inspect, tsnode = false, break: _break, resolve = false, trace_warnings = false, ..._options } = {}) {
    return term(exe_nodejs, [
        ...trace_warnings ? ['--trace-warnings'] : [],
        ...resolve ? ['--experimental-specifier-resolution=node'] : [],
        ...title ? [`--title=${title}`] : [],
        ...inspect ? [
            _break ? `--inspect-brk=0.0.0.0:${inspect}` : `--inspect=0.0.0.0:${inspect}`
        ] : [],
        ...tsnode ? [
            '--loader=ts-node/esm',
            '--experimental-specifier-resolution=node'
        ] : [],
        fp_js,
        ...args
    ], {
        title,
        ..._options,
    });
}
const short_exe_names = {
    [exe_winterm]: 'term',
    [exe_nodejs]: 'node',
};
//# sourceMappingURL=process.js.map