import path from 'upath';
import byte_size from 'byte-size';
import EmojiRegex from 'emoji-regex';
import strip_ansi from 'strip-ansi';
import chalk from 'chalk';
chalk.level = 2;
import { t } from './i18n/instance.js';
export const emoji_regex = EmojiRegex();
export { chalk };
export function to_method_property_descriptors(methods) {
    return Object.fromEntries(Object.entries(methods)
        .map(([name, value]) => ([name, {
            configurable: true,
            writable: true,
            enumerable: false,
            value,
        }])));
}
export function to_getter_property_descriptors(getters) {
    return Object.fromEntries(Object.entries(getters)
        .map(([name, get]) => ([name, {
            configurable: true,
            enumerable: false,
            get,
        }])));
}
export const cjk = '([\u2e80-\u9fff\uf900-\ufaff])';
export const quotes = {
    single: "'",
    double: '"',
    backtick: '`',
};
export const brackets = {
    round: ['(', ')'],
    square: ['[', ']'],
    curly: ['{', '}'],
    pointy: ['<', '>'],
    corner: ['「', '」'],
    fat: ['【', '】'],
    tortoise_shell: ['〔', '〕'],
};
const color_map = Object.fromEntries(['red_', 'green_', 'yellow_', 'blue_', 'magenta_', 'cyan_'].map(color => [color, `${color.slice(0, -1)}Bright`]));
if (!global.my_prototype_defined) {
    // ------------------------------------ String.prototype
    Object.defineProperties(String.prototype, {
        ...to_getter_property_descriptors({
            width() {
                const s = strip_ansi(this.replace(emoji_regex, '  '));
                let width = 0;
                for (let i = 0; i < s.length; i++) {
                    const code = s.codePointAt(i);
                    if ((code <= 0x1f || (code >= 0x7f && code <= 0x9f)) || // ignore control characters
                        code >= 0x300 && code <= 0x36f // ignore combining characters
                    )
                        continue;
                    // surrogates
                    if (code > 0xffff)
                        i++;
                    width += is_codepoint_fullwidth(code) ? 2 : 1;
                }
                return width;
            }
        }),
        // ------------ 文本处理工具方法
        ...to_method_property_descriptors({
            /** 截取字符串不超过 width 显示宽度的部分，并保留颜色
                找到并记录能容纳 字符串 + … 的最后一个字符的位置 i_fitted
                  - 若完整的字符串长度超过 width，返回 slice(0, i_fitted + 1) + …
                  - 否则                          返回 this
             */
            truncate(width) {
                const color_bak = this.startsWith('\u001b') ? this.slice(0, 5) : '';
                const s = strip_ansi(this);
                if (width <= 2)
                    return this.slice(0, width);
                let i_fitted = 0;
                let fitted_width = 0;
                let cur_width = 0;
                for (let i = 0; i < s.length; i++) {
                    const code = s.codePointAt(i);
                    if ((code <= 0x1F || (code >= 0x7F && code <= 0x9F)) || // Ignore control characters
                        code >= 0x300 && code <= 0x36F // Ignore combining characters
                    )
                        continue;
                    // surrogates (codepoint 需要用两个 utf-16 编码单位表示，因此这里跳过第二个编码单位，防止重复计算显示宽度)
                    if (code > 0xFFFF)
                        i++;
                    const w = is_codepoint_fullwidth(code) ? 2 : 1;
                    if (cur_width + w + 2 <= width) {
                        i_fitted = i;
                        fitted_width += w;
                    }
                    cur_width += w;
                    if (cur_width > width) {
                        const i_fitted_next = i_fitted + 1;
                        const t = s.slice(0, i_fitted_next) + ' '.repeat(width - 2 - fitted_width) + '…';
                        return color_bak ? color_bak + t + '\u001b[39m' : t;
                    }
                }
                return this;
            },
            pad(width, { character = ' ', position = 'right' } = {}) {
                const _width = this.width;
                if (_width >= width)
                    return this;
                if (position === 'right')
                    return this + character.repeat((width - _width) / character.width);
                return character.repeat(width - _width) + this;
            },
            limit(width, { character = ' ', position = 'right' } = {}) {
                return this.pad(width, { character, position }).truncate(width);
            },
            to_regexp(preservations, flags = '') {
                const preserved_chars = new Set(preservations);
                const replace_chars = Array.prototype.filter.call('|\\{}()[]^$+*?.-', (c) => !preserved_chars.has(c))
                    .map((c) => c === ']' ? '\\]' : c).join('');
                return new RegExp(this.replace(new RegExp(`[${replace_chars}]`, 'g'), '\\$&'), flags);
            },
            to_bool() {
                return this.length && this !== '0' && this.toLowerCase() !== 'false';
            },
            refmt(pattern, pattern_, preservations = '', flags = '', transformer = (name, value) => value || '', pattern_placeholder = /\{.*?\}/g) {
                // --- 转换 pattern 为 pattern_regx
                let last_end = 0;
                // placeholder matched group indexes
                let $placeholders = {};
                let regx_parts = [];
                function add_part(left, right) {
                    const part = pattern.slice(left, right);
                    if (part)
                        regx_parts.push(part.to_regexp(preservations).source.bracket());
                }
                pattern.replace(pattern_placeholder, ($0, offset) => {
                    add_part(last_end, offset);
                    last_end = offset + $0.length;
                    const placeholder = $0.slice(1, -1);
                    let [placeholder_name, placeholder_pattern] = placeholder.split(':').map(s => s.trim());
                    let optional = false;
                    if (placeholder_name.endsWith('?')) {
                        placeholder_name = placeholder_name.slice(0, -1);
                        optional = true;
                    }
                    $placeholders[placeholder_name] = regx_parts.push(placeholder_pattern ?
                        `${placeholder_pattern.bracket()}${optional ? '?' : ''}`
                        :
                            '(.*?)');
                    return '';
                });
                add_part(last_end);
                // 最后一个 (.*?) 改为贪心匹配，满足 .{suffix} 的需要
                regx_parts = regx_parts.filter(part => part);
                if (regx_parts.last === '(.*?)')
                    regx_parts[regx_parts.length - 1] = '(.*)';
                const pattern_regx = new RegExp(regx_parts.join(''), flags);
                // --- 根据 pattern_regx 去匹配原有字符串，获取匹配结果，生成 placeholders 词典
                const matches = pattern_regx.exec(this);
                if (!matches)
                    return this;
                const placeholders = Object.fromEntries(Object.entries($placeholders)
                    .map(([name, $i]) => [
                    [name, matches[$i]],
                    [`${name}.before`, matches[$i - 1] || ''],
                    [`${name}.after`, matches[$i + 1] || ''],
                ])
                    .flat());
                // --- 转换 pattern_ 为 replacement_str，如果有 transformer 则在遇到 placeholder 时应用
                last_end = 0;
                let replacement_parts = [];
                pattern_.replace(pattern_placeholder, ($0, offset) => {
                    replacement_parts.push(pattern_.slice(last_end, offset));
                    last_end = offset + $0.length;
                    const placeholder_name = $0.slice(1, -1);
                    replacement_parts.push(transformer(placeholder_name, placeholders[placeholder_name], placeholders));
                    return '';
                });
                replacement_parts.push(pattern_.slice(last_end));
                return this.replace(pattern_regx, replacement_parts.join(''));
            },
            find(pattern, preservations = '', flags = '', pattern_placeholder = /\{.*?\}/g) {
                // --- 转换 pattern 为 pattern_regx
                let last_end = 0;
                // placeholder matched group index
                let $placeholders = {};
                let regx_parts = [];
                function add_part(left, right) {
                    const part = pattern.slice(left, right);
                    if (part)
                        regx_parts.push(part.to_regexp(preservations).source.bracket());
                }
                pattern.replace(pattern_placeholder, ($0, offset) => {
                    add_part(last_end, offset);
                    last_end = offset + $0.length;
                    const placeholder = $0.slice(1, -1);
                    let [placeholder_name, placeholder_pattern] = placeholder.split(':').map(s => s.trim());
                    let optional = false;
                    if (placeholder_name.endsWith('?')) {
                        placeholder_name = placeholder_name.slice(0, -1);
                        optional = true;
                    }
                    $placeholders[placeholder_name] = regx_parts.push(placeholder_pattern ?
                        `${placeholder_pattern.bracket()}${optional ? '?' : ''}`
                        :
                            '(.*?)');
                    return '';
                });
                add_part(last_end);
                // 最后一个 (.*?) 改为贪心匹配，满足 .{suffix} 的需要
                regx_parts = regx_parts.filter(part => part);
                if (regx_parts[regx_parts.length - 1] === '(.*?)')
                    regx_parts[regx_parts.length - 1] = '(.*)';
                const pattern_regx = new RegExp(regx_parts.join(''), flags);
                // --- 根据 pattern_regx 去匹配原有字符串，获取匹配结果，生成 placeholders 词典
                const matches = pattern_regx.exec(this);
                if (!matches)
                    return {};
                return Object.fromEntries(Object.entries($placeholders)
                    .map(([name, $i]) => [name, matches[$i] || '']));
            },
            quote(type = 'single') {
                if (type === 'psh')
                    return `& ${this.quote()}`;
                return this.surround(quotes[type]);
            },
            bracket(shape = 'round') {
                return this.surround(...brackets[shape]);
            },
            surround(left, right) {
                return left + this + (right || left);
            },
            surround_tag(tag_name) {
                return '<' + tag_name + '>' + this + '</' + tag_name + '>';
            },
            to_lf() {
                return this.replace(/\r\n/g, '\n');
            },
            to_crlf() {
                return this.replace(/\n/g, '\r\n');
            },
            rm(pattern, flags = 'g') {
                if (typeof pattern === 'string')
                    pattern = new RegExp(pattern, flags);
                return this.replace(pattern, '');
            },
            split_lines(delimiter = /\r?\n/) {
                let lines = this.split(delimiter);
                if (lines.last === '')
                    lines.pop();
                return lines;
            },
            split_indent() {
                let i = 0;
                let indent = 0;
                for (; i < this.length; i++)
                    if (this[i] === ' ')
                        indent++;
                    else if (this[i] === '\t')
                        indent += 4;
                    else
                        break;
                return {
                    indent,
                    text: this.slice(i)
                };
            },
            trim_doc_comment() {
                return `/** ${this.slice(3, -2).replace(/\s*\*\s*/g, '  ').replace(/@(param|params|return) \{.*?\}\s*/g, '').trim()} */`;
            },
            to_base64() {
                return Buffer.from(this).toString('base64');
            },
            decode_base64(buffer = false) {
                const buf = Buffer.from(this, 'base64');
                if (buffer)
                    return buf;
                return buf.toString();
            },
            strip_ansi() {
                return strip_ansi(this);
            },
            space() {
                if (!this)
                    return this;
                let text_;
                text_ = this
                    .replace(new RegExp(cjk + '([\'"])', 'g'), '$1 $2')
                    .replace(new RegExp('([\'"])' + cjk, 'g'), '$1 $2')
                    .replace(/(["']+)\s*(.+?)\s*(["']+)/g, '$1$2$3')
                    .replace(new RegExp(cjk + '([\\+\\-\\*\\/=&\\\\\\|<>])([A-Za-z0-9])', 'g'), '$1 $2 $3')
                    .replace(new RegExp('([A-Za-z0-9])([\\+\\-\\*\\/=&\\\\\\|<>])' + cjk, 'g'), '$1 $2 $3');
                const text_bak = text_;
                text_ = text_.replace(new RegExp(cjk + '([\\(\\[\\{<\u201c]+(.*?)[\\)\\]\\}>\u201d]+)' + cjk, 'g'), '$1 $2 $4');
                if (text_ === text_bak)
                    text_ = text_
                        .replace(new RegExp(cjk + '([\\(\\[\\{<\u201c>])', 'g'), '$1 $2')
                        .replace(new RegExp('([\\)\\]\\}>\u201d<])' + cjk, 'g'), '$1 $2');
                return text_
                    .replace(/([\(\[\{<\u201c]+)(\s*)(.+?)(\s*)([\)\]\}>\u201d]+)/g, '$1$3$5')
                    .replace(new RegExp(cjk + '([~!;:,\\.\\?\u2026])([A-Za-z0-9])', 'g'), '$1$2 $3')
                    .replace(new RegExp(cjk + '([A-Za-z0-9`\\$%\\^&\\*\\-=\\+\\\\\\|\\/@\u00a1-\u00ff\u2022\u2027\u2150-\u218f])', 'g'), '$1 $2')
                    .replace(new RegExp('([A-Za-z0-9`\\$%\\^&\\*\\-=\\+\\\\\\|\\/@\u00a1-\u00ff\u2022\u2027\u2150-\u218f])' + cjk, 'g'), '$1 $2');
            }
        }),
        // ------------ chalk colors
        ...Object.fromEntries([
            'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'grey',
            'red_', 'green_', 'yellow_', 'blue_', 'magenta_', 'cyan_',
            'underline',
        ].map(color => ([color, {
                configurable: true,
                get() {
                    return chalk[color_map[color] || color](this);
                }
            }]))),
        // ------------ 文件路径操作
        ...to_getter_property_descriptors({
            fdir() {
                const dir = path.dirname(this);
                return dir.endsWith('/') ? dir : `${dir}/`;
            },
            fname() {
                return `${path.basename(this)}${this.endsWith('/') ? '/' : ''}`;
            },
            fext() {
                return path.extname(this);
            },
        }),
        ...to_method_property_descriptors({
            to_slash() {
                if (!this)
                    return this;
                return path.normalizeSafe(this);
            },
            to_backslash() {
                return this.replaceAll('/', '\\');
            },
        })
    });
    // ------------------------------------ Date.prototype
    Object.defineProperties(Date.prototype, to_method_property_descriptors({
        to_str(ms) {
            const [ampm, hour] = (() => {
                let hour = this.getHours();
                if (hour <= 6)
                    return [t('凌晨'), hour];
                if (hour <= 8)
                    return [t('清晨'), hour];
                if (hour <= 9)
                    return [t('早上'), hour];
                if (hour <= 10)
                    return [t('上午'), hour];
                if (hour <= 12)
                    return [t('中午'), hour];
                hour -= 12;
                if (hour <= 5)
                    return [t('下午'), hour];
                if (hour <= 10)
                    return [t('晚上'), hour];
                return [t('深夜'), hour];
            })();
            const zero_padding = { character: '0', position: 'left' };
            return '' +
                // year.month.date
                this.getFullYear() + '.' +
                String(this.getMonth() + 1).pad(2, zero_padding) + '.' +
                String(this.getDate()).pad(2, zero_padding) + ' ' +
                // 上午
                ampm + ' ' +
                // 10:03:02
                String(hour).pad(2, zero_padding) + ':' +
                String(this.getMinutes()).pad(2, zero_padding) + ':' +
                String(this.getSeconds()).pad(2, zero_padding) +
                (ms ?
                    '.' + String(this.getMilliseconds()).pad(3, zero_padding)
                    :
                        '');
        },
        to_date_str() {
            return this.to_str().split(' ')[0];
        },
        to_time_str(ms) {
            const [, ampm, time] = this.to_str(ms).split(' ');
            return `${ampm} ${time}`;
        },
    }));
    // ------------------------------------ Number.prototype
    Object.defineProperties(Number.prototype, to_method_property_descriptors({
        to_fsize_str(units = 'iec') {
            const { value, unit } = byte_size(this, { units });
            return `${value} ${unit.replace('i', '')}`;
        },
        to_bin_str() {
            return `0b${this.toString(2)}`;
        },
        to_hex_str(length) {
            const s = this.toString(16);
            // 长度自动对齐到 4 的倍数
            if (!length)
                length = Math.ceil(s.length / 4) * 4;
            return `0x${'0'.repeat(length - s.length)}${s}`;
        },
        to_oct_str() {
            return `0o${this.toString(8)}`;
        },
    }));
    // ------------------------------------ Array.prototype
    Object.defineProperties(Array.prototype, {
        ...to_getter_property_descriptors({
            last() {
                return this[this.length - 1];
            }
        }),
        // --- 文本处理工具方法
        ...to_method_property_descriptors({
            log(limit = 10000) {
                const text = this.join('\n') + '\n';
                if (limit === -1 || this.length <= limit)
                    console.log(text);
                else if (limit > 0)
                    console.log(text.slice(0, limit) + '\n...'.blue);
                else
                    console.log('...\n'.blue + text.slice(limit));
            },
            trim_lines({ trim_line = true, rm_empty_lines = true, rm_last_empty_lines = false } = {}) {
                if (!this.length)
                    return this;
                let lines = this;
                if (trim_line)
                    lines = lines.map(line => line.trim());
                if (rm_empty_lines)
                    return lines.filter(line => line);
                if (rm_last_empty_lines) {
                    lines.reverse();
                    const i_not_empty = lines.findIndex(line => line);
                    if (i_not_empty !== -1)
                        lines = lines.slice(i_not_empty);
                    lines.reverse();
                    return lines;
                }
                return lines;
            },
            trim_license() {
                const i = this.indexOf('/*');
                const j = this.indexOf('*/');
                if (i === 0 && this[i + 1].includes('License'))
                    return this.slice(j + 1);
                else
                    return this;
            },
            split_indents() {
                return this.map(line => line.split_indent());
            },
            indent(width, character = ' ') {
                return this.map(line => character.repeat(width) + line);
            },
            indent2to4() {
                return this.split_indents()
                    .map(line => ' '.repeat(line.indent * 2) + line.text);
            },
            join_lines(append = true) {
                return `${this.join('\n')}${append ? '\n' : ''}`;
            }
        })
    });
    Object.defineProperties(BigInt.prototype, to_method_property_descriptors({
        toJSON() {
            return this.toString();
        }
    }));
    Object.defineProperties(Error.prototype, to_method_property_descriptors({
        toJSON() {
            return Object.fromEntries(Object.getOwnPropertyNames(this)
                .map(name => [name, this[name]]));
        }
    }));
}
export function to_json(obj, replacer) {
    return JSON.stringify(obj, replacer, 4) + '\n';
}
export function to_json_safely(obj, replacer) {
    return to_json(obj, replacer)
        .replace(/\u2028/g, '\\u2028')
        .replace(/\u2029/g, '\\u2029')
        .replace(/<\/script>/g, '<\\/script>');
}
export function is_codepoint_fullwidth(codepoint) {
    // code points are derived from:
    // http://www.unix.org/Public/UNIDATA/EastAsianWidth.txt
    return (!Number.isNaN(codepoint) &&
        codepoint >= 0x1100 &&
        (codepoint <= 0x115f || // hangul jamo
            codepoint === 0x201c || codepoint === 0x201d || // 
            codepoint === 0x2026 || // …
            codepoint === 0x203b || // ※
            // arrows
            (0x2190 <= codepoint && codepoint <= 0x21FF) ||
            codepoint === 0x2329 || // left-pointing angle bracket
            codepoint === 0x232a || // right-pointing angle bracket
            // ①
            (0x2460 <= codepoint && codepoint <= 0x24ff) ||
            // box drawing
            (0x2500 <= codepoint && codepoint <= 0x257f) ||
            // shapes, symbols, …
            (0x2580 <= codepoint && codepoint <= 0x2bef) ||
            // cjk radicals supplement .. enclosed cjk letters and months
            (0x2e80 <= codepoint && codepoint <= 0x3247 && codepoint !== 0x303f) ||
            // enclosed cjk letters and months .. cjk unified ideographs extension a
            (0x3250 <= codepoint && codepoint <= 0x4dbf) ||
            // cjk unified ideographs .. yi radicals
            (0x4E00 <= codepoint && codepoint <= 0xA4C6) ||
            // hangul jamo extended-a
            (0xa960 <= codepoint && codepoint <= 0xa97c) ||
            // hangul syllables
            (0xac00 <= codepoint && codepoint <= 0xd7a3) ||
            // cjk compatibility ideographs
            (0xf900 <= codepoint && codepoint <= 0xfaff) ||
            // vertical forms
            (0xfe10 <= codepoint && codepoint <= 0xfe19) ||
            // cjk compatibility forms .. small form variants
            (0xfe30 <= codepoint && codepoint <= 0xfe6b) ||
            // halfwidth and fullwidth forms
            (0xff01 <= codepoint && codepoint <= 0xff60) ||
            (0xffe0 <= codepoint && codepoint <= 0xffe6) ||
            // kana supplement
            (0x1b000 <= codepoint && codepoint <= 0x1b001) ||
            // enclosed ideographic supplement
            (0x1f200 <= codepoint && codepoint <= 0x1f251) ||
            // cjk unified ideographs extension b .. tertiary ideographic plane
            (0x20000 <= codepoint && codepoint <= 0x3fffd)));
}
//# sourceMappingURL=prototype.js.map