import nvm from 'vm';
import repl from 'repl';
import process from 'process';
import { fileURLToPath } from 'url';
import path from 'upath';
import ts from 'typescript';
const { factory, createPrinter, SyntaxKind, NodeFlags, ModifierFlags, isIdentifier, isNamedImports, isImportDeclaration, isAwaitExpression, isExpressionStatement, isVariableStatement, isNamespaceImport, isClassDeclaration, isCallExpression, isReturnStatement, isBinaryExpression, isFunctionDeclaration, } = ts;
import { t } from './i18n/instance.js';
import './prototype.js';
import { log_line, delay, inspect, set_inspect_options, Timer, delta2str } from './utils.js';
import { fread, fwrite } from './file.js';
import { Remote } from './net.js';
set_inspect_options();
let server;
let inspection_limit = 10000;
let printing_compiled_js = false;
/** 谨慎使用，webpack 打包后可能会变成 /d:/1/mod/node_modules/xshell/ 这样的编译期路径 */
export const fpd_root = path.dirname(fileURLToPath(import.meta.url)) + '/';
export function set_inspection_limit(limit = 10000) {
    if (limit === -1)
        limit = 50 * 10 ** 4;
    inspection_limit = limit;
}
export function set_printing_compiled_js(enabled) {
    printing_compiled_js = enabled;
}
// ------------------------------------ Code Compilers, Transformers
function resolve_kind(kind) {
    return kind === SyntaxKind.VariableStatement ? 'VariableStatement' : SyntaxKind[kind];
}
function print_ast(nodes, sourceFile) {
    function print_node(node) {
        console.log(`${resolve_kind(node.kind)}: ${node.getText(sourceFile)}`);
    }
    if (Array.isArray(nodes))
        nodes.forEach(node => { print_node(node); });
    else
        print_node(nodes);
}
export function parse_code(code) {
    return ts.createSourceFile('repl.ts', code, ts.ScriptTarget.ESNext);
}
export function generate_code(stmts) {
    return createPrinter({ omitTrailingSemicolon: true, removeComments: false, newLine: ts.NewLineKind.LineFeed })
        .printFile(factory.updateSourceFile(ts.createSourceFile('output.ts', '', ts.ScriptTarget.ESNext), stmts));
}
function trans_import_2_require(import_decl) {
    if (!isImportDeclaration(import_decl))
        return import_decl;
    const { importClause: import_clause, moduleSpecifier: module_specifier } = import_decl;
    // require('module_specifier')
    const require_module_stmt = factory.createCallExpression(factory.createIdentifier('require'), [], [module_specifier]);
    // aaa, bbb as ccc, default as ddd, ...
    // (import_clause.namedBindings as NamedImports).elements
    return [
        // import mod from 'mod'
        // global.mod = __importDefault(require("mod")).default
        ...import_clause.name ? [
            factory.createExpressionStatement(factory.createAssignment(factory.createPropertyAccessExpression(factory.createIdentifier('global'), import_clause.name), factory.createPropertyAccessExpression(factory.createCallExpression(factory.createIdentifier('__importDefault'), [], [require_module_stmt]), 'default')))
        ] : [],
        // import * as mod from 'mod'
        // global.mod = __importStar(require("mod"))
        ...import_clause.namedBindings && isNamespaceImport(import_clause.namedBindings) ? [
            factory.createExpressionStatement(factory.createAssignment(factory.createPropertyAccessExpression(factory.createIdentifier('global'), import_clause.namedBindings.name), factory.createCallExpression(factory.createIdentifier('__importStar'), [], [require_module_stmt])))
        ] : [],
        // import { element2 as element2_, element3 as element3_, other, default as mod2 } from 'mod2'
        ...import_clause.namedBindings && isNamedImports(import_clause.namedBindings) ? [
            // var { element2: element2_, element3 } = __importDefault(require("mod"))
            factory.createVariableStatement([], [factory.createVariableDeclaration(factory.createObjectBindingPattern(
                // ImportSpecifier[] -> ObjectBindingPattern.BindingElement[]
                import_clause.namedBindings.elements.map(import_specifier => factory.createBindingElement(undefined, import_specifier.propertyName, import_specifier.name, import_specifier.propertyName?.text === 'default' ? require_module_stmt : undefined))), undefined, undefined, require_module_stmt)]),
            // Object.assign(global, { element2_, element3 })
            factory.createExpressionStatement(factory.createCallExpression(factory.createPropertyAccessExpression(factory.createIdentifier('Object'), 'assign'), [], 
            // ImportSpecifier[] -> ObjectBindingPattern.BindingElement[]
            [factory.createIdentifier('global'), factory.createObjectLiteralExpression(import_clause.namedBindings.elements.map(import_specifier => factory.createShorthandPropertyAssignment(import_specifier.name)), false)])),
            factory.createExpressionStatement(import_clause.namedBindings.elements.length === 1 ?
                // (a)
                factory.createParenthesizedExpression(import_clause.namedBindings.elements[0].name)
                : // ({ a: xxx, b: xxx })
                    factory.createObjectLiteralExpression(import_clause.namedBindings.elements.map(import_specifier => factory.createShorthandPropertyAssignment(import_specifier.name)), false))
        ] : [],
    ];
}
/** export function foo () { } */
function trans_export_stmt(stmt) {
    if (ts.isExportAssignment(stmt))
        return factory.createExpressionStatement(stmt.expression);
    function is_export_modifier(modifier) {
        return modifier.flags & ModifierFlags.Export || modifier.flags & ModifierFlags.ExportDefault || modifier.kind === SyntaxKind.ExportKeyword;
    }
    // @ts-ignore
    if (stmt.modifiers?.some(modifier => is_export_modifier(modifier))) {
        // @ts-ignore
        stmt.modifiers = factory.createNodeArray(
        // @ts-ignore
        stmt.modifiers.filter(modifier => !is_export_modifier(modifier)));
        return stmt;
    }
    return stmt;
}
/** const a = 123, b = 234, { c, d: e } = obj */
function trans_variable_decl_2_var(var_decl) {
    if (!isVariableStatement(var_decl))
        return var_decl;
    // @ts-ignore
    var_decl.declarationList.flags = NodeFlags.None;
    return var_decl;
}
function trans_class_decl_2_expr(class_decl) {
    if (!isClassDeclaration(class_decl))
        return class_decl;
    // export class C { } 这样的语句已经在 trans_export_stmt 中被去掉了 export modifier
    return factory.createVariableStatement([], factory.createVariableDeclarationList([
        factory.createVariableDeclaration(class_decl.name, undefined, undefined, factory.createClassExpression(undefined, class_decl.name, undefined, undefined, class_decl.members))
    ]));
}
function get_expr_of_stmt(statement) {
    if (isExpressionStatement(statement))
        return statement;
    if (isVariableStatement(statement)) {
        const { declarations } = statement.declarationList;
        return factory.createExpressionStatement(declarations.length === 1 ?
            (() => {
                const { name } = declarations[0];
                if (isIdentifier(name)) // const a = c
                    return name;
                else // const { a, b } = c
                    return factory.createObjectLiteralExpression(name.elements
                        .filter(({ name }) => name && isIdentifier(name))
                        .map(({ name }) => factory.createShorthandPropertyAssignment(name)));
            })()
            :
                factory.createObjectLiteralExpression(declarations.map(var_decl => factory.createShorthandPropertyAssignment(var_decl.name)), true));
    }
    if (isFunctionDeclaration(statement))
        return factory.createExpressionStatement(statement.name);
    return statement;
}
function trans_return_2_expr(stmt) {
    if (!isReturnStatement(stmt))
        return stmt;
    return factory.createExpressionStatement(stmt.expression);
}
function assign_var_decl(var_decl, destination = 'global') {
    if (!isVariableStatement(var_decl))
        return [var_decl];
    // VariableDeclaration[] -> ObjectLiteral[]
    // const a = 123, b = 234, { c, d: e } = obj
    const obj_literal = factory.createObjectLiteralExpression(var_decl.declarationList.declarations.map(declaration => isIdentifier(declaration.name) ?
        [factory.createShorthandPropertyAssignment(declaration.name)]
        : // ObjectBindingPattern
            declaration.name.elements.map((element) => factory.createShorthandPropertyAssignment(element.name))).flat(), true);
    return [
        var_decl,
        factory.createExpressionStatement(factory.createCallExpression(factory.createPropertyAccessExpression(factory.createIdentifier('Object'), 'assign'), [], [factory.createIdentifier(destination), obj_literal])),
        factory.createExpressionStatement(obj_literal)
    ];
}
function return_last_expr(statements) {
    if (!statements.length)
        return statements;
    const last_expr = get_expr_of_stmt(statements.last);
    if (isExpressionStatement(last_expr))
        return [...statements.slice(0, -1), factory.createReturnStatement(last_expr.expression)];
    else
        return statements;
}
function wrap_await_stmt(statements, code) {
    function wrap(stmts) {
        return [
            factory.createExpressionStatement(factory.createCallExpression(factory.createFunctionExpression([factory.createModifier(SyntaxKind.AsyncKeyword)], undefined, 'async_wrapper', [], [], undefined, factory.createBlock(return_last_expr(stmts.map(statement => assign_var_decl(statement)).flat()), true)), [], [])),
        ];
    }
    if (!code.includes('await'))
        return statements;
    if (code.includes('await') && !code.includes('async '))
        return wrap(statements);
    if (statements.some(stmt => (isExpressionStatement(stmt) && isAwaitExpression(stmt.expression) ||
        isVariableStatement(stmt) && stmt.declarationList.declarations.some(var_decl => var_decl.initializer && isAwaitExpression(var_decl.initializer)) ||
        isExpressionStatement(stmt) && isCallExpression(stmt.expression) && stmt.expression.arguments.some(expr => isAwaitExpression(expr)) ||
        isExpressionStatement(stmt) && isBinaryExpression(stmt.expression) && isAwaitExpression(stmt.expression.right))))
        return wrap(statements);
    return statements;
}
export async function compile_ts({ fp, code, print = printing_compiled_js, save = false, } = {}) {
    if (!code && fp)
        code = await fread(fp);
    const source_file = parse_code(code);
    let statements = [...source_file.statements];
    statements = statements.map(trans_import_2_require).flat();
    statements = statements.map(trans_export_stmt);
    statements = statements.map(trans_class_decl_2_expr);
    statements = wrap_await_stmt(statements, code);
    statements = statements.map(trans_variable_decl_2_var);
    statements = statements.map(trans_return_2_expr);
    if (statements.length) {
        const last_stmt = statements[statements.length - 1];
        const last_expr = get_expr_of_stmt(last_stmt);
        if (last_expr !== last_stmt)
            statements = [...statements, last_expr];
    }
    code = generate_code(statements);
    let { diagnostics, outputText: output_text } = ts.transpileModule(code, { compilerOptions: ts_options_commonjs_repl });
    if (diagnostics.length) {
        console.log(diagnostics.join('\n\n\n'));
        log_line();
    }
    if (print) {
        console.log(output_text);
        log_line();
    }
    if (fp && save)
        await fwrite(fp.replace(/\.ts(x?)$/, '.js$1'), output_text);
    return output_text;
}
/** tsconfig.compilerOptions */
export let ts_options;
export let ts_options_commonjs;
export let ts_options_commonjs_repl;
export async function load_tsconfig() {
    const fp = `${fpd_root}tsconfig.json`;
    ({ config: { compilerOptions: ts_options } } = ts.parseConfigFileTextToJson(fp, await fread(fp, { print: false })));
    ts_options_commonjs = {
        ...ts_options,
        module: 'CommonJS',
        incremental: false,
        // determine CommonJS module require method
        esModuleInterop: true,
    };
    ts_options_commonjs_repl = {
        ...ts_options,
        module: 'CommonJS',
        esModuleInterop: true,
        // nvm.runInThisContext doesn't support inline source map
        sourceMap: false,
    };
}
let eval_id = 0;
export async function eval_ts(code) {
    return nvm.runInThisContext(await compile_ts({ code }), `repl/${eval_id++}.ts`);
}
// ------------------------------------ repl
export async function repl_ts(code) {
    console.log(); // 离之前的输出有一个空行
    const timer = new Timer();
    const __ = global.__ = await eval_ts(code);
    timer.stop();
    console.log(inspect(__, { limit: inspection_limit }));
    if (timer.get() >= 500)
        timer.print();
    return __;
}
export async function start_repl() {
    // ------------ 加载库
    console.log(t('xshell 开始启动').yellow);
    // --- prevent from exiting
    process.on('uncaughtException', error => {
        console.error(error);
    });
    // --- start nodejs repl
    repl.start({
        prompt: '',
        replMode: repl.REPL_MODE_SLOPPY,
        useGlobal: true,
        useColors: true,
        terminal: true,
    });
    console.log(t('nodejs.repl 启动成功'));
    process.title = 'xshell';
    await Promise.all([
        pollute_global(),
        load_tsconfig(),
        (async () => {
            // --- http server
            let { Server } = await import('./server.js');
            server = new Server({
                port: 8421,
                remote: new Remote({
                    funcs: {
                        echo({ data: [data] }) {
                            console.log('echo:', data);
                            return [data];
                        },
                        async repl_ts({ data: [code] }) {
                            return [await repl_ts(code)];
                        }
                    }
                })
            });
            await server.start();
            console.log(t('server 启动成功'));
        })(),
    ]);
    console.log(('-'.repeat(30) + '\n' +
        t('xshell 启动成功，用时 ') + delta2str(new Date().getTime() - global.started_at.getTime()) + '\n' +
        t('正在监听: http://localhost:8421\n') +
        '-'.repeat(30)).green);
}
export async function stop() {
    console.log(t('xshell 正在退出').red);
    server.stop();
}
export async function exit() {
    await stop();
    await delay(100);
    process.exit();
}
export async function pollute_global() {
    Object.assign(global, {
        __importDefault: (mod) => ((mod?.__esModule) ? mod : { default: mod }),
        __importStar: (mod) => {
            if (mod?.__esModule)
                return mod;
            let result = {};
            // eslint-disable-next-line eqeqeq
            if (mod != null)
                for (let k in mod)
                    if (Object.hasOwnProperty.call(mod, k))
                        result[k] = mod[k];
            result.default = mod;
            return result;
        },
        exports: global
    });
    await Promise.all([
        pollute_module_default_export('upath', 'path'),
        pollute_module_default_export('lodash/omit.js', 'omit'),
        pollute_module_default_export('lodash/sortBy.js', 'sort_by'),
        pollute_module_default_export('lodash/groupBy.js', 'group_by'),
        pollute_module_default_export('qs', 'qs'),
        pollute_module_exports('./prototype.js'),
        pollute_module_exports('./utils.js'),
        pollute_module_exports('./process.js'),
        pollute_module_exports('./file.js'),
        pollute_module_exports('./net.js'),
        pollute_module_exports('./server.js'),
        pollute_module_exports('./repl.js'),
    ]);
    console.log(t('所有模块加载成功'));
}
export async function pollute_module_exports(fp_mod) {
    Object.assign(global, await import(fp_mod));
}
export async function pollute_module_default_export(fp_mod, name) {
    global[name] = (await import(fp_mod)).default;
}
//# sourceMappingURL=repl.js.map