import { createServer as http_create_server } from 'http';
import zlib from 'zlib';
import { createReadStream } from 'fs';
import { Readable } from 'stream';
import util from 'util';
// --- 3rd party
import upath from 'upath';
import qs from 'qs';
import resolve_safely from 'resolve-path';
import { WebSocketServer } from 'ws';
// --- koa & koa middleware
import { default as Koa } from 'koa';
import KoaCors from '@koa/cors';
import KoaCompress from 'koa-compress';
import { userAgent as KoaUserAgent } from 'koa-useragent';
// --- my libs
import { request as _request } from './net.js';
import { stream_to_buffer, inspect, output_width, assert } from './utils.js';
import { fstat } from './file.js';
// ------------ my server
export class Server {
    /** proxy 时需要丢弃的 resposne headers */
    static drop_response_headers = new Set([
        // : 开头的 key
        'connection',
        'content-encoding',
        'keep-alive',
        'transfer-encoding',
        'x-powered-by'
    ]);
    app;
    handler;
    port;
    server_http;
    server_ws;
    remote;
    colors = util.inspect.defaultOptions.colors;
    constructor({ port, remote } = {}) {
        if (port !== undefined)
            this.port = port;
        if (remote)
            this.remote = remote;
    }
    /** start http server and listen */
    async start() {
        // --- init koa app
        let app = new Koa();
        app.on('error', (error, ctx) => {
            console.error(error);
            console.log(ctx);
        });
        app.use(this.entry.bind(this));
        app.use(KoaCompress({
            br: {
                chunkSize: 64 * 1024,
                // https://nodejs.org/api/zlib.html#zlib_class_brotlioptions
                params: {
                    [zlib.constants.BROTLI_PARAM_MODE]: zlib.constants.BROTLI_MODE_TEXT,
                    [zlib.constants.BROTLI_PARAM_QUALITY]: 3 // 默认为 11 (最大压缩)，会导致 news/get 生成的 14 mb 的 json 压缩时长高达 24s
                },
            },
            threshold: 512
        }));
        app.use(KoaCors({ credentials: true }));
        app.use(KoaUserAgent);
        app.use(this._router.bind(this));
        this.app = app;
        this.handler = this.app.callback();
        this.server_http = http_create_server(this.handler);
        // websocket rpc
        if (this.remote) {
            this.server_ws = new WebSocketServer({
                noServer: true,
                skipUTF8Validation: true,
                perMessageDeflate: true
            });
            this.server_ws.on('connection', (ws, request) => {
                ws.addEventListener('message', ({ data }) => {
                    this.remote.handle(new Uint8Array(data), ws);
                });
            });
            this.server_http.on('upgrade', (request, socket, head) => {
                // url 只有路径部分
                const { url, headers: { host = '', 'user-agent': ua }, } = request;
                const ip = request.socket.remoteAddress.replace(/^::ffff:/, '');
                console.log(`${new Date().to_time_str()}    ${(ip || '').pad(40)}  ${(ua || '').limit(40)}  ${'websocket'.pad(10).magenta}    ${'connect'.pad(10).magenta}${host.pad(20)}  ${url.pad(60).yellow}`);
                switch (url) {
                    case '/':
                        this.server_ws.handleUpgrade(request, socket, head, ws => {
                            ws.binaryType = 'arraybuffer';
                            this.server_ws.emit('connection', ws, request);
                        });
                        return;
                    default:
                        console.log(`${' '.repeat(13)}    connect 404: ${url}`.red);
                        socket.destroy();
                }
            });
        }
        await new Promise(resolve => {
            this.server_http.listen(this.port, resolve);
        });
    }
    stop() {
        this.server_http.close();
    }
    async entry(ctx, next) {
        let { response } = ctx;
        await this.parse(ctx);
        // ------------ next
        try {
            await next();
        }
        catch (error) {
            if (error.status !== 404)
                console.error(error);
            response.status = error.status || 500;
            response.body = inspect(error, { colors: false });
            response.type = 'text/plain';
        }
    }
    /**
        parse req.body to request.body
        process request.ip
    */
    async parse(ctx) {
        let { req, request } = ctx;
        const buf = await stream_to_buffer(req);
        if (buf.length)
            req.body = buf;
        // --- parse request.ip
        request.ip = (request.headers['x-real-ip'] || request.ip).replace(/^::ffff:/, '');
        // --- parse body
        if (!req.body)
            return;
        if (ctx.is('application/json') || ctx.is('text/plain'))
            request.body = JSON.parse(req.body.toString());
        else if (ctx.is('application/x-www-form-urlencoded'))
            request.body = qs.parse(req.body.toString());
        else if (ctx.is('multipart/form-data'))
            throw new Error('multipart/form-data is not supported');
        else
            request.body = req.body;
    }
    async _router(ctx, next) {
        let { request } = ctx;
        const _path = request._path = decodeURIComponent(request.path);
        Object.defineProperty(request, 'path', {
            value: _path,
            configurable: true,
            enumerable: true,
            writable: true
        });
        // ------------ log
        this.logger(ctx);
        // ------------ repl_router hook
        if (await global.repl_router?.(ctx))
            return;
        if (await this.router(ctx))
            return;
        await next?.();
    }
    async router(ctx) {
        return false;
    }
    logger(ctx) {
        const { colors } = this;
        const { request } = ctx;
        const { query, body, path, _path, protocol, host, req: { httpVersion: http_version }, ip, } = request;
        let { method } = request;
        const ua = ctx.userAgent;
        let s = '';
        // --- time
        s += `${new Date().to_time_str()}    `;
        // --- ip
        s += (ip || '').pad(40) + '  ';
        // --- ua
        s += (() => {
            let t = '';
            if (ua.isMobile)
                t += 'mobile';
            if (ua.isDesktop)
                t += 'desktop';
            if (ua.isBot)
                t += `${t ? ' ' : ''}${colors ? 'robot'.blue : 'robot'}`;
            if (ua.platform !== 'unknown' && !ua.os.startsWith('Windows'))
                t += '／' + ua.platform.toLowerCase().replace('apple mac', 'mac');
            if (ua.os !== 'unknown' && ua.platform !== 'Android')
                t += '／' + ua.os.toLowerCase();
            if (ua.browser !== 'unknown')
                t += '／' + ua.browser.toLowerCase();
            if (ua.isWechat)
                t += '／weixin';
            if (ua.version !== 'unknown')
                t += '／' + ua.version.split('.').slice(0, 2).join('.');
            return t;
        })().pad(40) + '  ';
        // --- https／2.0
        s += `${`${protocol.pad(5)}／${http_version}`.pad(10)}    `;
        // --- method
        method = method.toLowerCase();
        s += (method === 'get' || !colors) ? method.pad(10) : method.pad(10).yellow;
        // --- host
        s += `${host.pad(20)}  `;
        // --- path
        s += (() => {
            if (path.toLowerCase() !== _path.toLowerCase())
                return `${_path.blue} → ${path}`;
            if (!path.includes('.'))
                return colors ? path.yellow : path;
            return path;
        })();
        // --- query
        if (Object.keys(query).length) {
            let t = inspect(query, { compact: true })
                .replace('[Object: null prototype] ', '');
            if (t.endsWith('\n'))
                t = t.slice(0, -1);
            s += (s + t).width > output_width ? '\n' : '    ';
            s += t;
        }
        // --- body
        if (body && Object.keys(body).length)
            s += '\n' + inspect(body).replace('[Object: null prototype] ', '');
        // --- print log
        console.log(s);
    }
    async proxy(ctx, url, headers_ = {}) {
        const { request: { method, headers, query, body } } = ctx;
        let { response } = ctx;
        try {
            const response_ = await _request(url, {
                method: method,
                queries: query,
                body,
                headers: {
                    ...headers,
                    ...headers_
                },
                raw: true,
            });
            response.status = response_.status;
            response.set(Server.filter_response_headers(response_.headers));
            response.body = Readable.fromWeb(response_.body);
        }
        catch (error) {
            if (error.response?.status !== 404)
                console.log(error);
            if (error.response) {
                const { status, headers, text } = error.response;
                if (status === 404)
                    console.log(method, '404:', url);
                else
                    console.log(error);
                response.status = status;
                response.set(Server.filter_response_headers(headers));
                response.body = text;
            }
            else {
                console.log(error);
                response.status = 500;
                response.body = inspect(error, { colors: false });
            }
        }
    }
    static filter_response_headers(headers) {
        let headers_ = {};
        for (const [key, value] of headers)
            if (!key.startsWith(':') && !this.drop_response_headers.has(key))
                headers_[key] = value;
        return headers_;
    }
    async try_send(ctx, fp, { root, log_404, }) {
        const { request: { _path, path, method }, response, } = ctx;
        if (!(typeof response.body === 'undefined') || response.status !== 404)
            return true;
        if (method !== 'HEAD' && method !== 'GET')
            return false;
        function _log_404() {
            let s = `${' '.repeat(13)}    ${method.toLowerCase()} 404: ${path}`;
            if (_path !== path)
                s += ` ${_path.bracket()}`;
            console.log(s.red);
        }
        try {
            await this.fsend(ctx, fp, { root });
            return true;
        }
        catch (error) {
            if (error.status !== 404)
                throw error;
            if (log_404)
                _log_404();
            return false;
        }
    }
    /** send file at `path` with the  given `options` to the koa `ctx`. */
    async fsend(ctx, path, { root, absolute } = {}) {
        const { request } = ctx;
        let { response } = ctx;
        if (!absolute && !root)
            throw new Error('fsend with `!absolute && !root`');
        if (absolute)
            path = upath.resolve(path);
        else {
            if (path.startsWith(root))
                path = path.slice(root.length);
            if (path.startsWith('/'))
                path = path.slice(1);
            try {
                path = upath.normalize(resolve_safely(root, path));
            }
            catch (error) {
                error.message += `, path: ${path}`;
                throw error;
            }
        }
        // stat
        let stats;
        try {
            stats = await fstat(path);
        }
        catch (error) {
            if (['ENOENT', 'ENAMETOOLONG', 'ENOTDIR'].includes(error.code)) {
                error.status = 404;
                throw error;
            }
            error.status = 500;
            error.message = `fs.stat 出错: ${error.message}`;
            throw error;
        }
        // size limit
        // if (stats.size >= 100 * 2**20) {
        //     let error = new Error('body.length >= 100 mb')
        //     ;(error as any).status = 500
        //     throw error
        // }
        // https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-Ranges
        // advertise server support of partial requests
        response.set('accept-ranges', 'bytes');
        if (!response.get('cache-control'))
            response.set('cache-control', 'max-age=0, must-revalidate');
        if (!response.get('last-modified'))
            response.set('last-modified', stats.mtime.toUTCString());
        const fext = path.fext;
        if (!response.type)
            response.type = fext;
        if (fext === '.pdf')
            response.set('content-disposition', `attachment; filename="${encodeURIComponent(path.fname)}"`);
        if (request.fresh) {
            response.status = 304;
            // 以上会自动设置 response.body = null
            return path;
        }
        if (request.headers.range)
            try {
                const range_header = request.headers.range;
                const range_value = /=(.*)$/.exec(range_header)[1];
                const range = /^[\w]*?(\d*)-(\d*)$/.exec(range_value);
                assert(stats.size <= Number.MAX_SAFE_INTEGER);
                let start = range[1] ? parseInt(range[1]) : undefined;
                let end = range[2] ? parseInt(range[2]) : Number(stats.size) - 1;
                if (start === undefined) {
                    start = Number(stats.size) - end;
                    end = Number(stats.size) - 1;
                }
                const chunksize = (end - start + 1);
                response.status = 206;
                response.set('content-length', String(chunksize));
                response.set('content-range', `bytes ${start}-${end}/${stats.size}`);
                response.body = createReadStream(path, { start, end });
            }
            catch (err) {
                response.status = 416;
                response.set('content-length', String(stats.size));
                response.set('content-range', `bytes */${stats.size}`);
                response.body = createReadStream(path);
            }
        else {
            response.set('content-length', String(stats.size));
            response.body = createReadStream(path);
        }
        return path;
    }
}
//# sourceMappingURL=server.js.map