export declare const noop: () => void;
export declare function assert(assertion: any, message?: string): never | void;
export declare function delay(milliseconds: number): Promise<void>;
export declare function timeout(milliseconds: number): Promise<void>;
/** https://stackoverflow.com/questions/63297164/how-to-only-accept-arraybuffer-as-parameter */
export type StrictArrayBuffer = ArrayBuffer & {
    buffer?: undefined;
};
export interface Deferred<TValue> extends Promise<TValue> {
    resolve(value: TValue | PromiseLike<TValue>): void;
    reject(reason?: Error): void;
}
/** 创建一个 promise，后续可调用 promise.resolve, promise.reject 方法设置其状态和值
    - initial?: `undefined` 传入非 undefined 值（包括 null）时直接设置为 resolved 状态
    注: 下面的方法不能标记为 aysnc function, 否则会对返回值再做一层 Promise.resolve() 导致 reject, resolve 属性丢失 */
export declare function defer<TValue>(initial?: TValue): Deferred<TValue>;
export interface LockedAction<TResource, TResult> {
    (resource: TResource): TResult | Promise<TResult>;
}
/** @example
    let lock = new Lock(redis)
    
    // 锁定资源后在 action 回调中操作资源，回调 promise 完成后自动释放资源，
    // 三秒内未成功独占资源直接抛出 TimeoutError (开始执行后不受 signal abort 影响)
    await lock.request(async redis => {
        const value = await redis.get('key')
        await redis.set('key', value * 2)
    }, AbortSignal.timeout(3000))
    
    参考:
    https://developer.mozilla.org/en-US/docs/Web/API/Web_Locks_API
    https://github.com/metarhia/web-locks
    https://www.npmjs.com/package/async-lock  */
export declare class Lock<TResource = void> {
    /** 如果操作不需要独占资源，可以直接通过 lock.resource 访问，否则需要通过 await lock.request() 独占资源后再访问 */
    resource: TResource;
    /** 等待链，新的 await lock.request() 调用会等待当前等待链尾部的 promise 完成，并作为新的尾部 */
    ptail: Deferred<void>;
    /** 查询当前资源是否属于被锁定的状态，方便在资源闲置时做一些可选操作（操作前仍需锁定），或者做一些状态展示 */
    locked: boolean;
    /** 可以不传 resource，表示管理某个抽象或虚拟的资源 */
    constructor(resource?: TResource);
    /** 通过 await lock.request() 锁定资源以便独占访问
        成功返回之后由调用方负责调用 release 方法释放资源
        在 signal aborted 时抛出错误结束等待，且由内部实现自动释放资源 */
    request<TResult>(action: LockedAction<TResource, TResult>, signal?: AbortSignal): Promise<TResult>;
}
export declare function debug(milliseconds?: number): Promise<void>;
/** 字符串字典序比较 */
export declare function strcmp(l: string, r: string): 0 | 1 | -1;
/** 拼接 TypedArrays 生成一个完整的 Uint8Array */
export declare function concat(arrays: ArrayBufferView[]): Uint8Array;
/** 时间间隔 (milliseconds) 格式化 */
export declare function delta2str(delta: number): string;
/** generate random id */
export declare function genid(): number;
export declare class Timer {
    started: number;
    ended: number;
    /** 停止秒表，保存读数 */
    stop(): void;
    /** 如果秒表未停止，获取当前秒表读数；
        如果秒表已停止，获取停止时的秒表读数; */
    get(): number;
    getstr(): string;
    print(): void;
}
