import { Stream, Writable, Transform } from 'stream';
import util from 'util';
import omit from 'lodash/omit.js';
import { t } from './i18n/instance.js';
import './prototype.js';
export const noop = () => { };
/** `230` term 字符宽度 （实际上有 240） */
export const output_width = 230;
export function set_inspect_options(colors = true) {
    util.inspect.defaultOptions.maxArrayLength = 40;
    util.inspect.defaultOptions.maxStringLength = 10000;
    util.inspect.defaultOptions.breakLength = output_width;
    util.inspect.defaultOptions.colors = colors;
    util.inspect.defaultOptions.compact = false;
    util.inspect.defaultOptions.getters = true;
    util.inspect.defaultOptions.depth = 2;
    util.inspect.defaultOptions.sorted = false;
    util.inspect.defaultOptions.showProxy = true;
    util.inspect.styles.number = 'green';
    util.inspect.styles.string = 'cyan';
    util.inspect.styles.boolean = 'blue';
    util.inspect.styles.date = 'magenta';
    util.inspect.styles.special = 'white';
}
export function assert(assertion, message) {
    if (!assertion) {
        debugger;
        throw Object.assign(new Error(`断言失败: ${message ? `${message}: ` : ''}${inspect(assertion, { colors: false, compact: true })}`), { assertion });
    }
}
export function dedent(templ, ...values) {
    let strings = Array.from(typeof templ === 'string' ? [templ] : templ.raw);
    // 1. remove trailing whitespace
    strings[strings.length - 1] = strings[strings.length - 1].replace(/\r?\n([\t ]*)$/, '');
    // 2. find all line breaks to determine the highest common indentation level
    const indent_lengths = strings.reduce((arr, str) => {
        const matches = str.match(/\n[\t ]+/g);
        if (matches)
            return arr.concat(matches.map(match => match.length - 1));
        return arr;
    }, []);
    // 3. remove the common indentation from all strings
    if (indent_lengths.length) {
        const pattern = new RegExp(`\n[\t ]{${Math.min(...indent_lengths)}}`, 'g');
        strings = strings.map(str => str.replace(pattern, '\n'));
    }
    // 4. remove leading whitespace
    strings[0] = strings[0].replace(/^\r?\n/, '');
    // 5. perform interpolation
    let string = strings[0];
    values.forEach((value, i) => {
        string += value + strings[i + 1];
    });
    string += '\n';
    return string;
}
/** 数组或 iterable 去重（可按 selector 去重）
    - selector?: 可以是 key (string) 或 (obj: any) => any
*/
export function unique(iterable, selector) {
    if (!selector)
        return [...new Set(iterable)];
    let map = new Map();
    const is_str_selector = typeof selector === 'string';
    for (const x of iterable)
        map.set(is_str_selector ? x[selector] : selector(x), x);
    return [...map.values()];
}
/** 排序对象中 key 的顺序，返回新的对象 */
export function sort_keys(obj) {
    return Object.fromEntries(Object.entries(obj)
        .sort(([key_l], [key_r]) => strcmp(key_l, key_r)));
}
/** 字符串字典序比较 */
export function strcmp(l, r) {
    if (l === r)
        return 0;
    if (l < r)
        return -1;
    return 1;
}
/** 拼接 TypedArrays 生成一个完整的 Uint8Array */
export function concat(arrays) {
    let length = 0;
    for (const a of arrays)
        length += a.byteLength;
    let buf = new Uint8Array(length);
    let offset = 0;
    for (const a of arrays) {
        const uint8view = new Uint8Array(a.buffer, a.byteOffset, a.byteLength);
        buf.set(uint8view, offset);
        offset += uint8view.byteLength;
    }
    return buf;
}
export function typed_array_to_buffer(view) {
    return Buffer.from(view.buffer, view.byteOffset, view.byteLength);
}
/** 时间间隔 (milliseconds) 格式化 */
export function delta2str(delta) {
    if (delta < 1)
        return '0 ms';
    // [1, 1000) ms
    if (delta < 1000)
        return `${delta} ms`;
    // 1.123 s
    if (1000 <= delta && delta < 1000 * 60)
        return `${(delta / 1000).toFixed(1)} s`;
    // 1 min 12 s [1 min 0s, 60 min)
    const seconds = Math.trunc(delta / 1000);
    if (seconds < 60 * 60)
        return `${Math.trunc(seconds / 60)} min ${seconds % 60} s`;
    const hour = Math.trunc(seconds / 3600);
    return `${hour} h ${Math.trunc((seconds - 3600 * hour) / 60)} min ${seconds % 60} s`;
}
/** generate random id */
export function genid() {
    return Math.random() * 2 ** 53;
}
export class Timer {
    started = new Date().getTime();
    ended;
    /** 停止秒表，保存读数 */
    stop() {
        this.ended = new Date().getTime();
    }
    /** 如果秒表未停止，获取当前秒表读数；
        如果秒表已停止，获取停止时的秒表读数; */
    get() {
        return (this.ended || new Date().getTime()) - this.started;
    }
    /** 获取时间表示字符串，如 1.2 s
        - bracket?: `true` 字符串前后加上括号，如 (1.2 s) */
    getstr(bracket = false) {
        let s = delta2str(this.get());
        if (bracket)
            return s.bracket();
        else
            return s;
    }
    print() {
        console.log(this.getstr(true));
    }
}
export function log_line() {
    console.log('---');
}
export async function delay(milliseconds) {
    return new Promise(resolve => {
        setTimeout(resolve, milliseconds);
    });
}
export async function timeout(milliseconds) {
    const error = new Error(t('超时错误'));
    await delay(milliseconds);
    throw error;
}
/** 创建一个 promise，后续可调用 promise.resolve, promise.reject 方法设置其状态和值
    - initial?: `undefined` 传入非 undefined 值（包括 null）时直接设置为 resolved 状态
    注: 下面的方法不能标记为 aysnc function, 否则会对返回值再做一层 Promise.resolve() 导致 reject, resolve 属性丢失 */
// eslint-disable-next-line @typescript-eslint/promise-function-async
export function defer(initial) {
    if (initial === undefined) {
        let resolve;
        let reject;
        let promise = new Promise((_resolve, _reject) => {
            resolve = _resolve;
            reject = _reject;
        });
        return Object.assign(promise, { resolve, reject });
    }
    else
        return Object.assign(Promise.resolve(initial), {
            resolve() { },
            reject() { }
        });
}
/** @example
    let lock = new Lock(redis)
    
    // 锁定资源后在 action 回调中操作资源，回调 promise 完成后自动释放资源，
    // 三秒内未成功独占资源直接抛出 TimeoutError (开始执行后不受 signal abort 影响)
    await lock.request(async (redis) => {
        const value = await redis.get('key')
        await redis.set('key', value * 2)
    }, AbortSignal.timeout(3000))
    
    参考:
    https://developer.mozilla.org/en-US/docs/Web/API/Web_Locks_API
    https://github.com/metarhia/web-locks
    https://www.npmjs.com/package/async-lock  */
export class Lock {
    /** 如果操作不需要独占资源，可以直接通过 lock.resource 访问，否则需要通过 await lock.request() 独占资源后再访问 */
    resource;
    /** 等待链，新的 await lock.request() 调用会等待当前等待链尾部的 promise 完成，并作为新的尾部 */
    ptail = defer(null);
    /** 查询当前资源是否属于被锁定的状态，方便在资源闲置时做一些可选操作（操作前仍需锁定），或者做一些状态展示 */
    locked = false;
    /** 可以不传 resource，表示管理某个抽象或虚拟的资源 */
    constructor(resource) {
        this.resource = resource;
    }
    /** 通过 await lock.request() 锁定资源以便独占访问
        成功返回之后由调用方负责调用 release 方法释放资源
        在 signal aborted 时抛出错误结束等待，且由内部实现自动释放资源 */
    // async request (action: (resource: TResource) => Promise<void>, signal?: AbortSignal): Promise<void>
    // async request <TResult> (action: <TResult> (resource: TResource) => Promise<TResult>, signal?: AbortSignal): Promise<TResult>
    async request(action, signal) {
        signal?.throwIfAborted();
        const ptail = this.ptail;
        let pcurrent = this.ptail = defer();
        this.locked = true;
        return new Promise((resolve, reject) => {
            // 下面两种情况，先发生的决定 request 返回的 promise 状态
            // 不管是否 aborted, 都要等资源被前一次调用释放，先 aborted 只是将控制权交回给调用者，在锁定资源后不执行 action， 直接释放
            /** 防止执行过程中 signal abort */
            let executing = false;
            signal?.addEventListener('abort', () => {
                if (!executing)
                    // 这里不能释放锁，需要等 ptail resolve 后拿到资源再释放
                    reject(signal.reason);
            }, { once: true });
            ptail.then(async () => {
                // 这里已经能保证独占访问资源
                // 如果 aborted, 可以理解为独占资源失败，调用者不会使用资源，直接释放
                if (signal?.aborted)
                    reject(signal.reason);
                else
                    // 由调用者去操作资源
                    try {
                        executing = true;
                        resolve(await action(this.resource));
                    }
                    catch (error) {
                        reject(error);
                    }
                // 下面开始释放锁
                // assert(this.locked)
                if (!this.locked)
                    reject(new Error(t('尝试释放未锁定的锁，这不应该发生')));
                this.locked = false;
                pcurrent.resolve();
            });
        });
    }
}
// ------------ text
export function has_chinese(str) {
    return /[\u4E00-\u9FA5]/.test(str);
}
export function escape_line_feed(str) {
    return str.replace(/\n/g, '\\n');
}
/** util.inspect(obj)
    - options
        - limit?: `10000`
        - omit?: string[] */
export function inspect(obj, options = {}) {
    if (options.omit)
        obj = omit(obj, [inspect.custom, ...(options.omit || [])]);
    let text = util.inspect(obj, options);
    if (!('limit' in options))
        options.limit = 10000;
    if (options.limit && text.length > options.limit)
        return `${text.slice(0, options.limit)}……'\u001b[39m\n`;
    else
        return text;
}
(function (inspect) {
    inspect.custom = util.inspect.custom;
})(inspect || (inspect = {}));
// ------------------------------------ stream
/** npm map-stream
    filter will reemit the data if cb(err,pass) pass is truthy
    
    reduce is more tricky
    maybe we want to group the reductions or emit progress updates occasionally
    the most basic reduce just emits one 'data' event after it has recieved 'end'
    
    create an event stream and apply function to each .write,
    emitting each response as data unless it's an empty callback
 */
export function map_stream(mapper, options) {
    options = options || {};
    let inputs = 0, outputs = 0, ended = false, paused = false, destroyed = false, last_written = 0, in_next = false;
    let stream = Object.assign(new Stream(), {
        readable: true,
        writable: true,
        write(data) {
            if (ended)
                throw new Error('map stream is not writable');
            in_next = false;
            inputs++;
            try {
                // catch sync errors and handle them like async errors
                const written = wrapped_mapper(data, inputs, next);
                paused = (written === false);
                return !paused;
            }
            catch (err) {
                // if the callback has been called syncronously, and the error has occured in an listener, throw it again.
                if (in_next)
                    throw err;
                next(err);
                return !paused;
            }
        },
        end(data) {
            if (ended)
                return;
            _end(data);
        },
        destroy() {
            ended = destroyed = true;
            stream.writable = stream.readable = paused = false;
            process.nextTick(function () {
                stream.emit('close');
            });
        },
        pause() {
            paused = true;
        },
        resume() {
            paused = false;
        }
    });
    let error_event_name = options.failures ? 'failure' : 'error';
    // Items that are not ready to be written yet (because they would come out of order) get stuck in a queue for later.
    let write_queue = {};
    function queue_data(data, number) {
        let next_to_write = last_written + 1;
        if (number === next_to_write) {
            // If it's next, and its not undefined write it
            if (data !== undefined)
                stream.emit('data', data);
            last_written++;
            next_to_write++;
        }
        else
            // Otherwise queue it for later.
            write_queue[number] = data;
        // If the next value is in the queue, write it
        if (Object.prototype.hasOwnProperty.call(write_queue, next_to_write)) {
            let data_to_write = write_queue[next_to_write];
            delete write_queue[next_to_write];
            return queue_data(data_to_write, next_to_write);
        }
        outputs++;
        if (inputs === outputs) {
            if (paused) {
                paused = false;
                stream.emit('drain'); // written all the incoming events
            }
            if (ended)
                _end();
        }
    }
    function next(err, data, number) {
        if (destroyed)
            return;
        in_next = true;
        if (!err || options.failures)
            queue_data(data, number);
        if (err)
            stream.emit(error_event_name, err);
        in_next = false;
    }
    /** Wrap the mapper function by calling its callback with the order number of the item in the stream. */
    function wrapped_mapper(input, number, callback) {
        return mapper.call(null, input, function (err, data) {
            callback(err, data, number);
        });
    }
    function _end(data) {
        // if end was called with args, write it, 
        ended = true; // write will emit 'end' if ended is true
        stream.writable = false;
        if (data !== undefined)
            return queue_data(data, inputs);
        else if (inputs === outputs) { // wait for processing 
            stream.readable = false;
            stream.emit('end');
            stream.destroy();
        }
    }
    return stream;
}
export async function stream_to_buffer(stream) {
    let chunks = [];
    for await (const chunk of stream)
        chunks.push(chunk);
    return Buffer.concat(chunks);
}
export async function* stream_to_lines(stream) {
    let buf = '';
    for await (const chunk of stream) {
        let i = 0, j = 0;
        for (; (i = chunk.indexOf('\n', j)) >= 0;) {
            let line = chunk.slice(j, i);
            if (buf) {
                line = buf + line;
                buf = '';
            }
            j = i + 1;
            yield line;
        }
        buf = chunk.slice(j);
    }
}
export class WritableMemoryStream extends Writable {
    chunks = [];
    pbuffer = defer();
    constructor() {
        super({ highWaterMark: 2 ** 30, decodeStrings: false });
    }
    _write(chunk, encoding, callback) {
        this.chunks.push(chunk);
        callback();
    }
    _writev(chunks, callback) {
        for (const { chunk } of chunks)
            this.chunks.push(chunk);
        callback();
    }
    _final(callback) {
        this.pbuffer.resolve(Buffer.concat(this.chunks));
        this.chunks = null;
        callback();
    }
}
/** 用 TextDecoder 通过 .pipe 方法持续的解码 process.stdout 等流，
    主要在 process.ts#start 中使用
    https://nodejs.org/api/stream.html#implementing-a-transform-stream */
export class DecoderStream extends Transform {
    decoder;
    constructor(encoding) {
        super({ encoding: 'utf-8' });
        this.decoder = new TextDecoder(encoding);
    }
    _transform(chunk, encoding, callback) {
        assert(chunk instanceof Uint8Array);
        const str = this.decoder.decode(chunk, { stream: true });
        if (str.length)
            this.push(str, 'utf-8');
        callback();
    }
    _flush(callback) {
        const str = this.decoder.decode();
        if (str.length)
            this.push(str, 'utf-8');
        callback();
    }
}
//# sourceMappingURL=utils.js.map